/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc    CORBAGeneral
///////////////////////////////////////////////////////////////////////////////

#ifndef _CORBAHelper_inl
#define _CORBAHelper_inl

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <definition.h>
#include <CyrException.h>
///////////////////////////////////////////////////////////////////////////////
// CORBA includes
#include <omniORB2/CORBA.h>
#include <omniORB2/Naming.hh>
///////////////////////////////////////////////////////////////////////////////
//Standard includes
#include <string>
using std::string;
#ifdef typename
#undef typename
#endif
///////////////////////////////////////////////////////////////////////////////
// @func    This template function narrows CORBA objects to be of a more
//          specific type.  If any CORBA exceptions are thrown they are caught
//          and converted to be a Cyrano exception.  If the object cannot be
//          narrowed to the requested type an exception is thrown as well.
// @rdesc   A reference to the type of object we are narrowing to.
//
// @tfarg   The type of object we are narrowing to.
// @parm    The CORBA object to be narrowed.
//
// @devnote This function relies on the newly added constructs _ptr_type
//          and _var_type.  If they are not available in the ORB mapping
//          being used then they can be added quite easily as typedefs within
//          the class declaration.  The function could be written so as not to
//          use these constructs, but it's not as elegant a solution.
//           
//          e.g <pre>typedef NamingContext_ptr _ptr_type</pre>
template<class T>
inline typename T::_ptr_type narrow_object( T* dummy, CORBA::Object_ptr obj )
{
    typename T::_var_type ref;
	const char* where = "narrow_object";
	const char* error = NULL;

    // Check to see if object is nil to start with
    if ( CORBA::is_nil( obj ))
        error = "Object reference is nil";
    else
    {
        try
        {
            // Narrow object to specific class
            ref  = T::_narrow( obj );
        }
        catch( CORBA::Exception )
        {
            error = "CORBA exception whilst narrowing reference";
        }
        
        // Check to see if object narrowed correctly.  If not then throw an
        // exception.
        if ( CORBA::is_nil( ref ) )
            error = "Reference has incorrect type";
    }

	if ( error != NULL )
		throw CyrCORBAException( error, where, CYR_CORBA_NARROW_FAILED );

	// Return pointer as well.
	return ref._retn();
}

///////////////////////////////////////////////////////////////////////////////
// @func    This template function resolves an initial reference to a named
//          CORBA object such as the NameService or the EventService.  If the
//          requested object cannot be found then an exeption is thrown.
//          Strictly speaking this function should be a method of the
//          <c CORBAGeneral> class, but the required template specialisation is
//          not available in all compilers.
// @rdesc   A reference to the type of object we are looking for.
//
// @tfarg   The type of object we are looking for.
// @parm    A reference to the currently running ORB.
// @parm    A string that represents the object we are looking for.
//
// @devnote This function relies on the newly added constructs _ptr_type
//          and _var_type.  If they are not available in the ORB mapping
//          being used then they can be added quite easily as typedefs within
//          the class declaration.  The function could be written so as not to
//          use these constructs, but it's not as elegant a solution.
//           
//          e.g <pre>typedef NamingContext_ptr _ptr_type</pre>
template<class T>
inline typename T::_ptr_type resolve_init(T* dummy, CORBA::ORB_ptr orb, const char* id )
{
	const char* where = "resolve_init";
	string error;

	CORBA::Object_var obj;
	try
	{
		// Obtain a reference to the named object.
		obj = orb->resolve_initial_references( id );

		// Narrow object returned to be of specified type
		return narrow_object(dummy, (CORBA::Object_ptr)obj );
	}
	catch( CORBA::Exception )
	{
		// Build error string
		error = "Could not connect to ";
		error += id;
		throw CyrCORBAException( error.c_str(), where, CYR_CORBA_OBJECT_NOT_FOUND );
	}

	// Return a nil reference
	return T::_nil();
}		
//NO LONGER REQUIRED now use CORBAExString this requires no dynamic casting and therefore no RTTI
//(negligible overhead but seeing as where it is used we a perfectly safe with static casting see no need to keep it)
///////////////////////////////////////////////////////////////////////////////
// @func    Given a CORBA exception this method will return an erro message
//          that is suitable for use in a standard Cyrano exception.
// @rdesc   The error message.
// @parm    The CORBA exception.
/*inline const char* CorbaExceptionString( CORBA::Exception& ex )
{
	// Default error message
	const char* error = "Unknown CORBA exception";

	// Try to determine an error code by using rtti
	if ( dynamic_cast<CORBA::UserException*>(&ex) != 0 )
		error = "Unknown CORBA user exception";
	else if ( dynamic_cast<CORBA::BAD_PARAM*>(&ex) != 0 )
		error = "CORBA method was passed an invalid parameter";
	else if ( dynamic_cast<CORBA::UNKNOWN*>(&ex) != 0 )
		error = "Unknown CORBA system exception";
	else if ( dynamic_cast<CORBA::NO_MEMORY*>(&ex) != 0 )
		error = "CORBA NO_MEMORY exception";
	else if ( dynamic_cast<CORBA::IMP_LIMIT*>(&ex) != 0 )
		error = "CORBA IMP_LIMIT exception";
	else if ( dynamic_cast<CORBA::COMM_FAILURE*>(&ex) != 0 )
		error = "CORBA COMM_FAILURE exception";
	else if ( dynamic_cast<CORBA::INV_OBJREF*>(&ex) != 0 )
		error = "CORBA INV_OBJREF exception";
	else if ( dynamic_cast<CORBA::OBJECT_NOT_EXIST*>(&ex) != 0 )
		error = "CORBA OBJECT_NOT_EXIST exception";
	else if ( dynamic_cast<CORBA::NO_PERMISSION*>(&ex) != 0 )
		error = "CORBA NO_PERMISSION exception";
	else if ( dynamic_cast<CORBA::INTERNAL*>(&ex) != 0 )
		error = "CORBA INTERNAL exception";
	else if ( dynamic_cast<CORBA::MARSHAL*>(&ex) != 0 )
		error = "CORBA MARSHAL exception";
	else if ( dynamic_cast<CORBA::INITIALIZE*>(&ex) != 0 )
		error = "CORBA INITIALIZE exception";
	else if ( dynamic_cast<CORBA::NO_IMPLEMENT*>(&ex) != 0 )
		error = "CORBA NO_IMPLEMENT exception";
	else if ( dynamic_cast<CORBA::BAD_TYPECODE*>(&ex) != 0 )
		error = "CORBA BAD_TYPECODE exception";
	else if ( dynamic_cast<CORBA::BAD_OPERATION*>(&ex) != 0 )
		error = "CORBA BAD_OPERATION exception";
	else if ( dynamic_cast<CORBA::NO_RESOURCES*>(&ex) != 0 )
		error = "CORBA NO_RESOURCES exception";
	else if ( dynamic_cast<CORBA::NO_RESPONSE*>(&ex) != 0 )
		error = "CORBA NO_RESPONSE exception";
	else if ( dynamic_cast<CORBA::PERSIST_STORE*>(&ex) != 0 )
		error = "CORBA PERSIST_STORE exception";
	else if ( dynamic_cast<CORBA::BAD_INV_ORDER*>(&ex) != 0 )
		error = "CORBA BAD_INV_ORDER exception";
	else if ( dynamic_cast<CORBA::TRANSIENT*>(&ex) != 0 )
		error = "CORBA TRANSIENT exception";
	else if ( dynamic_cast<CORBA::FREE_MEM*>(&ex) != 0 )
		error = "CORBA FREE_MEM exception";
	else if ( dynamic_cast<CORBA::INV_IDENT*>(&ex) != 0 )
		error = "CORBA INV_IDENT exception";
	else if ( dynamic_cast<CORBA::INV_FLAG*>(&ex) != 0 )
		error = "CORBA INV_FLAG exception";
	else if ( dynamic_cast<CORBA::INTF_REPOS*>(&ex) != 0 )
		error = "CORBA INTF_REPOS exception";
	else if ( dynamic_cast<CORBA::BAD_CONTEXT*>(&ex) != 0 )
		error = "CORBA BAD_CONTEXT exception";
	else if ( dynamic_cast<CORBA::OBJ_ADAPTER*>(&ex) != 0 )
		error = "CORBA OBJ_ADAPTER exception";
	else if ( dynamic_cast<CORBA::DATA_CONVERSION*>(&ex) != 0 )
		error = "CORBA DATA_CONVERSION exception";
	else if ( dynamic_cast<CORBA::TRANSACTION_REQUIRED*>(&ex) != 0 )
		error = "CORBA TRANSACTION_REQUIRED exception";
	else if ( dynamic_cast<CORBA::TRANSACTION_ROLLEDBACK*>(&ex) != 0 )
		error = "CORBA TRANSACTION_ROLLEDBACK exception";
	else if ( dynamic_cast<CORBA::INVALID_TRANSACTION*>(&ex) != 0 )
		error = "CORBA INVALID_TRANSACTION exception";
	else if ( dynamic_cast<CORBA::WRONG_TRANSACTION*>(&ex) != 0 )
		error = "CORBA WRONG_TRANSACTION exception";
	else if ( dynamic_cast<CORBA::SystemException*>(&ex) != 0 )
		error = "Unknown CORBA system exception";

	return error;	
}*/
///////////////////////////////////////////////////////////////////////////////
// @func Convert CORBA::Exception so that a text string can be output to ostreams
//	Based on code from 
//	"Advanced CORBA Programming with C++(Michi Henning,Steve Vinoski)"
static inline std::ostream& operator<<(std::ostream &os,const CORBA::Exception &e)
{   
    CORBA::Any anytmp;
    anytmp<<=e;
    CORBA::TypeCode_var tc = anytmp.type();
    const char* p=tc->name();
    if(*p != '\0' )
	os << p;
    else
	os << tc->id();
    return os;
}
///////////////////////////////////////////////////////////////////////////////
// @func Convert CORBA::Exception to a std::string
//	Based on code from 
//	"Advanced CORBA Programming with C++(Michi Henning,Steve Vinoski)"
inline const char* CorbaExceptionString( const CORBA::Exception& e )
{   
    CORBA::Any anytmp;
    anytmp<<=e;
    CORBA::TypeCode_var tc = anytmp.type();
    CORBA::String_var p=tc->name();
    if(strcmp("",(const char*)p)==0)
	p=tc->id();
    return CORBA::string_dup(p);
}
inline string CORBAExString(const CORBA::Exception &e)
{   
    CORBA::Any anytmp;
    anytmp<<=e;
    CORBA::TypeCode_var tc = anytmp.type();
    string p=tc->name();
    if(p.empty())
	p=tc->id();
    return p;
}

#endif


