/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbClass.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor is responsible for creating a CORBA helper class
//          and for connecting to a remote Factory.  This Factory must be
//          one that can create OpenSTARep interface objects.
// @parm    Pointer to a CORBA helper class.
// @parm    The name of the Factory as registered with the Name Service.
CECOrbClass::CECOrbClass( CORBAServer* corba, const char* factory )
{
    // Error information
    const char* where = "CECOrbClass::CECOrbClass";

    try
    {
        // Create CORBA functionality class
        m_corba = corba;

        // Connect to Factory
        CORBA::Object_ptr obj = m_corba->GetObjectFromClass( Factory::CLASS_NAME, factory );
        m_factory = narrow_object( (Factory*)NULL, obj );
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor should be used when Factory and CORBA helper
//          objects are already available.
// @parm    Pointer to a CORBA helper class.
// @parm    Reference to a Factory.
CECOrbClass::CECOrbClass( CORBAServer* corba, Factory_ptr factory )
{
    // Assign member variables
    m_corba = corba;
    m_factory = factory;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   As well as tidying up any allocated memory, the destructor is
//          responsible for disposing of the CORBA object in the Factory.
CECOrbClass::~CECOrbClass()
{
    //TBD call the factory to _dispose of the underlying object(or have ping method)
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the Factory object that maintains
//          the underlying servant object.  This reference should be used
//          to maintain the underlying servant object and can also be used
//          to create other OpenSTARep interfaces.  This call will increment the
//          reference count and therefore the returned value should be
//          released.
// @rdesc   The reference to the servant object.

Factory_ptr CECOrbClass::GetFactory()
{
    return Factory::_duplicate( m_factory );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a pointer to the CORBA helper class,
//          CORBAServer, that exists as a member within this class.  This will
//          prevent having to create more than one of these.
// @rdesc   A pointer to the CORBA helper class.
CORBAServer* CECOrbClass::GetHelper()
{
    return m_corba;
}
