/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc Factory
///////////////////////////////////////////////////////////////////////////////

#ifndef _Factory_i_h
#define _Factory_i_h

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CORBA/CORBAServer.h>
#include <Factory/Factory.h>
#include <Factory/ProviderReg.h>
#include <Factory/Provider.h>

///////////////////////////////////////////////////////////////////////////////
// @class   This class represents the main Factory object.  It maintains a list
//          of Providers and classes.  Client requests are received by this
//          object and passed on to the relevant Provider for handling.
//
//          This class is derived from a CORBA IDL generated class.  It
//          implements all the relevant functionality and registers itself with
//          the NameService.  There should be no need to derive further from
//          this class.
//
//          A single exception type is thrown by this class at the server end -
//          <c CyrCORBAException>.  Any CORBA exceptions that are thrown are
//          standard CORBA exceptions and are not generated by this code
//          directly.  All other exceptions are trapped internally
//          and converted to be of the type <c CyrCORBAException>.
// @base    public | _sk_Factory
//
// @ex      The following example demonstrates the correct usage for starting
//          a factory: |
//          int main( int argc, char** argv )
//          {
//              // Initialise ORB and BOA
//              CORBA::ORB_ptr orb = CORBA::ORB_Init( argc, argv, "omniORB2" );
//              CORBA::BOA_ptr boa = orb->BOA_Init( argc, argv, "omniORB2_BOA" );
//
//              // Create Factory
//              FactoryImpl* factory = new FactoryImpl( "ACME", orb );
//
//              ...
//
//              // Dispose of Factory and tidy up CORBA references
//              factory->_dispose();
//              CORBA::release( boa );
//              CORBA::release( orb );
//              return 1;
//          }
//
// @ex      This example demonstrates a standard method for accessing a
//          Factory as a client application: |
//          int main( int argc, char** argv )
//          {
//              // Initialise ORB and BOA
//              CORBAClient* pClient = new CORBAClient( argc, argv );
//
//              // Connect to Factory
//              CORBA::Object_ptr obj = pClient->GetObjectFromClass( Factory::CLASS_NAME, "ACME" );
//              Factory_ptr factory = pClient->Narrow( &factory, obj );
//
//              ...
//
//              // Release Factory reference
//              CORBA::release( factory );
//              return 1;
//          }
class FactoryImpl : public virtual _sk_Factory
{
    // @access Public Members
public:
    FactoryImpl( CORBAServer* pServer, const char* name );  // @cmember Constructor.
    virtual ~FactoryImpl();               // @cmember Destructor.

    // @access IDL Generated Methods
    CORBA::Object_ptr New( ParamList_ptr params );    // @cmember Creates a new CORBA object.
    void Dispose( const char* cls, const char* name );          // @cmember Disposes of an existing object.

    void Register( const char* cls, const char* provider, long type );  // @cmember Registers a single class.
    void Deregister( const char* cls, const char* provider );   // @cmember Deregisters a single class.
    void DeregisterAll( const char* provider );                 // @cmember Deregisters all classes.

    char* GetFirstClass();                                      // @cmember Retrieves a class name.
    char* GetFirstClassFromProvider( const char* provider );    // @cmember Retrieves a class name.
    char* GetFirstClassFromType( long type );                   // @cmember Retrieves a class name.
    char* GetFirstProvider();                                   // @cmember Retrieves a Provider name.
    char* GetFirstProviderFromClass( const char* cls );         // @cmember Retrieves a Provider name.
    char* GetNext();                                            // @cmember Gets next entry in search list.

    void Reaper();      // @cmember Reaps all dead objects.
    void Borrow( const char* cls, const char* name, long period );      // @cmember Extends lease on a particular object.

    // @cmember Logs an event
    void LogEvent( long level, const char* event );
    void Shutdown();

    // @access Private Members
private:
    Provider_ptr    GetProvider( string name ) const;      // @cmember Gets Provider based on its name.

    ProviderReg     m_Reg;          // @cmember Provider/class list.
    CORBAServer*    m_pServer;      // @cmember CORBA server functionality.
    string      m_name;         // @cmember The name of the Factory.
}
;

#endif
