/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Provider
///////////////////////////////////////////////////////////////////////////////

#ifndef _Provider_i_h
#define _Provider_i_h

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CORBA/CORBAServer.h>
#include <Factory/Factory.h>

///////////////////////////////////////////////////////////////////////////////
#include <deque>
using std::deque;
///////////////////////////////////////////////////////////////////////////////
// The following constants are used to represent the types of class that a
// Provider can provide. They should be used as the 'type' parameter for the
// registration method.
//
// @const long | PROVIDER_DATA | This value indicates a data object.
// @const long | PROVIDER_INITIATOR | This value indicates an initiator object.
// @const long | PROVIDER_MANAGER | This value indicates a manager object.
// @const long | PROVIDER_EXECUTOR | This value indicates an executor object.
const long PROVIDER_DATA         =  0;
const long PROVIDER_INITIATOR    =  1;
const long PROVIDER_MANAGER      =  2;
const long PROVIDER_EXECUTOR     =  3;

///////////////////////////////////////////////////////////////////////////////
// Typedefs
// @struct  LEASED_OBJECT | This typedef is a structure that holds a pointer to
//          a CORBA object and also the period of time that that object will
//          exist for in minutes.
// @field   CORBA::Object_var | obj | Pointer to the CORBA object.
// @field   string | name | The name of the CORBA object.
// @field   string | cls | The class of the CORBA object.
// @field   long | period | Length of time remaining for object in minutes.
typedef struct tagsLEASED_OBJECT
{
    CORBA::Object* obj;
    string name;
    string cls;
    long period;
}
LEASED_OBJECT, *PLEASED_OBJECT;
// @type PLEASED_OBJECT | This type is a pointer to a <t LEASED_OBJECT> structure.

// @type ObjectList | This typedef is a list to hold CORBA objects.
typedef deque<PLEASED_OBJECT> ObjectList;

///////////////////////////////////////////////////////////////////////////////
// @class   This class represents the base class for a Provider component.
//          Provider components are responsible for creating and maintaining
//          CORBA objects within the architecture.  The Factory will
//          utilise these classes and make them appear as one entity.
//
//          This is a CORBA object that registers itself with the naming
//          service. Specific instances of a Provider should be derived from
//          here.  This class is a mechanism to hide the CORBA implementation
//          and contains any generic functionality.
//
//          The single method 'New' is declared as a pure virtual method and
//          must be overridden in a derived class.
//
//          A single exception type is thrown by this class -
//          <c CyrCORBAException>.  All other exceptions are trapped internally
//          and converted to be an exception of this type.
// @base    public | _sk_Provider
//
// @ex      This example demonstrates a typical implementation of the
//          <mf ProviderImpl::New> method: |
//
//          class SampleProvider : public ProviderImpl
//          {
//              ...
//          public:
//              CORBA::Object_ptr New( ParamList_ptr params );
//              ...
//          };
//
//          CORBA::Object_ptr SampleProvider::New( ParamList_ptr params )
//          {
//              CORBA::String_var cls = params->GetValue( Factory::PARAM_CLASS );
//              CORBA::String_var name = params->GetValue( Factory::PARAM_NAME );
//              CORBA::String_var pd = params->GetValue( Factory::PARAM_PERIOD );
//              long period = atol( pd );
//
//              // Check to see which product we wish to create
//              if ( strcmp( cls, "Product1" ) == 0 )
//              {
//                  // Create Product, register it with NameService
//                  // and send reference back to Factory
//                  Product1Impl* p1 = new Product1Impl();
//                  m_pServer->RegisterObject( p1, cls, name );
//                  Store( p1, cls, name, period );
//                  return p1->_this();
//              }
//              else if ( strcmp( cls, "Product2" ) == 0 )
//              {
//                  // Create Product, register it with NameService
//                  // and send reference back to Factory
//                  Product2Impl* p2 = new Product2Impl();
//                  m_pServer->RegisterObject( p2, cls, name );
//                  Store( p2, cls, name, period );
//                  return p2->_this();
//              }
//
//              return CORBA::Object::_nil();
//          }
class ProviderImpl : public virtual _sk_Provider
{
    // @access Public Members
public:
    ProviderImpl( CORBAServer* pServer, const char* name, const char* owner );     // @cmember Constructor.
    virtual void Shutdown();

    // @access IDL Generated Methods
    void Dispose( const char* cls, const char* name );              // @cmember Disposes of existing object in Provider.
    void Reaper();      // @cmember Reaps all dead objects.
    void Borrow( const char* cls, const char* name, long period );  // @cmember Extends lease on a particular object.

    // @access Protected Members
protected:
    CORBAServer* m_pServer;  // @cmember CORBA server functionality.
    Factory_var  m_factory;  // @cmember Parent Factory reference

    // @cmember Stores a newly created object for later destruction.
    void Store( CORBA::Object* obj, const char* cls, const char* name, long period );
    // @cmember Registers a class type with the parent Factory.
    void Register( const char* cls, const char* provider, long type );
    // @cmember Deregisters a class type from parent Factory.
    void Deregister( const char* cls, const char* provider );
    // @cmember Deregisters all class types from parent Factory.
    void DeregisterAll( const char* provider );

    string GetName() const
    {
        return m_name;
    }
    ;         // @cmember Name accessor method

    // @cmember Locates an element in the object list
    size_t Find( size_t start, size_t stop, const char* cls, const char* name );
    // @cmember Removes a section of the list and destroys the objects
    void EmptyList( long start, long stop, bool reap );

    // @cmember Logs events to standard logging mechanism
    void LogEvent( long level, const char* event );

    // @cmember,mfunc  Returns underlying list of objects
    ObjectList* GetObjectList()
    {
        return &m_lstObj;
    }

    // @access Private Members
private:
    ObjectList      m_lstObj;       // @cmember List of all objects created by Provider.
    string       m_name;         // @cmember Name of Provider object.
}
;

#endif
