/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     SpawnProc
// @module  Spawn Process | This module contains the implementation for the
//          SpawnProcessImpl class.  This code relies on the CyrTools2 library
//          and the CyrUtils library.  There are sections of code within this
//          module that are necessarily platform specific and have been
//          #define'd depending on the build platform.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef _WIN32
#pragma warning( disable: 4275 )
#endif

#include <stdlib.h>
#include <iostream>
//
//    'C' format routines
//
extern "C"
{
#include "psl_macros.h"
#include "target_rtndef.h"
}

//
//    VMS specific includes
//
#ifdef VMS
extern "C"
{
#include "psl_rtndef.h"                 // PSL Routines 
}


#include <starlet.h>                    // VMS System Services
#include <itemlist.h>                   // Itemlists 
#pragma nostandard                      // Keep Alpha Happy!
#include <descrip.h>                    // Descriptors
#include <lnmdef.h>                     // Logical name symbols
#pragma standard
#endif

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Objects/SpawnProc_i.h>

#include <Logging/CECLog.h>

extern CECLog* flogger;


///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor is responsible for registering this object with the
//          NameService.
// @parm    Pointer to the CORBAServer class used to register object.
// @parm    The name of this object.
SpawnProcessImpl::SpawnProcessImpl( CORBAServer* pServer, const char* name )
{
    // Assign member variables
    m_pServer = pServer;
    m_name = name;

    // Register object with NameService
    //m_pServer->RegisterObject( this, SpawnProcess::CLASS_NAME, name );
    m_pServer->ReadyObject(this);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor is responsible for deregistering the object from the
//          NameService.
SpawnProcessImpl::~SpawnProcessImpl()
{
    // Deregister object
    //m_pServer->DeregisterObject( SpawnProcess::CLASS_NAME, m_name );
    delete m_process;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method waits until the particular process terminates.
// @rdesc   Returns -1 or the exit status of the process.
long SpawnProcessImpl::Wait()
{
    return m_process->wait();
}

///////////////////////////////////////////////////////////////////////////////
// The following section of code is WIN32 specific.
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method starts a new process on the current machine.
// @rdesc   The process ID.
// @parm    The command line to execute.
long SpawnProcessImpl::Start (const char* image_ptr )
{
    m_process = new CTProcess( image_ptr, FALSE );
    m_process->start();

#ifdef _WIN32
    return m_process->getActualPid();
#else
    return m_process->getPid();
#endif
}

///////////////////////////////////////////////////////////////////////////////
//
// @mfunc   This method starts a new process on the current machine.
// @rdesc   The process ID.
// @parm    ID of the target for whom the executable is to be started
// @parm    ID of the image to be executed
// @parm    The name of the product (directory beneath CYRANO_HOME) in which the image resides
// @parm    A list of the arguments to be passed to the image on the command line
// @parm    A flag to show ignore login information
// @parm    Username to log in under
// @parm    Password to log in under
// @rdesc   As returned from Start
//
// Exceptions : throws CyrSpawnException if the command line cannot be formatted
//
long SpawnProcessImpl::StartProductImage    (      long  TargetID     ,long  TargetImage
        ,const char* pProduct     ,const char* pParameters)

{
    //
    //    As we came from a CORBA call we could not pass real ENUM's to this routine so we shall
    //    convert them now.
    //
    enum  TARGETID RealTargetID = (enum TARGETID) TargetID;
    enum  TARGETIMAGE RealTargetImage = (enum TARGETIMAGE) TargetImage;

    //
    //    If requested, validate the login information passed to us.
    //

    //
    //    From the supplied product and image name, format the name of the executable that
    //    we are going to start and call the base start function to start it.
    //
    char    ImagePath[512];
    long Status = FormatExecutable (RealTargetID ,RealTargetImage ,pProduct ,ImagePath);
    if (Status != 0)
        throw CyrSpawnException ("Unable to translate CYRANO_HOME"
                                 ,"SpawnProcessImpl::StartProductImage"
                                 ,-1);

    //
    //    Add on the parameters to the command line and invoke it.
    //
    (void)sprintf (ImagePath ,"%s %s" ,ImagePath ,pParameters);
    return Start (ImagePath);
}


//
//    FormatExecutable
//    ================
//
// @mfunc   This private method is a platform specific method for formatting the full path
//          for an executable that is to be started.
// @parm    The ID of the target for which the image is to be formatted (see Target_Rtndef.h)
// @parm    The ID of the image for which the executable name is to be formatted (see Target_Rtndef.h)
// @parn    The product directory in which the executable is to be found
// @parm    Buffer to return the formatted executable name into
// @rdesc   0 if success, -1 if an error
//
//    The following is an example of a WIN32 path returned from this routine
//
//      <CYRANO_HOME>\<PRODUCT>\<SUB-PRODUCT>\TExecuter_syb.exe
//
#ifdef WIN32
int SpawnProcessImpl::FormatExecutable  (enum  TARGETID TargetID ,enum  TARGETIMAGE TargetImage
        ,const char* pProduct ,char* pImagePath)
{
    //
    //    If we have specified product then ensure that it has a trailing slash
    //
    char Product[32];
    Product[0] = '\0';

    if (pProduct != NULL)
    {
        (void)strcpy (Product ,pProduct);
        (void)strcat (Product ,"\\");
    }

    //
    //    Recover the CYRANO_HOME environment variable and ensure it has a trailing slash also
    //
    char    Path[512];
    int Status = GetEnvironmentVariable("CYRANO_HOME" ,Path ,sizeof (Path));

    std::cout << "CYRANO_HOME is " << Path << std::endl;
    if (Status <= 0) return -1;

    int Length = strlen (Path);
    if (Path[Length-1] != '\\') strcat (Path ,"\\");

    //
    //    Recover the sub-directory name (which is the same as the target name) and add on a trailing slash
    //
    char SubProduct[32];
    if (TargetID != TARGET_UNKNOWN)
    {
        Status = Tgt_GetTarget (TargetID ,SubProduct ,sizeof (SubProduct) ,FALSE);
        if (Status != 0) return -1;
        (void)strcat (SubProduct ,"\\");
    }
    else
    {
        SubProduct[0] = '\0';
    }

    std::cout << "SubProduct is " << SubProduct << std::endl;

    //
    //    Recover the image name (which is based on the target image)
    //
    char ImageName[32];
    Tgt_GetTargetExecutable (TargetID ,TargetImage ,ImageName ,sizeof (ImageName));

    //
    //    ...and format the executable name.
    //
    (void)sprintf(pImagePath ,"\"%s%s%s%s\"" ,Path ,Product ,SubProduct ,ImageName);
    return 0;
}


#elif UNIX
int SpawnProcessImpl::FormatExecutable  (enum  TARGETID TargetID ,enum  TARGETIMAGE TargetImage
        ,const char* pProduct ,char* pImagePath)
{
    //
    //    If we have specified product then ensure that it has a trailing slash
    //
    char Product[32];
    Product[0] = '\0';

    if (pProduct != NULL)
    {
        (void)strcpy (Product ,pProduct);
        (void)strcat (Product ,"/");
    }

    //
    //    Recover the CYRANO_HOME environment variable and ensure it has a trailing slash also
    //
    char    Path[512];
    char*   pEnvVar;
    pEnvVar = getenv("CYRANO_HOME");
    if (pEnvVar == NULL) return -1;
    (void)strcpy (Path ,pEnvVar);
    int Length = strlen (Path);
    if (Path[Length-1] != '/') strcat (Path ,"/");

    //
    //    Recover the sub-directory name (which is the same as the target name) and add on a trailing slash
    //
    char SubProduct[32];
    if (TargetID != TARGET_UNKNOWN)
    {
        int Status = Tgt_GetTarget (TargetID ,SubProduct ,sizeof (SubProduct) ,FALSE);
        if (Status != 0) return -1;
        (void)strcat (SubProduct ,"//");
    }
    else
    {
        SubProduct[0] = '\0';
    }

    //
    //    Recover the image name (which is based on the target image)
    //
    char ImageName[32];
    Tgt_GetTargetExecutable (TargetID ,TargetImage ,ImageName ,sizeof (ImageName));

    //
    //    ...and format the executable name.
    //
    (void)sprintf(pImagePath ,"%s%s%s%s" ,Path ,Product ,SubProduct ,ImageName);
    return 0;
}



#elif VMS
int SpawnProcessImpl::FormatExecutable  (enum  TARGETID TargetID ,enum  TARGETIMAGE TargetImage
        ,const char* pProduct ,char* pImagePath)
{
    struct   dsc$descriptor_s tabnam_dsc;   // Table descriptor
    struct   dsc$descriptor_s lognam_dsc;   // Logical name descriptor
    static   char CyranoHome[] = "CYRANO_HOME";
    // CYRANO_HOME logical name
    static   char SystemTable[] = "LNM$SYSTEM_TABLE";

    //
    //    Itemlists for VMS system calls.
    //
    struct   itemlist itemlist [2];         // System call Itemlist
    char    Path[512];             // Path to image
    unsigned short   PathLen;               // Length of path returned
    unsigned long    iret;                  // Status return

    //
    //    Assume no path to start with.
    //
    (void)memset(Path ,'\0' ,sizeof (Path));

    //
    //    Recover the CYRANO_HOME logical name which is defined in the system logical name table.
    //
    lognam_dsc.dsc$a_pointer = CyranoHome;
    lognam_dsc.dsc$w_length  = strlen (CyranoHome);

    tabnam_dsc.dsc$a_pointer = SystemTable;
    tabnam_dsc.dsc$w_length  = strlen (SystemTable);

    itemlist [0].itl_item = LNM$_STRING;
    itemlist [0].itl_buflen = sizeof (Path);
    itemlist [0].itl_bufadr = Path;
    itemlist [0].itl_lenadr = &PathLen;

    itemlist [1].itl_buflen = 0;
    itemlist [1].itl_item   = 0;

    iret = sys$trnlnm (NULL, &tabnam_dsc, &lognam_dsc, NULL, itemlist);
    if (!psl_test_status(iret)) return -1;

    //
    //    Strip off any trailing closing bracket as we may want to add on a sub-directory
    //
    int Length = strlen (Path);
    if (Path[Length-1] == ']') Path[Length-1] = '\0';

    //
    //    If we have specified product then add it to the path
    //
    if (pProduct != NULL)
    {
        (void)strcat (Path ,".");
        (void)strcat (Path ,pProduct);
    }


    //
    //    Recover the sub-directory name (which is the same as the target name) and add it to the path
    //
    char SubProduct[32];
    if (TargetID != TARGET_UNKNOWN)
    {
        int Status = Tgt_GetTarget (TargetID ,SubProduct ,sizeof (SubProduct) ,FALSE);
        if (Status != 0) return -1;
        (void)strcat (Path ,".");
        (void)strcat (Path ,SubProduct);
    }

    //
    //    End of the path - complete with a closing bracket
    //
    (void)strcat (Path ,"]");

    //
    //    Recover the image name (which is based on the target image)
    //
    char ImageName[32];
    Tgt_GetTargetExecutable (TargetID ,TargetImage ,ImageName ,sizeof (ImageName));

    //
    //    ...and add it to the path.
    //
    strcat (Path ,ImageName);
    return 0;
}

#else
#error Platform not supported for recovering executable paths
#endif
