/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// Hooker.cpp: implementation of the CHooker class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Hooker.h"
#include <ImageHlp.h>
#include "Psapi.h"
#include <algorithm>
#include <iterator>

#ifdef _DEBUG
#   include<fstream>
    using std::endl;
    using std::flush;
    extern std::ofstream fout("c:\\temp\\fout.txt"); // My trace file.
#endif


// Link directives.
#pragma comment(lib, "ImageHlp.lib")
#pragma comment(lib, "psapi.lib")
// End Link directives.


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////

// Static data initialization.
HMODULE             CHooker::s_hHookerModule = GetModuleHandleW(L"CaptureBHO.dll");
CHooker::HookList   CHooker::s_HookList;
CHooker::ModuleList CHooker::s_ModuleList;
CRITICAL_SECTION    CHooker::s_cs = {0};
// END Static data initialization.


// Static Initialization.
BOOL CHooker::Init()
{
#ifdef _DEBUG
    fout << "~~Creating the built-in hooks..." << endl;
#endif

    InitializeCriticalSection(&s_cs);

    // Create the initial list of modules.
    ModuleList newModules;
    if(GetNewModules(newModules))
        s_ModuleList = newModules;

    s_Hook_LoadLibraryA   = new CHooker("LoadLibraryA",   "Kernel32.dll", (PROC)CHooker::Hook_LoadLibraryA);
    s_Hook_LoadLibraryW   = new CHooker("LoadLibraryW",   "Kernel32.dll", (PROC)CHooker::Hook_LoadLibraryW);
    s_Hook_LoadLibraryExA = new CHooker("LoadLibraryExA", "Kernel32.dll", (PROC)CHooker::Hook_LoadLibraryExA);
    s_Hook_LoadLibraryExW = new CHooker("LoadLibraryExW", "Kernel32.dll", (PROC)CHooker::Hook_LoadLibraryExW);
    s_Hook_GetProcAddress = new CHooker("GetProcAddress", "Kernel32.dll", (PROC)CHooker::Hook_GetProcAddress);

#ifdef _DEBUG
    fout << "~~Built-in hooks created." << endl;
#endif

    return TRUE;
}

// Static Uninitialization.
BOOL CHooker::Uninit()
{
#ifdef _DEBUG
    fout << "~~Deleting the built-in hooks..." << endl;
#endif

    if(s_Hook_LoadLibraryA)
    {
        delete s_Hook_LoadLibraryA;
        s_Hook_LoadLibraryA = NULL;
    }
    if(s_Hook_LoadLibraryW)
    {
        delete s_Hook_LoadLibraryW;
        s_Hook_LoadLibraryW = NULL;
    }
    if(s_Hook_LoadLibraryExA)
    {
        delete s_Hook_LoadLibraryExA;
        s_Hook_LoadLibraryExA = NULL;
    }
    if(s_Hook_LoadLibraryExW)
    {
        delete s_Hook_LoadLibraryExW;
        s_Hook_LoadLibraryExW = NULL;
    }
    if(s_Hook_GetProcAddress)
    {
        delete s_Hook_GetProcAddress;
        s_Hook_GetProcAddress = NULL;
    }

    DeleteCriticalSection(&s_cs);

#ifdef _DEBUG
    fout << "~~Built-in hooks deleted." << endl;
#endif

    return TRUE;
}


//////////////////////////////////////////////////////////////////////


// Hook constructor.
CHooker::CHooker(PCSTR zFunctionName, PCSTR zExportingModuleName, PROC pfnHookFunction)
    : m_zFunctionName(zFunctionName),
      m_zExportingModuleName(zExportingModuleName),
      m_pfnHookFunction(pfnHookFunction),
      m_pfnOriginalFunction(NULL)
{
    // Get the handle of the exporting module.
    HMODULE hExportingModule = GetModuleHandleA(zExportingModuleName);
    if(hExportingModule == NULL)
    {
#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": ";

        fout << "\t!!! Creating unresolved hook for function "
             << zFunctionName
             << " because host DLL is not loaded."
             << endl;
#endif
    }
    else
    {
        // Get the original funtion's address.
        m_pfnOriginalFunction = GetProcAddress(hExportingModule, m_zFunctionName);

        // Check that the function to hook really exists!
        if(m_pfnOriginalFunction == NULL)
        {
#ifdef _DEBUG
            DWORD dwThread = GetCurrentThreadId();
            fout << dwThread << ": ";

            fout << "\t!!! Failed to create hook for function "
                 << zFunctionName
                 << " because it was not found in host DLL."
                 << endl;
#endif
            return; // This hook must never be installed, because the function doesn't exist.
        }
    }

#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": ";

        fout << "\t### Installing new hook for function "
             << m_zFunctionName
             << "."
             << endl;
#endif

    // Insert the new hook in the list.
    s_HookList.push_front(this);

    if(m_pfnOriginalFunction)
    {
        // Hook the function in all modules.
        HookInAllModules(m_zFunctionName, m_zExportingModuleName, m_pfnOriginalFunction,
            pfnHookFunction, TRUE);
    }
}

// Hook destructor.
CHooker::~CHooker()
{
    // Uninstall the hook for this function in all modules.
    HookInAllModules(m_zFunctionName, m_zExportingModuleName, m_pfnHookFunction, m_pfnOriginalFunction, FALSE);

    // Delete the hook from the list.
    s_HookList.remove(this);
}

// Building a list of the newly loaded modules.
// It is assmued that newModules is empty, and that s_ModuleList is sorted.
BOOL CHooker::GetNewModules(ModuleList& newModules)
{
    BOOL bRet = TRUE;
    HMODULE hm[512] = {0};
    DWORD dwNeeded = 0;
    if(EnumProcessModules(GetCurrentProcess(), hm, sizeof(hm), &dwNeeded))
    {
        DWORD dwNum = dwNeeded / sizeof(HMODULE);

        // Sort the module handles.
        std::sort(hm, hm + dwNum);

        // Compute the set difference.
        std::set_difference(hm, hm + dwNum, s_ModuleList.begin(), s_ModuleList.end(),
            std::back_inserter(newModules));

#ifdef _DEBUG
        // Log the new modules.
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": ";
        fout << newModules.size() << " new module(s):" << endl;

        for(ModuleList::const_iterator it = newModules.begin(); it != newModules.end(); ++it)
        {
            char zFileName[MAX_PATH] = {0};
            if(GetModuleFileNameA(*it, zFileName, sizeof(zFileName)/sizeof(zFileName[0])) == 0)
                lstrcpyA(zFileName, "[NULL]");
            fout << dwThread << ": \t"
                 << "0x"
                 << std::hex << (DWORD)(*it) << std::dec
                 << ": "
                 << zFileName
                 << endl;
        }
#endif
    }
    else
        bRet = FALSE;

    return bRet;
}

// Try to resolve any unresolved hook among a set of newly loaded modules.
// Return TRUE if at least one hook was resolved.
BOOL CHooker::ResolveHooks(const ModuleList& newModules)
{
    BOOL bRet = FALSE;
    for(ModuleList::const_iterator it = newModules.begin(); it != newModules.end(); ++it)
    {
        if(ResolveHooks(*it))
            bRet = TRUE;
    }

    return bRet;
}

// Walk the hook list, and try to resolve any unresolved hook.
// Return TRUE if at least one hook was resolved.
BOOL CHooker::ResolveHooks(HMODULE hModule)
{
    BOOL bRet = FALSE;

    for(HookList::const_iterator it = s_HookList.begin(); it != s_HookList.end(); ++it)
    {
        CHooker *pHook = *it;
        if(pHook->m_pfnOriginalFunction == NULL
            && GetModuleHandleA(pHook->m_zExportingModuleName) == hModule)
        {
            // Try to resolve.
            pHook->m_pfnOriginalFunction = GetProcAddress(hModule, pHook->m_zFunctionName);
            if(pHook->m_pfnOriginalFunction)
            {
                bRet = TRUE;
#ifdef _DEBUG
                DWORD dwThread = GetCurrentThreadId();
                fout << dwThread << ": ";
                fout << "!!! Resolved hook for "
                     << pHook->m_zFunctionName
                     << "( "
                     << pHook->m_zExportingModuleName
                     << " )"
                     << endl;
#endif
                // Apply the resolved hook to the module list.
                HookInAllModules(pHook->m_zFunctionName, pHook->m_zExportingModuleName,
                    pHook->m_pfnOriginalFunction, pHook->m_pfnHookFunction, TRUE);
            }
        }
    }

    return bRet;
}

// Hooks a function in a module. If the hook was installed, then returns TRUE.
// bReplace is used to indicate if we are installing a hook or restoring the
// original function.
BOOL CHooker::HookInModule(HMODULE hImportingModule, PCSTR zFunctionName, PCSTR zExportingModuleName,
   PROC pfnFunctionToHook, PROC pfnHookFunction, BOOL bReplace, BOOL bVerbose)
{
    BOOL bRet = FALSE;

    // Get the import section.
    ULONG ulSize;
    PIMAGE_IMPORT_DESCRIPTOR pImportDesc =
        (PIMAGE_IMPORT_DESCRIPTOR)ImageDirectoryEntryToData(hImportingModule,
            TRUE, IMAGE_DIRECTORY_ENTRY_IMPORT, &ulSize);

    if(pImportDesc != NULL)         // There is an import section.
    {
        // Look for the appropriate import descriptor.
        while(pImportDesc->Name)
        {
            // Build absolute module name address.
            PSTR zCurModuleName = (PSTR)hImportingModule;
            zCurModuleName += pImportDesc->Name;
            if(lstrcmpiA(zCurModuleName, zExportingModuleName) == 0) 
                break;   // Found
            ++pImportDesc;
        }

        if(pImportDesc->Name != 0)  // The module imports functions from zExportingModuleName.
        {
            // Get the corresponding IAT.
            PIMAGE_THUNK_DATA pThunk = (PIMAGE_THUNK_DATA)((BYTE*)hImportingModule + pImportDesc->FirstThunk);

            // Loop on the imported functions.
            while(pThunk->u1.Function)
            {
                // Is this the function to hook?
				// Cast of pThunk->u1.Function to be removed if SDK installed
                if((DWORD)pThunk->u1.Function == (DWORD)pfnFunctionToHook)
                {
                    // OK to hook!
                    WriteProcessMemory(GetCurrentProcess(), (PVOID)&pThunk->u1.Function, (PVOID)&pfnHookFunction, sizeof(pfnHookFunction), NULL);
#ifdef _DEBUG
                    if(bVerbose)
                    {
                        DWORD dwThread = GetCurrentThreadId();
                        fout << dwThread << ": ";

                        char zFileName[MAX_PATH] = {0};
                        if(GetModuleFileNameA(hImportingModule, zFileName, sizeof(zFileName)/sizeof(zFileName[0])) == 0)
                            lstrcpyA(zFileName, "[NULL]");

                        if(bReplace)
                            fout << "\t>>Hooked ";
                        else
                            fout << "\t<<Restored ";
                        fout << zFunctionName
                             << "( "
                             << zExportingModuleName
                             << " ) in "
                             << zFileName
                             << "."
                             << endl;
                    }
#endif
                    bRet = TRUE;    // New hook installed.
                    break;
                }
                ++pThunk;
            }
        }
    }

    return bRet;
}

// Hooks a function in all modules. If at least one hook was
// installed, then returns TRUE.
// bReplace is used to indicate if we are installing a hook or
// restoring the original function.
BOOL CHooker::HookInAllModules(PCSTR zFunctionName, PCSTR zExportingModuleName,
   PROC pfnFunctionToHook, PROC pfnHookFunction, BOOL bReplace)
{
    BOOL bRet = FALSE;

#ifdef _DEBUG
    fout << "\t\t" << s_ModuleList.size() << " modules." << endl;
#endif
    for(ModuleList::const_iterator it = s_ModuleList.begin(); it != s_ModuleList.end(); ++it)
    {
        HMODULE hModule = *it;
        if(hModule != s_hHookerModule)  // Do not hook in the hooker module!
        {
            char zFileName[MAX_PATH] = {0};
            DWORD dwLen = GetModuleFileNameA(hModule, zFileName, sizeof(zFileName)/sizeof(zFileName[0]));
#ifdef _DEBUG
            DWORD dwThread = GetCurrentThreadId();
            fout << dwThread << ": ";

            if(dwLen == 0)
                lstrcpyA(zFileName, "[NULL]");
            fout << "\t\t"
                 << zFileName;
#endif
            if(dwLen)
            {
                // Hook this function in this module.
                if(HookInModule(hModule, zFunctionName, zExportingModuleName,
                    pfnFunctionToHook, pfnHookFunction, bReplace, FALSE))
                {
                    bRet = TRUE;
#ifdef _DEBUG
                    fout << "\t=> Success";
#endif
                }
            }
            else
            {
#ifdef _DEBUG
                fout << "\t=> Unloaded";
#endif
            }
#ifdef _DEBUG
            fout << endl;
#endif
        }
        else
        {
#ifdef _DEBUG
            DWORD dwThread = GetCurrentThreadId();
            fout << dwThread << ": ";

            char zFileName[MAX_PATH] = {0};
            if(GetModuleFileNameA(hModule, zFileName, sizeof(zFileName)/sizeof(zFileName[0])) == 0)
                lstrcpyA(zFileName, "[NULL]");
            fout << "\t\t"
                 << zFileName
                 << "\t=> N/A"
                 << endl;
#endif
        }
    }

    return bRet;
}

// This function installs all the current hooks in the modules of newModules.
// If at least one hook was installed, then returns TRUE.
BOOL CHooker::InstallHooksInNewModules(const ModuleList& newModules)
{
    BOOL bRet = FALSE;

    for(ModuleList::const_iterator it = newModules.begin(); it != newModules.end(); ++it)
    {
        if(InstallHooksInNewModule(*it))
            bRet = TRUE;
    }

    return bRet;
}

// This function installs all the current hooks in one module.
// If at least one hook was installed, then returns TRUE.
BOOL CHooker::InstallHooksInNewModule(HMODULE hNewModule)
{
    BOOL bRet = FALSE;

    for(HookList::const_iterator it = s_HookList.begin(); it != s_HookList.end(); ++it)
    {
        if(HookInModule(hNewModule, (*it)->m_zFunctionName, (*it)->m_zExportingModuleName,
           (*it)->m_pfnOriginalFunction, (*it)->m_pfnHookFunction, TRUE))
            bRet = TRUE;
    }

    return bRet;
}


// -----------------------------------------------------------------------------------
// ---------------------------------- Built-in hooks ---------------------------------
// -----------------------------------------------------------------------------------

CHooker* CHooker::s_Hook_LoadLibraryA    = NULL;
CHooker* CHooker::s_Hook_LoadLibraryW    = NULL;
CHooker* CHooker::s_Hook_LoadLibraryExA  = NULL;
CHooker* CHooker::s_Hook_LoadLibraryExW  = NULL;
CHooker* CHooker::s_Hook_GetProcAddress  = NULL;
// END Hook global variables.


// Called when a LoadLibrary* was detected, to resolve and install hooks.
BOOL CHooker::ProcessNewModules()
{
    BOOL bRet = FALSE;

    // The LoadLibrary* has loaded a new module, plus some
    // modules it implicitely depends on.
    // Get the list of all new modules.
    ModuleList newModules;
    if(GetNewModules(newModules))
    {
        // These new modules may allow to resolve some yet unresolved hooks.
        // Attempt resolution, and apply newly resolved hooks to the existing
        // module list (which does not yet include the new modules).
        ResolveHooks(newModules);

        // Install all hooks in the new modules.
        InstallHooksInNewModules(newModules);

        // Add the new modules to the current module list.
        s_ModuleList.merge(newModules);

        bRet = TRUE;
    }

    return bRet;
}

// Hook functions.
HMODULE WINAPI CHooker::Hook_LoadLibraryA(PCSTR pFileName)
{
    EnterCriticalSection(&s_cs);

    HMODULE hModule = ::LoadLibraryA(pFileName);
    if(hModule)
    {
#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": "
             << "LoadLibraryA( "
             << pFileName
             << " )"
             << endl;
#endif
        ProcessNewModules();
    }

    LeaveCriticalSection(&s_cs);
    return hModule;
}

HMODULE WINAPI CHooker::Hook_LoadLibraryW(PCWSTR pFileName)
{
    EnterCriticalSection(&s_cs);

    HMODULE hModule = ::LoadLibraryW(pFileName);
    if(hModule)
    {
#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        char zFileA[MAX_PATH];
        WideCharToMultiByte(CP_ACP, 0, pFileName, -1, zFileA, sizeof(zFileA), NULL, NULL);
        fout << dwThread << ": "
             << "LoadLibraryW( "
             << zFileA
             << " )"
             << endl;
#endif
        ProcessNewModules();
    }

    LeaveCriticalSection(&s_cs);
    return hModule;
}

HMODULE WINAPI CHooker::Hook_LoadLibraryExA(PCSTR pFileName, HANDLE hFile, DWORD dwFlags)
{
    EnterCriticalSection(&s_cs);

    HMODULE hModule = ::LoadLibraryExA(pFileName, hFile, dwFlags);
    if(hModule && (dwFlags & LOAD_LIBRARY_AS_DATAFILE) == 0)
    {
#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": "
             << "LoadLibraryExA( "
             << pFileName
             << " )"
             << endl;
#endif
        ProcessNewModules();
    }

    LeaveCriticalSection(&s_cs);
    return hModule;
}

HMODULE WINAPI CHooker::Hook_LoadLibraryExW(PCWSTR pFileName, HANDLE hFile, DWORD dwFlags)
{
    EnterCriticalSection(&s_cs);

    HMODULE hModule = ::LoadLibraryExW(pFileName, hFile, dwFlags);
    if(hModule && (dwFlags & LOAD_LIBRARY_AS_DATAFILE) == 0)
    {
#ifdef _DEBUG
        DWORD dwThread = GetCurrentThreadId();
        char zFileA[MAX_PATH];
        WideCharToMultiByte(CP_ACP, 0, pFileName, -1, zFileA, sizeof(zFileA), NULL, NULL);
        fout << dwThread << ": "
             << "LoadLibraryExW( "
             << zFileA
             << " )"
             << endl;
#endif
        ProcessNewModules();
    }

    LeaveCriticalSection(&s_cs);
    return hModule;
}

FARPROC WINAPI CHooker::Hook_GetProcAddress(HMODULE hModule, PCSTR pProcName)
{
#ifdef _DEBUG
        char zFileName[MAX_PATH] = {0};
        if(GetModuleFileNameA(hModule, zFileName, sizeof(zFileName)/sizeof(zFileName[0])) == 0)
            lstrcpyA(zFileName, "[NULL]");
        DWORD dwThread = GetCurrentThreadId();
        fout << dwThread << ": "
             << "GetProcAddress( "
             << zFileName
             << ", ";
        if((((DWORD)pProcName) & 0xffff0000) == 0)
            fout << "ordinal: " << (DWORD)pProcName;  // Ordinal.
        else
            fout << pProcName;         // Function name.
        fout << " )"
             << endl;
#endif

    // Get the function's real address.
    FARPROC pfnFunctionAddress = ::GetProcAddress(hModule, pProcName);

    if(pfnFunctionAddress)
    {
        // If that function needs to be hooked, then hook it.
        for(HookList::const_iterator it = s_HookList.begin(); it != s_HookList.end(); ++it)
        {
            if(pfnFunctionAddress == (*it)->m_pfnOriginalFunction)
            {
#ifdef _DEBUG
                DWORD dwThread = GetCurrentThreadId();
                fout << dwThread << ": ";

                fout << "\t>>Dynamically Hooked "
                     << pProcName
                     << "( "
                     << (*it)->m_zExportingModuleName
                     << " )."
                     << endl;
#endif
                // Install the hook.
                pfnFunctionAddress = (*it)->m_pfnHookFunction;
                break;
            }
        }
    }

    return pfnFunctionAddress;
}
// END Hook functions.
