/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "Address.hpp"
#include <iterator>

const char* Address::AddressTypeToString(Address::AddressType addrType)
{
    const char *pRet = NULL;
    switch(addrType)
    {
        case ADDRESS_TEXT:
            pRet = "TEXT";
            break;
        case ADDRESS_SCRIPT:
            pRet = "SCRIPT";
            break;
        case ADDRESS_COMMENT:
            pRet = "COMMENT";
            break;
        case ADDRESS_ANONYMOUS_ATTRIBUTE:
            pRet = "ANONYMOUS ATTRIBUTE";
            break;
        case ADDRESS_ATTRIBUTE:
            pRet = "ATTRIBUTE";
            break;
        default:
            pRet = "";
            break;
    }
    return pRet;
}

Address::AddressType Address::StringToAddressType(const std::string& strAddrType)
{
    int iAddrType;
    for(iAddrType = ADDRESS_NONE + 1; iAddrType != ADDRESS_LAST; ++iAddrType)
    {
        if(strAddrType == AddressTypeToString((AddressType)iAddrType))
            break;
    }
    return (AddressType)iAddrType;
}

Address::Address(const char* zStringAddress)
{
    Parse(std::string(zStringAddress));
}

void Address::Parse(std::string& strAddr)
{
    addrType_ = ADDRESS_NONE;

    // Extract the tag hierarchy.
    std::string::size_type pos = strAddr.find(':');
    if(pos == std::string::npos)
        return; // Malformed address.
    std::string strTags = strAddr.substr(0, pos);
    strTags += "/ ";
    std::string strRest = strAddr.substr(pos + 1);

    // Build the tag hierarchy.
    std::string::size_type newpos = 0;
    pos = 0;
    AddressNode node;
    while((newpos = strTags.find('/', pos)) != std::string::npos)
    {
        // Extract the tag_label and tag_childnumber.
        std::string strTag = strTags.substr(pos, newpos - pos - 1); // Strip off the right parenthesis.
        pos = strTag.rfind('(');
        if(pos == std::string::npos)
            return; // Malformed address.
        std::string strChildNum = strTag.substr(pos + 1);
        std::istringstream is(strChildNum);
        is >> node.iChild_;
        strTag = strTag.substr(0, pos);
        node.eTag_ = nameToTag[strTag.c_str()];
        PushNode(node);

        pos = newpos + 1;   // Skip to the next tag.
    }

    // Extract the address type.
    pos = strRest.find(':');
    if(pos == std::string::npos)
        return; // Malformed address.
    strAddr = strRest.substr(0, pos);
    AddressType addressType = StringToAddressType(strAddr);

    // Extract the optional attribute_name.
    strAddr = strRest.substr(pos + 1);
    pos = strAddr.find('(');
    if(pos == std::string::npos)
        return; // Malformed address.
    if(pos != 0)
        strAtt_ = strAddr.substr(0, pos);

    // Extract the childnumber.
    int iChild;
    strAddr = strAddr.substr(pos + 1);
    std::istringstream is(strAddr);
    is >> iChild;
    SetChildNumber(addressType, iChild);

    // Nothing should remain after the childnumber and the ')'.
    char cParenth = 0;
    is >> cParenth;
    if(cParenth != ')')
        return; // Malformed address.
    std::string str;
    is >> str;
    if(!str.empty())
        return; // Malformed address.

    // Address is correct: set the address type.
    addrType_ = addressType;
}

std::string Address::ToString() const
{
    BOOL bFirst = TRUE;
    std::ostringstream os;
    for(TagHierarchy::const_iterator it = tagHierarchy_.begin(); it != tagHierarchy_.end(); it++)
    {
        if(bFirst)
            bFirst = FALSE;
        else
            os << '/';
        os << tagToName[it->eTag_] << '(' << it->iChild_ << ')';
    }
    os << ':' << AddressTypeToString(addrType_) << ':';
    if(!strAtt_.empty())
        os << strAtt_;
    os << '(' << GetChildNumber() << ')';
    return os.str();
}

#ifdef _DEBUG
void Address::Dump(std::ostream& os) const
{
    os << ToString().c_str();
}
#endif

AddressQuery::AddressQuery(const std::string& strVariable, const char* zStringAddressQuery)
    : strVariable_(strVariable)
{
    std::string strAq = zStringAddressQuery;
    std::string::size_type n = strAq.size();

    // Is the last character a ']' or a '['?
    if(n > 5 && (strAq[n-1] == ']' || strAq[n-1] == '['))
    {
        // Find the beginning of the address query extensions.
        std::string::size_type iSize = strAq.find_first_of("[]");
        std::string strExtension = strAq.substr(iSize);

        char c = 0;
        std::istringstream is(strExtension);
        is >> c;

        // First boundary's exclusion flag.
        if(c != '[' && c != ']')
            return;         // Bad format.
        begin_.bIncluded = (c == '[');

        // Offset.
        c = is.peek();
        if(c == '>')
        {
            is.ignore();        // Skip the >

            // Offset's exclusion flag.
            c = is.peek();
            offset_.bIncluded = (c == '=');
            if(c == '=')
                is.ignore();    // Skip the =

            // Offset's case sensitivity.
            char ci = is.peek();
            offset_.bCaseSensitive = (ci != 'I');
            if(ci == 'I')
                is.ignore();    // Skip the I

            // Offset value.
            c = is.peek();
            offset_.rep = (c == '\'' ? IS_STRING : IS_NUMBER);
            if(c == '\'')
            {
                // Get the offset string.
                if(!ExtractString(is, offset_.str))
                    return;     // Bad format.
            }
            else
            {
                if(ci == 'I')
                    return;     // Bad format.
                is >> offset_.iNumber;
            }

            // Skip the comma.
            is >> c;
            if(c != ',')
                return;         // Bad format.
        }

        // First boundary.
        // First boundary's case sensitivity.
        char ci = is.peek();
        begin_.bCaseSensitive = (ci != 'I');
        if(ci == 'I')
            is.ignore();    // Skip the I

        // First boundary value.
        c = is.peek();
        begin_.rep = (c == '\'' ? IS_STRING : IS_NUMBER);
        if(c == '\'')
        {
            // Get the first boundary string.
            if(!ExtractString(is, begin_.str))
                return;     // Bad format.
        }
        else
        {
            if(ci == 'I')
                return;     // Bad format.
            is >> begin_.iNumber;
        }

        // Skip the comma.
        is >> c;
        if(c != ',')
            return;         // Bad format.

        // Second boundary.
        // Second boundary's case sensitivity.
        ci = is.peek();
        end_.bCaseSensitive = (ci != 'I');
        if(ci == 'I')
            is.ignore();    // Skip the I

        // Second boundary value.
        c = is.peek();
        end_.rep = (c == '\'' ? IS_STRING : IS_NUMBER);
        if(c == '\'')
        {
            // Get the second boundary string.
            if(!ExtractString(is, end_.str))
                return;     // Bad format.
        }
        else
        {
            if(ci == 'I')
                return;     // Bad format.
            is >> end_.iNumber;
        }

        // Second boundary's exclusion flag.
        is >> c;
        end_.bIncluded = (c == ']');
        if(c != '[' && c != ']')
            return;         // Bad format.

        if(is >> c)
            return;         // Bad format: should be End-of-stream.

        // Chop off the pos and size indication.
        strAq = strAq.substr(0, iSize);
    }
    address_.Parse(strAq);
}

BOOL AddressQuery::ExtractString(std::istringstream& is, std::string& str)
{
    BOOL bRet = FALSE;

    str.erase();
    std::istreambuf_iterator<char> inpos(is);
    std::istreambuf_iterator<char> endpos;   // End-of-stream iterator.

    // First single quote.
    char c = *inpos;
    if(c == '\'')
    {
        ++inpos;
        BOOL bEscape = FALSE;
        while(inpos != endpos)
        {
            char c = *inpos;
            ++inpos;

            if(c == '\\')
                bEscape = TRUE;
            else
            {
                if(c == '\'' && !bEscape)
                {
                    bRet = TRUE;
                    break;
                }
                bEscape = FALSE;
                str += c;
            }
        }
    }

    return bRet;
}
