/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*****************************************************************************/
/*                                       */
/*                **** C O M M O N  ****                         */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_CNV_RTNS                    */
/*          =============                                */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains conversion routines for use by PSL utilities    */
/*                                       */
/*  PSL_CNV_RECFMT  - Converts byte ordering of data, produces or reads  */
/*            comma seperated lists representing some data       */
/*                                       */
/*****************************************************************************/
#include <psl_macros.h>

/*
 *  Standard library functions.
 */
#include <stdio.h>
#include <string.h>

/*
 *  Function prototypes
 */
#include psl_common_file (psl_rtndef.h)     /* PSL routines              */

/*
 *  Structure and symbolic names
 */
#include psl_common_file (psldiag_prmdef.h) /* PSL dignostic messages    */
#include psl_common_file (pslerr_prmdef.h)  /* PSL error messages        */
#include psl_common_file (shrmsg.h)     /* PSL messages              */
#include psl_common_file (psl_cnv_prmdef.h) /* PSL conversion parameters */

/*===========================================================================*/

/*    PSL_CNV_RECFMT - Convert data byte ordering and representation         */
/*    ==============                                                         */

/*    This routine may be used change the byte ordering of data.             */
/*    The format list passed to the routine is an array of long words which  */
/*    define the internal structure of the input or output data.             */
/*    There are three fundamental modes of operation:                        */

/*    Convert the byte ordering of shorts and longs between unlike machines. */

/*    Produce a comma seperated ascii list representing the input data which */
/*    is machine independent.                                                */

/*    Produce a local binary representation of a machine independent ascii   */
/*    list as produced by the above mode of operation.                       */

/*    Full documemntation on the parameters and use can be found in the      */
/*    intternals documentation.                                              */

/*    Calling convention:                                                    */

/*  iret = PSL_CNV_RECFMT (inp_ptr, inptyp, fmt_ptr,                     */
/*                             out_ptr, outtyp, outlen,                      */
/*                             diag_ptr);                                    */

/*  Where:                                                               */

/*  iret        Unsigned long return status:                         */

/*          PSL_NORMAL  - Operation performed successfuly    */
/*          PSL_BADPARAM    - Invalid input or output buffer     */
/*                                    Invalid format list entry          */
/*                                        Output buffer too small for ascii  */
/*                                        list                               */

/*  inp_ptr     Pointer to void location of input buffer             */
/*                                                                       */
/*  inptyp      Unsigned long input data type                        */
/*                                                                       */
/*  fmt_ptr     Pointer to void location of format list              */
/*                                                                       */
/*  out_len     Unsigned long length of output buffer                */
/*                                                                       */
/*  out_ptr     Pointer to void location of output buffer            */
/*                                                                       */
/*  outtyp      Unsigned long output data type                       */
/*                                                                       */
/*  diag_ptr    Pointer to diagnostics stack                     */
/*                                                                       */
/*===========================================================================*/
unsigned long PSL_CNV_RECFMT (  void *inp_ptr,      /* Input buffer      */
                                long inptyp,        /* Input type        */
                                long *fmt_ptr,      /* Format list       */
                                void *out_ptr,      /* Output buffer     */
                                long outtyp,        /* Output type       */
                                unsigned long out_len,  /* Output buf length */
                                struct diagstk *diag_ptr)/* Diag pointer     */
{
    unsigned long   iret = PSL_NORMAL;  /* Return status                     */
    char    *src_ptr;       /* Working input buffer pointer      */
    char    *dst_ptr;       /* Working output buffer pointer     */
    char    swap[2];        /* Temporary buffer used in swaping  */
    char    *ascbuf;        /* Temp buffer when creating ascii   */
    char    *temp_ptr;      /* Temp pointer when reading ascii   */
    short   *short_ptr;     /* Temp pointer when reading ascii   */
    long    *long_ptr;      /* Temp pointer when reading ascii   */
    char    tempbuf[10];        /* Char buffer to hold temp strings  */
    long    space_left;         /* Amount of data in output buffer   */
    long    nochars;        /* Amount of space needed in output  */
    long    item;           /* Current item in format list       */
    long    op;             /* Current format list operand       */
    short   short_val;      /* Working value                     */
    long    long_val;       /* Working value                     */
    long    how_much;       /* Length of copy operation      */
    long    copy_len;       /* Length of copy operation      */


    /*
     *  Check pointers
     */
    if ((inp_ptr == NULL) || (out_ptr == NULL) ||  (fmt_ptr == NULL))
    {
        (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Invalid data or format buffers");
        iret = PSL_BADPARAM;
        goto end;
    }
    /*
     *  Check if parameters are meaningless
     */
    if ((inptyp == PSLRECFMT_COMMA) && (outtyp == PSLRECFMT_COMMA))
    {
        (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                   "Invalid conversion, two comma seperated lists");
        iret = PSL_BADPARAM;
        goto end;
    }
    /*
     *  Assume we need to do something, initialize data
     */
    src_ptr = inp_ptr;
    dst_ptr = out_ptr;
    ascbuf = out_ptr;
    space_left = out_len -1;        /* Space for terminator */
    item = 0;
    /*
     *  Do we want a comma seperated list ?
     */
    if (outtyp == PSLRECFMT_COMMA)
    {
        op = fmt_ptr[item];
        while (op != 0)
        {

            /*
             *  Is the data item to output a long (32bit value)?
             *  Get the number of characters that we need to represent this item by
             *  'printing' it to a temporary buffer first and recording the length.
             *  Then see if there is sufficient space in output buffer
             */
            if (op == -1)
            {
                long_ptr = (void *) src_ptr;
                long_val = *long_ptr;
                nochars = sprintf(tempbuf, "%ld", long_val);

                if (space_left >= nochars)
                {                           /* Yes !     */
                    (void) memcpy (ascbuf, tempbuf, nochars);
                    src_ptr += 4;
                    space_left -= nochars;
                    ascbuf += nochars;
                }
                else
                {                       /* Boooo, no */
                    (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Output buffer full");
                    iret = PSL_BADPARAM;
                    goto end;
                }
            }

            /*
             *  Is the data item to output a short (16bit value)?
             */
            else if (op == -2)
            {
                short_ptr = (void *) src_ptr;
                short_val = *short_ptr;
                nochars = sprintf(tempbuf, "%d", short_val);

                if (space_left >= nochars)
                {                       /* Yes !     */
                    (void) memcpy (ascbuf, tempbuf, nochars);
                    src_ptr += 2;
                    space_left -= nochars;
                    ascbuf += nochars;
                }
                else
                {
                    (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Output buffer full");
                    iret = PSL_BADPARAM;
                    goto end;
                }
            }
            /*
             *  Is the item a string ? Write out the string and space fill any output
             *  if the 'strlen' of the member is less than the member size.
             */
            else if (op > 0)
            {

                /*
                 *  See how much space we need for this string
                 */
                if (space_left >= op)
                {
                    /*
                     *  Is there a NUL in the member
                     */
                    temp_ptr = memchr (src_ptr, '\0', op);
                    if (temp_ptr != NULL)
                    {
                        how_much = temp_ptr - src_ptr;
                    }
                    else
                    {
                        how_much = op;
                    }
                    (void) memcpy (ascbuf, src_ptr, how_much);
                    /*
                     *    If there was a NUL in the member output spaces to make up the size
                     */
                    if (how_much < op)
                    {
                        (void) memset (ascbuf + how_much ,' ', op - how_much);
                    }
                    src_ptr += op;
                    space_left -= op;
                    ascbuf += op;
                }
                else
                    /*
                     *  Not enough space in output buffer
                     */
                {
                    (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Output buffer full");
                    iret = PSL_BADPARAM;
                    goto end;
                }
            }

            else
            {
                /*
                 *  If we cannot determine what the input type is, whinge
                 */
                (void) PSL_ADD_DIAGNOSTIC(diag_ptr,"Bad entry in format list");
                iret = PSL_BADPARAM;
                goto end;
            }

            /*
             *  If the next item in the format list is a not 0 i.e. not the end, write 
             *  out a comma
             */
            item += 1;
            op = fmt_ptr[item];
            if (op != 0)
            {
                if (space_left > 0)
                {
                    *ascbuf = ',';
                    space_left -= 1;
                    ascbuf += 1;
                }
                else
                {
                    (void) PSL_ADD_DIAGNOSTIC(diag_ptr,
                                              "Bad entry in format list");
                    iret = PSL_BADPARAM;
                    goto end;
                }
            }
            /*
             *  NUL terminate the list if there are no more items to output
             */
            else
            {
                *ascbuf = 0;
            }
        }
    }



    /*
     *  Are we to convert a comma seperated list to binary ?
     */
    else if (inptyp == PSLRECFMT_COMMA)
    {
        op = fmt_ptr[item];
        while (op != 0)
        {
            /*
             *      Read in the string and convert it to a long (32bit value)
             */
            if (op == -1)
            {
                nochars = sscanf(src_ptr, "%ld", dst_ptr);
                if (nochars != 1)
                {
                    (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                               "Invalid entry in comma seperated list");
                    iret = PSL_BADPARAM;
                    goto end;
                }
                else
                {
                    dst_ptr += 4;
                }
            }
            /*
             *  Read in the string and convert it to a short
             */
            else if (op == -2)
            {
                nochars = sscanf(src_ptr, "%d", dst_ptr);
                if (nochars != 1)
                {
                    (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Invalid entry in comma seperated list");
                    iret = PSL_BADPARAM;
                    goto end;
                }
                else
                {
                    dst_ptr += 2;
                }
            }
            /*
             *  Is the value a string ? Take the input up to the next comma and put it
             *  in the output. If the resulting section of text is greater than the
             *  available space we will truncate it. If the section is not long enough
             *  then space fill the record member.
             */
            else if (op > 0)
            {

                /*
                 *    Search for the end of this field. this is normally delimited by a comma
                 *    however we do allow quoted strings to contain commas and therefore we
                 *    look for an opening quote first.
                 */
                if (*src_ptr == '"')    /* If we find a double quote    */
                {
                    temp_ptr = strchr ((src_ptr+1) ,'"');
                    /* ...look for closing quote    */
                    if (temp_ptr != NULL) temp_ptr++;
                    /* ...include it if found   */
                }

                else if (*src_ptr == '\'')
                {           /* ...else if we find a single quote */
                    temp_ptr = strchr ((src_ptr+1) ,'\'');
                    /* ...use it as the closer  */
                    if (temp_ptr != NULL) temp_ptr++;
                    /* ...include it if found   */
                }

                else            /* ...else not quoted string    */
                {

                    /*
                     *  Find the next comma in the input buffer
                     */
                    temp_ptr =  strchr (src_ptr,',');
                }

                /*
                 *  If we didn't find one is this the last data item in the buffer?
                 *  see how many characters are remaining.
                 */
                if  (temp_ptr == NULL)
                {
                    how_much = strlen (src_ptr);
                }
                else
                {
                    how_much = temp_ptr - src_ptr;
                }
                /*
                 *  If there are more characters in the string than space in the record
                 *  member then reduce the amount that we will copy. Otherwise copy all we
                 *  can.
                 */
                if (how_much > op)
                {
                    copy_len = op;
                }
                else
                {
                    copy_len = how_much;
                }

                (void) memcpy (dst_ptr, src_ptr, copy_len);

                /*
                 *  If we copied less than the size of the record member we will space fill
                 *  any remaining characters.
                 */
                if (copy_len < op)
                {
                    (void) memset (dst_ptr + copy_len, ' ',op - copy_len);
                }
                dst_ptr += op;
                src_ptr += how_much;
            }
            /*
             *  If we cannot determine what the input type is, whinge
             */
            else
            {
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad entry in format list");
                iret = PSL_BADPARAM;
                goto end;
            }
            /*
             *  Advance to next data item after comma, if we cannot find a commma and
             *  there is another data item to read, flag an error
             */
            item += 1;
            op = fmt_ptr[item];
            temp_ptr = strchr(src_ptr,',');
            if ((temp_ptr == NULL) && (op != 0))
            {
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Input buffer exausted");
                iret = PSL_BADPARAM;
                goto end;
            }
            /*
             *  If the next item in the format list is a not NUL i.e. not the end, read
             *  past the comma
             */
            else
            {
                src_ptr = temp_ptr + 1;
            }
        }
    }
    /*
     *  Do we want to convert the input byte ordering instead?
     */
    else if ((inptyp != outtyp)  || (outtyp == PSLRECFMT_ALIEN))
    {
        op = fmt_ptr[item];
        while (op != 0)
        {
            /*
             *  Convert 32bit value
             */
            if (op == -1)
            {
                swap[0] = src_ptr[0];       /* Save copy of src[0..1] in */
                swap[1] = src_ptr[1];       /* case scr = dst            */
                dst_ptr[0] = src_ptr[3];
                dst_ptr[1] = src_ptr[2];
                dst_ptr[2] = swap[1];
                dst_ptr[3] = swap[0];
                dst_ptr += 4;
                src_ptr +=4;
            }
            else if (op == -2)
                /*
                 *  Convert 16bit value
                 */
            {
                swap[0] = src_ptr[0];       /* Save copy of src[0]       */
                dst_ptr[0] = src_ptr [1];
                dst_ptr[1] = swap [0];
                dst_ptr += 2;
                src_ptr +=2;
            }
            else if (op > 0)
            {
                /*
                 *  'Convert' string i.e. copy it
                 */
                (void) memmove(dst_ptr, src_ptr, op);
                dst_ptr += op;
                src_ptr += op;
            }
            else
            {
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad entry in format list");
                iret = PSL_BADPARAM;
                goto end;
            }
            item += 1;
            op = fmt_ptr[item];
        }
    }
    else
    {
        /*
         *  The input and output have the same byte ordering, copy the details
         */
        op = fmt_ptr[item];
        while (op != 0)
        {
            /*
             *  'Convert' 32bit value
             */
            if (op == -1)
            {
                (void) memmove (dst_ptr, src_ptr, 4);
                dst_ptr += 4;
                src_ptr += 4;
            }
            else if (op == -2)
            {
                (void) memmove (dst_ptr, src_ptr, 2);
                dst_ptr += 2;
                src_ptr += 2;
            }
            else if (op > 0)
            {
                (void) memmove (dst_ptr, src_ptr, op);
                dst_ptr += op;
                src_ptr += op;
            }
            else
            {
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad entry in format list");
                iret = PSL_BADPARAM;
                goto end;
            }
            item += 1;
            op = fmt_ptr[item];
        }
    }
    /*
     *  END:
     *  ====
     *  Jump point for irrecoverable errors and general exit point
     */
end:
    return iret;
}
