/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_DIAG_RTNS               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains a number of routines for managing a diagnostic  */
/*  stack for providing a meaningful error trace in nested utilities.    */
/*                                       */
/*===========================================================================*/


#include <psl_macros.h>         /* PSL Standard Macros */

#include <vtest_macros.h>       /* VTEST Subsystem Macros */

/*   Standard ANSI header files */

#include <stdio.h>          /* Standard I/O */
#include <stdlib.h>         /* Standard Library Routines */
#include <string.h>         /* String functions */
#include <stdarg.h>         /* Variable Argument List handling */
#include <errno.h>          /* Standard error numbers */

#ifdef WIN32
#include <windows.h>            /* Windows routines */
#endif /* WIN32 */

/*   Define Function Prototypes for all external routines */
/*   used by the module (and for ourselves) */

#include <psl_rtndef.h>         /* Define ourselves */
#include <trc_rtndef.h>         /* Trace Routines */

/*   Define Function Prototypes for all local routines */

static unsigned long ADD_DIAGNOSTIC (struct diagstk *diag_ptr,
                                     char *msg_ptr,
                                     unsigned long flags);

static void CHECK_TRAP_ERRORS (struct diagstk  *diag_ptr,
                               unsigned long   errcls,
                               unsigned long   errnum,
                               unsigned long   response,
                               jmp_buf         *jmp_ptr,
                               unsigned long   (*handler)());

static void CREATE_DIAG_BUFFER (struct diagstk *diag_ptr,
                                unsigned long context,
                                unsigned long width,
                                unsigned long wrap_flag,
                                char **bufptr_ptr,
                                unsigned long *count_ptr);

static void ERROR_EXIT (struct diagstk *diag_ptr, char *errmsg);

static void FLUSH_ERDIAG (struct diagstk *diag_ptr,
                          unsigned long context);

static void OUTPUT_TO_STDERR (struct diagstk *diag_ptr,
                              unsigned long context);

static void PUTDIAG_L (struct diagstk *usrdiag_ptr,
                       unsigned long numseg,
                       va_list *va_ptr,
                       unsigned long flags);

static struct diagstk *VALIDATE_DIAGPTR (struct diagstk *diag_ptr);
static   void   PSL_LOG_CIOERR_L (struct diagstk *diag_ptr,
                                  unsigned long msgnum,
                                  va_list *va_ptr);

#ifdef WIN32
static void PSL_LOG_LASTERROR_L (struct diagstk *diag_ptr,
                                 unsigned long msgnum,
                                 va_list *va_ptr);
#endif

/*   Define Data Structures */

#include <diagstkdef.h>         /* Define stack format */
#include <errtrapdef.h>         /* Error trapping definitions */
#include <shrmsg.h>         /* Message Statuses */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols */
#include <pslerr_prmdef.h>      /* Parameter Definitions */
#include <pslmsg_prmdef.h>      /* PSL_GET_MESSAGE flags */
#include <psl_trc_prmdef.h>     /* Standard Tracing Types */

/*   Define symbolic names for standard messages and internal flags */

#define MSG_ALLOCDIAGFAIL "Memory Allocation for Diagnostic Failed"
#define MSG_ALLOCSAVEFAIL "Memory Allocation for Save Buffer Failed"
#define PUTDIAG_SUPP 0x1        /* Supplementary Message */

/*   Define a global, static, diagnostics stack to use if all else  */
/*   fails.  We would normally avoid this technique, but there is   */
/*   nothing a user program can do if an attempt to allocate the    */
/*   initial diagnostics stack fails, so we must ensure there is    */
/*   always one available.                      */

/*   We also return a pointer to this to the user when he first     */
/*   calls PSL_ALLOCATE_DIAGNOSTICS, so we also need a flag to say  */
/*   whether we have done this or not.  This ensures that most      */
/*   (typically all) diagnostics will be on this stack so we can    */
/*   use it if the user loses track of where his diagnostics stack  */
/*   is.                                */

static struct diagstk fixed_diagstk;    /* Emergency Stack*/
static Boolean alloc_fixed = PFALSE;    /* Flag to say if we've allocated it */

/*   We also define a global flag to indicate whether we are in the */
/*   process of exiting due to a fatal error.  This allows      */
/*   TRM_FINISH to detect the fact so that it can refrain from      */
/*   refreshing the screen (and hence wiping out the diagnostics we */
/*   have just output).                         */

static Boolean errexit_flag = PFALSE;   /* Flag to say we're exiting */

/*   We also define a global flag to indicate whether we have ever  */
/*   reported an error or not so that PSL_EXIT can exit with the    */
/*   appropriate status.                        */

static Boolean error_reported = PFALSE; /* Flag to say error reported */

/************************************************************************/
/*                                  */
/*  PSL_ADD_DIAGNOSTIC - Add diagnostic to diagnostic stack     */
/*                                  */
/*  This routine adds a diagnostic to a diagnostic stack;       */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  ival = PSL_ADD_DIAGNOSTIC (diag_ptr, fmt_ptr, ...);     */
/*                                  */
/*  Where:                              */
/*                                  */
/*  ival     = Unsigned Long Context prior to added diagnostic. */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  fmt_ptr  = Pointer to Format String for use by printf.      */
/*  ...  = Arguments required by fmtstr format string.      */
/*                                  */
/************************************************************************/

unsigned long PSL_ADD_DIAGNOSTIC (struct diagstk *usrdiag_ptr,
                                  char *fmt_ptr, ...)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    static   char       msgbuf[1024];   /* Local Message Buffer */
    va_list    args;       /* Argument List Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Just set up the variable argument bits and format the message  */
    /*   via VSPRINTF                           */

    va_start (args, fmt_ptr);   /* Set up the argument pointer */
    (void) vsprintf (msgbuf, fmt_ptr, args); /* Format the diagnostic */
    va_end (args);          /* Tidy up the argument pointer */

    /*   Now add the message to the stack via our local routine and     */
    /*   return the subscript to our caller                 */

    return ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add message to stack */
}

/************************************************************************/
/*                                  */
/*  PSL_ADD_STATDIAG - Add static diagnostic to diagnostic stack    */
/*                                  */
/*  This routine adds a static diagnostic to a diagnostic stack */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  ival = PSL_ADD_STATDIAG (diag_ptr, msg_ptr);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  ival     = Unsigned Long Context prior to added diagnostic. */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  msg_ptr  = Pointer to Static Message String to add to stack.    */
/*                                  */
/************************************************************************/

unsigned long PSL_ADD_STATDIAG (struct diagstk *usrdiag_ptr, char *msg_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Just add the message to the stack via our local routine and    */
    /*   return the subscript to our caller                 */

    return ADD_DIAGNOSTIC (diag_ptr, msg_ptr, DIAGFL_STAT);
    /* Add message to stack */
}

/************************************************************************/
/*                                  */
/*  PSL_ALLOCATE_DIAGNOSTICS - Allocate a diagnostics stack     */
/*                                  */
/*  This routine allocates a diagnostics stack.         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  diag_ptr = PSL_ALLOCATE_DIAGNOSTICS ();             */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*                                  */
/************************************************************************/

struct diagstk *PSL_ALLOCATE_DIAGNOSTICS (void)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Stack Pointer */

    /*   Return a pointer to the fixed stack if it is available,        */
    /*   otherwise try to allocate a new one.               */

    if  (!alloc_fixed)      /* If fixed stack available */
    {
        diag_ptr = &fixed_diagstk;  /* Give the user that one */
        alloc_fixed = PTRUE;    /* And say we've allocated it */
    }
    else                /* Else (fixed is in use) */
    {
        diag_ptr = (struct diagstk *) calloc (1, sizeof(struct diagstk));
        /* Allocate a new one */
    }

    /*   Set up the structure type so that we can check for corruption  */
    /*   at a later date (if we want to).                   */

    if  (diag_ptr != NULL) (*diag_ptr).diag_strtypl = DIAG_STRNAM;
    /* Set up Structure Name */

    /*   Create a Cardfile entry */

    PSL_CARDFILE_ADD (diag_ptr, "DIA --> ");

    /*   Lastly, return the pointer to the caller */

    return diag_ptr;        /* Return pointer to diagnostics */
}

/************************************************************************/
/*                                  */
/*  PSL_CHK_ERREXIT - Check to see if we are doing an error exit    */
/*                                  */
/*  This routine checks to see if we are exiting via ERROR_EXIT */
/*  so that TRM_FINISH knows whether to reset the terminal or not.  */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  iret = PSL_CHK_ERREXIT ();                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret     = Unsigned Long PSL Routine Status:            */
/*       = PSL_NORMAL if exiting normally           */
/*       = PSL_BADPARAM otherwise               */
/*                                  */
/************************************************************************/

unsigned long PSL_CHK_ERREXIT (void)
{
    if  (errexit_flag)      /* If flag is set */
    {
        return PSL_BADPARAM;    /* Return Error Status */
    }
    else                /* Else (flag not set) */
    {
        return PSL_NORMAL;      /* Return Success */
    }

}

/************************************************************************/
/*                                  */
/*  PSL_DEALLOCATE_DIAGNOSTICS - Deallocate a diagnostics stack */
/*                                  */
/*  This routine deallocates a diagnostics stack.           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  Call PSL_DEALLOCATE_DIAGNOSTICS (diag_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*                                  */
/************************************************************************/

void    PSL_DEALLOCATE_DIAGNOSTICS (struct diagstk *usrdiag_ptr)
{


    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   First delete any stack entries */

    PSL_REMOVE_DIAGNOSTICS (diag_ptr, 0);
    /* Delete all entries */

    /*   Remove the Cardfile entry */

    PSL_CARDFILE_REMOVE ("DIA --> ");

    /*   If this was a pointer to the fixed stack then just mark that */
    /*   it is now available - otherwise free up the memory it uses. */

    if  (diag_ptr == &fixed_diagstk) /* If the fixed stack */
    {
        alloc_fixed = PFALSE;   /* just show it is now free */
    }
    else                /* Else (secondary stack) */
    {
        free (diag_ptr);        /* Free diagnostic stack */
    }
}

/************************************************************************/
/*                                  */
/*  PSL_DIAG_ERRNO - Add a system service error with errno      */
/*                                  */
/*  This routine adds a system service error for the specified  */
/*  function and a secondary error message of PSL_STDCERR with the  */
/*  text returned from strerror(errno).             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = PSL_DIAG_ERRNO (diag_ptr, rtn_name_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret         = PSL_SYSERVERR alays returned         */
/*  diag_ptr     = Pointer to the diagnostic stack          */
/*  rtn_name_ptr = Pointer to the system service name string    */
/*                                  */
/************************************************************************/
unsigned long PSL_DIAG_ERRNO(struct diagstk *usrdiag_ptr,
                             char *rtn_name_ptr)
{
    /*
     *  Add the message to the diagnostic stack, using the strerror function
     *  to pick up the text of the last error.
     *
     *  Note that currently PSL_PUTDIAG only supports FAO substitution for
     *  the first message segment, so we perform our own processing.
     */

    /*
     *  Local Data
     */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       msgbuf[255];    /* Local Message Buffer */
    char       fmtbuf[255];    /* Local Format Buffer */
    unsigned long       msglen;     /* Length of message text */
    char       *msg_ptr;       /* Pointer to next character in msg */
    char       *sysmsg_ptr;    /* Pointer to system message */
    unsigned long       context;        /* Saved context */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*
     *  Set the error context at the current position
     */
    PSL_SET_ERRORCTX(diag_ptr);

    /*
     *  Save the current last error context
     */
    context = PSL_GET_ERRORCTX (diag_ptr);

    /*
     *  Get the message prefix.
     *  Replace facility name with utility name. 
     */
#ifndef VMS_MSG_FORMAT
    msgbuf[0] = '-';            /* '-' for secondary status */
#endif
    (void) PSL_GET_MESSAGE (PSL_STDCERR,
#ifdef VMS_MSG_FORMAT
                            msgbuf,
                            sizeof(msgbuf),
#else
                            &msgbuf[1],
                            sizeof(msgbuf) - 1,
#endif
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */

    /*
     *  Get the message text
     */
    (void) PSL_GET_MESSAGE (PSL_STDCERR, fmtbuf, sizeof(fmtbuf), PSLMSG_TEXT);

    msglen = strlen(msgbuf);
    msg_ptr = &msgbuf[msglen];      /* Point to end of prefix */
    *msg_ptr = ',';         /* And add ", " */
    msg_ptr += 1;
    *msg_ptr = ' ';         /* Add ' ' */
    msg_ptr += 1;

    /*
     *  Now substitute system message and replace '%' with '-' as this is a
     *  secondary status.
     */
    sysmsg_ptr = strerror(errno);
    if (sysmsg_ptr != NULL)
    {
        PSL_FAO (msg_ptr,
                 sizeof(msgbuf) - (msglen+2),
                 fmtbuf,
                 sysmsg_ptr);
    }
    else
    {
        PSL_FAO (msg_ptr,
                 sizeof(msgbuf) - (msglen+2),
                 fmtbuf,
                 "no message available from errno");
    }
#ifdef VMS_MSG_FORMAT
    msgbuf[0] = '-';            /* Reset character at start */
#endif

    /*
     *  Add the secondary message first as the stack must be in reverse order.
     */
    (void) PSL_ADD_DIAGNOSTIC (diag_ptr, msgbuf);
    /* Add to diagnostics stack */
    /*
     *  Add the primary system service error message
     */
    PSL_PUTDIAG(diag_ptr, 1, PSL_SYSERVERR, rtn_name_ptr);


    /*
     *  Set the error context to point to before the messages we have added.
     */
    PSL_FIX_ERRORCTX (diag_ptr, context);

    return PSL_SYSERVERR;
}

/************************************************************************/
/*                                  */
/*  PSL_DIAG_INVERT - Invert Diagnostics for Latest Error       */
/*                                  */
/*  This routine reverses the order of all diagnostics logged   */
/*  as part of the latest error (i.e. since DIAG_LSTERRL).  This    */
/*  is necessary because the Diagnostics Stack is a LIFO stack  */
/*  while multiple error lines tend to be FIFO.         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      PSL_DIAG_INVERT (diag_ptr);                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr    = Pointer to User Diagnostics Stack.        */
/*                                  */
/************************************************************************/

void PSL_DIAG_INVERT (struct diagstk *usrdiag_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       *item_ptr;      /* Pointer to message */
    unsigned long       flags;      /* Flags for message */
    unsigned long       fstsub;     /* First item being inverted */
    unsigned long       lstsub;     /* Last item being inverted */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Check to see if there have been no diagnostics since the last  */
    /*   error was defined and, if so, just return.             */

    if  ((*diag_ptr).diag_lsterrl >=
            ((*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl))
        /* If nothing new since */
    {
        return;         /* Just return */
    }

    /*   We know we've got something to do, but we'll just check first  */
    /*   that we haven't got too much (i.e. the latest error has        */
    /*   over-filled the stack on it's own.                 */

    if  ((*diag_ptr).diag_lsterrl < (*diag_ptr).diag_biasl)
        /* If error started before top */
    {
        (*diag_ptr).diag_lsterrl = (*diag_ptr).diag_biasl;
        /* Reset to first message */
    }

    /*   We now need to move through the stack, from the first one      */
    /*   affected to the last, inverting the order of the pointers.     */

    for  (fstsub = (*diag_ptr).diag_lsterrl - (*diag_ptr).diag_biasl,
            /* Starting at first for error */
            lstsub = (*diag_ptr).diag_cursubl - 1;
            /* and ending at last in stack */
            fstsub < lstsub;      /* until we reach the middle */
            fstsub++, lstsub--)   /* moving both ways at once */
    {
        item_ptr = (*diag_ptr).diag_stackpc100[fstsub];
        /* Save first pointer */
        (*diag_ptr).diag_stackpc100[fstsub] =
            (*diag_ptr).diag_stackpc100[lstsub];
        /* Replace with last pointer */
        (*diag_ptr).diag_stackpc100[lstsub] = item_ptr;
        /* And reset last pointer */
        flags = (*diag_ptr).diag_flagsl100[fstsub];
        /* Save first flags */
        (*diag_ptr).diag_flagsl100[fstsub] =
            (*diag_ptr).diag_flagsl100[lstsub];
        /* Replace with last flags */
        (*diag_ptr).diag_flagsl100[lstsub] = flags;
        /* And reset last flags */
    }

    /*   All done - return to the caller */

    return;             /* Return to caller */
}

#ifdef WIN32
/************************************************************************/
/*                                  */
/*  PSL_DIAG_LASTERROR - Add a system service error with LastError  */
/*                                  */
/*  This routine adds a system service error for the specified  */
/*  function and a secondary error message of PSL_STDCERR with the  */
/*  text returned from GetLastError.                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      iret = PSL_DIAG_LASTERROR (diag_ptr, rtn_name_ptr);     */
/*                                  */
/*  Where:                              */
/*                                  */
/*      iret         = always PSL_SYSERVERR             */
/*      diag_ptr     = Pointer to diagnostic stack          */
/*      rtn_name_ptr = Pointer to routine name string       */
/*                                  */
/************************************************************************/

unsigned long PSL_DIAG_LASTERROR (struct diagstk *usrdiag_ptr,
                                  const char *rtn_name_ptr)
{
    /*
     *  Local data.
     */
    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       msgbuf[255];    /* Local Message Buffer */
    char       fmtbuf[255];    /* Local Format Buffer */
    char       errbuf[255];    /* Error Buffer */
    unsigned long       msglen;     /* Length of message text */
    char       *msg_ptr;       /* Pointer to next character in msg */
    unsigned long       context;        /* Saved context */

    DWORD  dwError = GetLastError();    /* Get last error */
    LPVOID lpMessageBuffer = NULL;      /* Pointer to error text buffer */

    char       socketerr[32];  /* Socket error message */
    char       *socketerr_text[] = {
        "WSAEWOULDBLOCK",   //         (WSABASEERR+35)
        "WSAEINPROGRESS",   //         (WSABASEERR+36)
        "WSAEALREADY",  //         (WSABASEERR+37)
        "WSAENOTSOCK",  //         (WSABASEERR+38)
        "WSAEDESTADDRREQ",  //         (WSABASEERR+39)
        "WSAEMSGSIZE",  //         (WSABASEERR+40)
        "WSAEPROTOTYPE",    //         (WSABASEERR+41)
        "WSAENOPROTOOPT",   //         (WSABASEERR+42)
        "WSAEPROTONOSUPPORT", //       (WSABASEERR+43)
        "WSAESOCKTNOSUPPORT", //       (WSABASEERR+44)
        "WSAEOPNOTSUPP",    //         (WSABASEERR+45)
        "WSAEPFNOSUPPORT",  //         (WSABASEERR+46)
        "WSAEAFNOSUPPORT",  //         (WSABASEERR+47)
        "WSAEADDRINUSE",    //         (WSABASEERR+48)
        "WSAEADDRNOTAVAIL", //         (WSABASEERR+49)
        "WSAENETDOWN",  //         (WSABASEERR+50)
        "WSAENETUNREACH",   //         (WSABASEERR+51)
        "WSAENETRESET", //         (WSABASEERR+52)
        "WSAECONNABORTED",  //         (WSABASEERR+53)
        "WSAECONNRESET",    //         (WSABASEERR+54)
        "WSAENOBUFS",   //         (WSABASEERR+55)
        "WSAEISCONN",   //         (WSABASEERR+56)
        "WSAENOTCONN",  //         (WSABASEERR+57)
        "WSAESHUTDOWN", //         (WSABASEERR+58)
        "WSAETOOMANYREFS",  //         (WSABASEERR+59)
        "WSAETIMEDOUT", //         (WSABASEERR+60)
        "WSAECONNREFUSED",  //         (WSABASEERR+61)
        "WSAELOOP",     //         (WSABASEERR+62)
        "WSAENAMETOOLONG",  //         (WSABASEERR+63)
        "WSAEHOSTDOWN", //         (WSABASEERR+64)
        "WSAEHOSTUNREACH",  //         (WSABASEERR+65)
        "WSAENOTEMPTY", //         (WSABASEERR+66)
        "WSAEPROCLIM",  //         (WSABASEERR+67)
        "WSAEUSERS",    //         (WSABASEERR+68)
        "WSAEDQUOT",    //         (WSABASEERR+69)
        "WSAESTALE",    //         (WSABASEERR+70)
        "WSAEREMOTE"    //         (WSABASEERR+71)
    };

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*
     *  Set the error context at the current position.
     */
    PSL_SET_ERRORCTX (diag_ptr);

    /*
     *  Save the current last error context.
     */
    context = PSL_GET_ERRORCTX (diag_ptr);

    /*
     *  Get the message prefix.
     *  Replace facility name with utility name. 
     */
    msgbuf[0] = '-';            /* '-' for secondary status */

    (void) PSL_GET_MESSAGE (PSL_STDCERR,
                            &msgbuf[1],
                            sizeof(msgbuf) - 1,
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */

    msglen = strlen(msgbuf);
    msg_ptr = &msgbuf[msglen];      /* Point to end of prefix */
    *msg_ptr = ',';         /* And add ", " */
    msg_ptr += 1;
    *msg_ptr = ' ';         /* Add ' ' */
    msg_ptr += 1;

    /*
     *  Get the message text.
     */
    (void) PSL_GET_MESSAGE (PSL_STDCERR, fmtbuf, sizeof(fmtbuf), PSLMSG_TEXT);

    /*
     *  Now get system message text.
     */
    if (dwError >= WSAEWOULDBLOCK)
    {
        if (dwError <= WSAEREMOTE)      /* If it's a socket error */
        {
            (void)strcpy (socketerr, socketerr_text[dwError - WSAEWOULDBLOCK]);
        }
        else if (dwError == WSASYSNOTREADY)         // (WSABASEERR+91)
        {
            (void)strcpy (socketerr, "WSASYSNOTREADY");
        }
        else if (dwError == WSAVERNOTSUPPORTED)         // (WSABASEERR+92)
        {
            (void)strcpy (socketerr, "WSAVERNOTSUPPORTED");
        }
        else if (dwError == WSANOTINITIALISED)          // (WSABASEERR+93)
        {
            (void)strcpy (socketerr, "WSANOTINITIALISED");
        }
        else if (dwError == WSAEDISCON)             // (WSABASEERR+101)
        {
            (void)strcpy (socketerr, "WSAEDISCON");
        }
        else
        {
            (void)sprintf (socketerr, "Unknown Error %d", dwError);
        }
        PSL_FAO (msg_ptr,
                 sizeof(msgbuf) - (msglen+2),
                 fmtbuf,
                 socketerr);
        msgbuf[0] = '-';            /* Reset character at start */
    }
    else
    {
        if (FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
                           FORMAT_MESSAGE_FROM_SYSTEM,
                           NULL,
                           dwError,
                           LANG_NEUTRAL,
                           (LPSTR) &lpMessageBuffer,
                           0,
                           NULL))       /* Get system message text */
        {
            PSL_FAO (msg_ptr,
                     sizeof(msgbuf) - (msglen+2),
                     fmtbuf,
                     lpMessageBuffer);
        }
        else                    /* It failed */
        {
            (void) sprintf (        /* Set up text by hand */
                errbuf,
                "no message available from FormatMessage (%d)",
                GetLastError());
            PSL_FAO (msg_ptr,
                     sizeof(msgbuf) - (msglen+2),
                     fmtbuf,
                     errbuf);
        }

        /*
         *  If a buffer was allocated to store the system message text
         *  we need to free it up.
         */
        if (lpMessageBuffer != NULL)        /* If buffer was allocated */
        {
            LocalFree ((HLOCAL) lpMessageBuffer);
            /* Free it */
        }
    }

    /*
     *  Add the secondary message first as the stack must be in reverse order.
     */
    (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add to diagnostics stack */

    /*
     *  Add the primary system service error message.
     */
    PSL_PUTDIAG (diag_ptr, 1, PSL_SYSERVERR, rtn_name_ptr);

    /*
     *  Set the error context to point to before the messages we have added.
     */
    PSL_FIX_ERRORCTX (diag_ptr, context);

    return PSL_SYSERVERR;       /* Return System Service Error */
}

#endif /* WIN32 */

/************************************************************************/
/*                                  */
/*  PSL_EXIT - Exit with Worst Error Status             */
/*                                  */
/*  This routine terminates the program, returning the      */
/*  'Worst Error' status to the caller.             */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  PSL_EXIT()                          */
/*                                  */
/************************************************************************/

void    PSL_EXIT (void)
{
    /*   Simply check the flag shared with RESET_STATUS to see whether  */
    /*   or not we have reported an error and exit with the appropriate */
    /*   status.                                */

    if  (error_reported)        /* If we have reported an error */
    {
        exit (EXIT_FAILURE);    /* exit with failure */
    }
    else
    {
        exit (EXIT_SUCCESS);    /* otherwise, exit with success */
    }
}

/************************************************************************/
/*                                  */
/*  PSL_FIX_ERRORCTX - Fix context of start of last error       */
/*                                  */
/*  This routine is called to fix the context of the start of   */
/*  the last error to the specified value.              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_FIX_ERRORCTX (diag_ptr, context);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context to set             */
/*                                  */
/************************************************************************/

void PSL_FIX_ERRORCTX (struct diagstk *usrdiag_ptr,
                       unsigned long context)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Fix the last error context */

    (*diag_ptr).diag_lsterrl = context;
    return;
}

/************************************************************************/
/*                                  */
/*  PSL_FLUSH_ERRORS - Flush items from the Error Registry.     */
/*                                  */
/*  This routine removes all items in the error registry that were  */
/*  registered after a particular error trap, including the     */
/*  trap itself.  It is primarily used internally by        */
/*  PSL_SIGNAL_ERROR, but is available for general use.     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_FLUSH_ERRORS (diag_ptr, context);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context of last trap to be deleted.    */
/*         0 => clear error registry.               */
/*                                  */
/************************************************************************/

void PSL_FLUSH_ERRORS (struct diagstk *usrdiag_ptr, unsigned long number)
{
    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    struct   errtrap    *err_ptr;       /* Pointer to Error Trap */
    struct   errtrap    *prev_ptr;      /* and to previous one */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Free errtrap structures until the context is found. */

    err_ptr = (*diag_ptr).diag_errtrpps;
    /* Get latest error trap */
    while ((err_ptr != NULL)    /* While more levels to check */
            &&  (number <= (*err_ptr).err_numberl))
        /* and we've not reached specified level*/
    {
        prev_ptr = (*err_ptr).err_prev_ptrps;
        /* Get pointer to previous level */
        free(err_ptr);      /* Release this level */
        err_ptr = prev_ptr;     /* And check previous one */
    }

    /*   We've found the latest acceptable one - store it as the current one */

    (*diag_ptr).diag_errtrpps = err_ptr;
    /* Set error trap pointer */

    /*   All done - return to the caller */

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_GET_DIAGNOSTIC - Get diagnostic from diagnostic stack   */
/*                                  */
/*  This routine retrieves a diagnostic from a diagnostic stack.    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  msg_ptr = PSL_GET_DIAGNOSTIC (diag_ptr, context_ptr);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*  msg_ptr     = Pointer to diagnostic text.           */
/*  diag_ptr    = Pointer to Diagnostics Stack.         */
/*  context_ptr = Pointer to context for operation.  The context    */
/*            should initially be set to zero and will be   */
/*            reset by the routine.             */
/*                                  */
/************************************************************************/

char *PSL_GET_DIAGNOSTIC (struct diagstk *usrdiag_ptr, unsigned long *ctx_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       *item_ptr;      /* Local Item Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   First see if the context has been initialized yet and, if not, */
    /*   initialize it.                         */

    if  ((*ctx_ptr) == 0)       /* If uninitialized */
    {
        *ctx_ptr = (*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl + 1;
        /* Point past most recent element */
    }

    /*   Now reset context pointer and return either a pointer to the   */
    /*   associated message or NULL if no more messages.  Note that if  */
    /*   the stack entry is NULL it means we ran out of memory so we    */
    /*   return a pointer to a fixed message in that case.          */

    *ctx_ptr -= 1;          /* Move back up the stack */
    if  ((*ctx_ptr) > (*diag_ptr).diag_biasl)
        /* If value still valid */
    {
        item_ptr = (*diag_ptr).diag_stackpc100
                   [*ctx_ptr - (*diag_ptr).diag_biasl-1];
        /* Get pointer to item for entry */
        if  (item_ptr != NULL)  /* If real message pointer */
        {
            return item_ptr;    /* Return it */
        }
        else            /* Else (no real message) */
        {
            return MSG_ALLOCDIAGFAIL;
            /* Return Standard (static) message */
        }
    }
    else                /* Else (end of stack) */
    {
        return NULL;        /* Return NULL */
    }
}

/************************************************************************/
/*                                  */
/*  PSL_GET_ERRORCTX - Get context of start of last error       */
/*                                  */
/*  This routine returns the diagnostics context associated     */
/*  with the start of the most recent error.            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  context = PSL_GET_ERRORCTX (diag_ptr);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*  context  = Unsigned Long Context of latest error        */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*                                  */
/************************************************************************/

unsigned long PSL_GET_ERRORCTX (struct diagstk *usrdiag_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Return the last error context. */

    return (*diag_ptr).diag_lsterrl;/* Return Last Error Context */
}

/************************************************************************/
/*                                  */
/*  PSL_GET_LSTERROR - Get details of most recent error.        */
/*                                  */
/*  This routine is used to fetch details of the most recent error  */
/*  signalled to the Error Registry.                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_GET_LSTERROR (diag_ptr, errcls_ptr, errnum_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr    = Pointer to Diagnostics Stack.         */
/*  errcls_ptr  = Pointer to variable to receive the error class    */
/*            (one of PSLERRCLS_*).             */
/*            Set to zero if no errors have been signalled. */
/*  errnum_ptr  = Pointer to variable to receive the error number.  */
/*            Set to zero if no errors have been signalled. */
/*                                  */
/************************************************************************/

void PSL_GET_LSTERROR (struct diagstk *usrdiag_ptr,
                       unsigned long *errcls_ptr,
                       unsigned long *errnum_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Extract and return the latest signalled error class and number */

    *errnum_ptr   = (*diag_ptr).diag_signuml;
    /* Return latest error number */
    *errcls_ptr   = (*diag_ptr).diag_sigclsl;
    /* and class */
    return;
}

/************************************************************************/
/*                                  */
/*  PSL_OUTPUT_DIAG - Output diagnostics from stack         */
/*                                  */
/*  This routine outputs some or all of the diagnostics in the  */
/*  stack until it is returned to a specified previous position (or */
/*  as close as we can get.                     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_OUTPUT_DIAG (diag_ptr, context, flags);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context of first diagnostic NOT to be  */
/*         output.                      */
/*  flags    = Unsigned Long Flags (DIOUFL_xxx as defined in    */
/*         PSLDIAG_PRMDEF).                 */
/*         DIOUFL_ALL    => Output entire Diagnostic Stack  */
/*         DIOUFL_LSTERR => Output Last Error           */
/*         If either of these flags is specified then the   */
/*         context parameter is ignored.            */
/*                                  */
/************************************************************************/

void PSL_OUTPUT_DIAG (struct diagstk *usrdiag_ptr,
                      unsigned long context,
                      unsigned long flags)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       *buf_ptr;       /* Pointer to diagnostics buffer */
    unsigned long       num_lines;      /* Number of lines in buffer */
    char       *wrk_buf_ptr;   /* Pointer to diagnostics buffer */

    /*  Pointer - TCB Action Routine */

    unsigned long       (*tcb_actrtn_ptr) (struct tcb *tcb_ptr,
                                           char *tcb_buf_ptr,
                                           long numlin,
                                           long outwid);
    /*  Pointer - LIST Action Routine */

    unsigned long       (*list_actrtn_ptr) (struct list_ctx *context_ptr,
                                            const char *list_buf_ptr,
                                            unsigned long numlin,
                                            unsigned long outwid);
    /*  Pointer - USER Action Routine */

    unsigned long       (*user_actrtn_ptr) (void *output_ptr,
                                            const char *user_buf_ptr,
                                            unsigned long numlin,
                                            unsigned long outwid);

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*    Because we can now manually fix the position of the last error    */
    /*    pointer it is possible that there is in fact no last error even   */
    /*    though the pointer suggests that there is. If this is the case,   */
    /*    reset the last error pointer to 0.                */

    if  ((*diag_ptr).diag_lsterrl >=
            ((*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl))
        /* If we've reached last error */
    {
        (*diag_ptr).diag_lsterrl = 0; /* Reset it */
    }

    /*   If the user requested the entire diagnostic stack, or the last */
    /*   error, set up the context accordingly.             */

    if  ((flags & DIOUFL_ALL) != 0)
    {
        context = 0;
    }
    else if ((flags & DIOUFL_LSTERR) != 0)
    {
        context = (*diag_ptr).diag_lsterrl;
    }

    /*   Next see if the user has specified one or more redirections of */
    /*   the output.  If not then just iterate through the stack        */
    /*   outputting entries until we either exhaust the stack or reach  */
    /*   the user-specified context.  If he has then we buffer up the   */
    /*   items and let the specified action routine(s) output the       */
    /*   information as they see fit.  To make life a bit easier, we    */
    /*   handle each type in a subroutine.                  */

    if (((*diag_ptr).diag_outtypl == 0)
            /* If no options set */
            || (((*diag_ptr).diag_outtypl & DIAG_OUTTYP_LIST) != 0))
        /* or output to Listing requested */
    {
        OUTPUT_TO_STDERR (      /* Just output to stderr */
            diag_ptr,       /* passing diagnostics pointer */
            context);       /* and context */
    }

    if ((*diag_ptr).diag_outtypl != 0)
        /* if some options set */
    {

        /*   First check for output to the listing routines */

        if  (((*diag_ptr).diag_outtypl & DIAG_OUTTYP_LIST) != 0)
            /* If Output to Listing requested */
        {
            CREATE_DIAG_BUFFER (    /* Create a buffer of messages */
                diag_ptr,       /* passing diagnostics pointer */
                context,        /* and context */
                (*diag_ptr).diag_list_outwidl,
                /* and width */
                (*diag_ptr).diag_outtypl & DIAG_OUTTYP_WRAP_LIST,
                /* status of wrap flag */
                &buf_ptr,       /* address of buffer pointer */
                &num_lines);    /* address of line count */
            if  (buf_ptr == NULL)   /* If allocation failed */
            {
                wrk_buf_ptr = MSG_ALLOCSAVEFAIL;
                /* Point to Error Message */
                num_lines = 1;  /* Set to only one line */
            }
            else            /* Else (allocation OK) */
            {
                wrk_buf_ptr = buf_ptr;
                /* Just copy buffer pointer */
            }           /* End of (if allocation failed) */
            list_actrtn_ptr = (*diag_ptr).diag_list_actrtnpf;
            /* Get pointer to Action Routine */
            (void) (*list_actrtn_ptr) (
                (*diag_ptr).diag_list_ctx_ptrps,
                wrk_buf_ptr,
                num_lines,
                (*diag_ptr).diag_list_outwidl);
            /* Call it */
            if  (buf_ptr != NULL)   /* If allocation succeeded */
            {
                free (buf_ptr); /* Free the buffer */
            }
        }               /* End of (if output to listing) */

        /*   Next check for output to the terminal  */

        if  (((*diag_ptr).diag_outtypl & DIAG_OUTTYP_TCB) != 0)
            /* If Output to Listing requested */
        {
            CREATE_DIAG_BUFFER (    /* Create a buffer of messages */
                diag_ptr,       /* passing diagnostics pointer */
                context,        /* and context */
                (*diag_ptr).diag_tcb_outwidl,
                /* and width */
                (*diag_ptr).diag_outtypl & DIAG_OUTTYP_WRAP_TCB,
                /* status of wrap flag */
                &buf_ptr,       /* address of buffer pointer */
                &num_lines);    /* address of line count */
            if  (buf_ptr == NULL)   /* If allocation failed */
            {
                wrk_buf_ptr = MSG_ALLOCSAVEFAIL;
                /* Point to Error Message */
                num_lines = 1;  /* Set to only one line */
            }
            else            /* Else (allocation OK) */
            {
                wrk_buf_ptr = buf_ptr;
                /* Just copy buffer pointer */
            }           /* End of (if allocation failed) */
            tcb_actrtn_ptr = (*diag_ptr).diag_tcb_actrtnpf;
            /* Get pointer to Action Routine */
            (void) (*tcb_actrtn_ptr) (
                (*diag_ptr).diag_tcb_ptrps,
                wrk_buf_ptr,
                num_lines,
                (*diag_ptr).diag_tcb_outwidl);
            /* Call it */
            if  (buf_ptr != NULL)   /* If allocation succeeded */
            {
                free (buf_ptr); /* Free the buffer */
            }
        }               /* End of (if output to terminal) */

        /*   Lastly check for output to a user routine */

        if  (((*diag_ptr).diag_outtypl & DIAG_OUTTYP_USER) != 0)
            /* If Output to User routine requested */
        {
            CREATE_DIAG_BUFFER (    /* Create a buffer of messages */
                diag_ptr,       /* passing diagnostics pointer */
                context,        /* and context */
                (*diag_ptr).diag_user_outwidl,
                /* and width */
                (*diag_ptr).diag_outtypl & DIAG_OUTTYP_WRAP_USER,
                /* status of wrap flag */
                &buf_ptr,       /* address of buffer pointer */
                &num_lines);    /* address of line count */
            if  (buf_ptr == NULL)   /* If allocation failed */
            {
                wrk_buf_ptr = MSG_ALLOCSAVEFAIL;
                /* Point to Error Message */
                num_lines = 1;  /* Set to only one line */
            }
            else            /* Else (allocation OK) */
            {
                wrk_buf_ptr = buf_ptr;
                /* Just copy buffer pointer */
            }           /* End of (if allocation failed) */
            user_actrtn_ptr = (*diag_ptr).diag_user_actrtnpf;
            /* Get pointer to Action Routine */
            (void) (*user_actrtn_ptr) (
                (*diag_ptr).diag_user_ptrp,
                wrk_buf_ptr,
                num_lines,
                (*diag_ptr).diag_user_outwidl);
            /* Call it */
            if  (buf_ptr != NULL)   /* If allocation succeeded */
            {
                free (buf_ptr); /* Free the buffer */
            }
        }               /* End of (if output to listing) */
    }               /* End of (if output options set) */

    /*   All done - return to caller */

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_PREDIAG - Add prefix diagnostics to the last error      */
/*                                  */
/*  This routine adds prefix diagnostics to the last error.     */
/*  It is needed so that a 'context' message can be prefixed    */
/*  to a more generic one.                      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_PREDIAG (diag_ptr, numseg, msg1, msg1prm1, msg1prm2, ...,   */
/*                     msg2, msg3, ...);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     msgi     = Message number for ith message.           */
/*     msg1prmj = jth Parameter for 1st message.            */
/*                                  */
/*  The number and type of the parameters for each string are   */
/*  determined by the insertion directives in the string itself.    */
/*  These directives are as defined for PSL_FAO.            */
/*                                  */
/*    This routine forms a shell around PSL_PREDIAG_L which does all of */
/*    the work.                             */
/*                                  */
/************************************************************************/

void PSL_PREDIAG (struct diagstk *usrdiag_ptr, unsigned long numseg, ...)
{
    /*
     *  Local Data   
     */
    va_list    args;            /* Argument List Pointer */

    /*  Set up the Variable Argument List pointer and call PSL_PREDIAG_L */

    va_start (args, numseg);        /* Initialize Pointer */

    PSL_PREDIAG_L (usrdiag_ptr, numseg, &args);
    /* Call PSL_PREDIAG_L */

    /*  All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_PREDIAG_L - Add prefix diagnostics to the last error    */
/*                                  */
/*  This routine adds prefix diagnostics to the last error.     */
/*  It is needed so that a 'context' message can be prefixed    */
/*  to a more generic one.                      */
/*                                  */
/*  It is similar to PSL_PREDIAG but uses a different calling   */
/*  format for instances when an argument list is required to be    */
/*  passed.                             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_PREDIAG_L (diag_ptr, numseg, va_ptr)            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     va_ptr  = Pointer to va_list structure for extra parameters  */
/*                                  */
/*  The routine works as follows:                   */
/*                                  */
/*  1. Save context of last error                   */
/*  2. Call PUTDIAG_L to add prefix diagnostics         */
/*  3. Fix the last error context to the saved value        */
/*                                  */
/*  Assume that the last error on the stack consists of 3 lines:    */
/*                                  */
/*      L3                          */
/*      L2                          */
/*      L1                          */
/*                                  */
/*  (reading the stack from the bottom).                */
/*                                  */
/*  Assume that step 2 adds another 3 line diagnostic. We then have:*/
/*                                  */
/*      L3                          */
/*      L2                          */
/*      L1                          */
/*      L6                          */
/*      L5                          */
/*      L4                          */
/*                                  */
/************************************************************************/

void PSL_PREDIAG_L (struct diagstk *usrdiag_ptr,
                    unsigned long numseg,
                    va_list *va_ptr)
{
    /*
     *    Local Data   
     */
    unsigned long       context;        /* Saved context */

    /*  Save the current last error context */

    context = PSL_GET_ERRORCTX (usrdiag_ptr);

    /*  Pass the argument list straight through to PUTDIAG_L */

    PUTDIAG_L (usrdiag_ptr, numseg, va_ptr, 0);
    /* Call PUTDIAG_L */

    /*  Fix the context of the last error to the saved value */

    PSL_FIX_ERRORCTX (usrdiag_ptr, context);

    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_PUTDIAG - Put a standard message onto the diagnostics stack */
/*                                  */
/*  This routine adds the text associated with a standard message   */
/*  message number to the diagnostics stack, substituting user  */
/*  parameters as it goes.                      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_PUTDIAG (diag_ptr, numseg, msg1, msg1prm1, msg1prm2, ...,   */
/*                     msg2, msg3, ...);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     msgi     = Message number for ith message.           */
/*     msg1prmj = jth Parameter for 1st message.            */
/*                                  */
/*  The number and type of the parameters for each string are   */
/*  determined by the insertion directives in the string itself.    */
/*  These directives are as defined for PSL_FAO.            */
/*                                  */
/************************************************************************/

void PSL_PUTDIAG (struct diagstk *usrdiag_ptr, unsigned long numseg, ...)
{

    /*
     *    Local Data   
     */
    va_list    args;            /* Argument List Pointer */

    /*
     *    Set up the Variable Argument List pointer and call PUTDIAG_L
     */
    va_start (args, numseg);        /* Initialize Pointer */

    PUTDIAG_L (usrdiag_ptr, numseg, &args, 0);
    /* Call PUTDIAG_L */

    /*   All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_PUTDIAG_L - Put a standard message onto the diagnostics     */
/*          stack.                      */
/*                                  */
/*  This routine is simply a user shell around PUTDIAG_L (for   */
/*  historic reasons.                       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_PUTDIAG_L (diag_ptr, numseg, va_ptr)            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     va_ptr  = Pointer to va_list structure for extra parameters  */
/*                                  */
/************************************************************************/

void PSL_PUTDIAG_L (struct diagstk *usrdiag_ptr, unsigned long numseg
                    ,va_list *va_ptr)

{
    /*   Just call PUTDIAG_L */

    PUTDIAG_L (usrdiag_ptr, numseg, va_ptr, 0);
    /* Call PUTDIAG_L */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_REMOVE_DIAGNOSTICS - Remove diagnostics from stack      */
/*                                  */
/*  This routine removes diagnostics from the stack until it is */
/*  returned to a specified previous position (or as close as we    */
/*  can get.                            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      PSL_REMOVE_DIAGNOSTICS (diag_ptr, context);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context to restore stack to.       */
/*         0 => remove all diagnostics.             */
/*                                  */
/************************************************************************/

void PSL_REMOVE_DIAGNOSTICS (struct diagstk *usrdiag_ptr, unsigned long context)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       **entry_ptr;    /* Local Entry Pointer */
    unsigned long       *flags_ptr;     /* Pointer to Diagnostic Entry Flags */


    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   We just iterate through the stack removing entries until we    */
    /*   either exhaust the stack or reach the user-specified context.  */

    while (((*diag_ptr).diag_cursubl > 0)
            /* While there is more on the stack */
            &&   (context < ((*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl)))
        /* And we haven't reached our target */
    {
        entry_ptr = &(*diag_ptr).diag_stackpc100
                    [(*diag_ptr).diag_cursubl - 1];
        /* Get pointer to current entry */

        flags_ptr = &(*diag_ptr).diag_flagsl100
                    [(*diag_ptr).diag_cursubl - 1];
        /* And to flags for entry */
        if  ((*flags_ptr & DIAGFL_STAT) == 0)
            /* If message is dynamic */
        {
            free (*entry_ptr);  /* Free the message string */
        }
        *entry_ptr = NULL;      /* Reset item pointer */
        *flags_ptr = 0;     /* and flags */

        (*diag_ptr).diag_cursubl -= 1;
        /* And decrement current */
    }

    /*   It is possible that the user context will be "above" the point */
    /*   at which the current stack starts.  If so then we just reset   */
    /*   the bias so that the top entry corresponds to the last one the */
    /*   user wanted to delete.                     */

    if   (context < (*diag_ptr).diag_biasl)
        /* If context above the top */
    {
        (*diag_ptr).diag_biasl = context;
        /* Reset bias value */
    }

    /*   If we have progressed beyond the start of the last error then  */
    /*   reset that as well. We allow the last error pointer to point   */
    /*   to the last item on the stack                  */

    if  ((*diag_ptr).diag_lsterrl >
            ((*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl))
        /* If we've reached last error */
    {
        (*diag_ptr).diag_lsterrl = 0; /* Reset it */
    }

    /*   Lastly flush the error trap mechanism as well. */

    FLUSH_ERDIAG (diag_ptr, context);

}

/************************************************************************/
/*                                  */
/*  PSL_REPLACE_DIAGNOSTIC - Replace current diagnostic on      */
/*      diagnostic stack                    */
/*                                  */
/*  This routine replaces the current diagnostic on the     */
/*  diagnostic stack with a different one.              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      PSL_REPLACE_DIAGNOSTIC (diag_ptr, fmt_ptr, ...);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  fmt_ptr  = Pointer to Format String for use by printf.      */
/*  ...  = Arguments required by fmtstr format string.      */
/*                                  */
/************************************************************************/

void PSL_REPLACE_DIAGNOSTIC (struct diagstk *usrdiag_ptr, char *fmt_ptr, ...)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    static   char       msgbuf[1024];   /* Local Message Buffer */
    long       msgsiz;     /* Formatted Message Size */
    char       **entry_ptr;    /* Local Entry Pointer */
    va_list    args;       /* Argument List Pointer */
    unsigned long       *flags_ptr;     /* Pointer to Diagnostic Entry Flags */


    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   First free the old entry on the stack if it was dynamic */

    entry_ptr = &(*diag_ptr).diag_stackpc100[(*diag_ptr).diag_cursubl - 1];
    /* Get pointer to entry */
    flags_ptr = &(*diag_ptr).diag_flagsl100[(*diag_ptr).diag_cursubl - 1];
    /* And pointer to flags for entry */
    if  ((*flags_ptr & DIAGFL_STAT) == 0)
        /* If message is dynamic */
    {
        free (*entry_ptr);      /* Free the message string */
    }

    /*   Now set up the variable argument bits and format the message   */
    /*   via VSPRINTF                           */

    va_start (args, fmt_ptr);   /* Set up the argument pointer */
    (void) vsprintf (msgbuf, fmt_ptr, args); /* Format the diagnostic */
    msgsiz = strlen (msgbuf);   /* And get the size */
    va_end (args);          /* Tidy up the argument pointer */

    /*   Now try to allocate a buffer to hold the message and copy the  */
    /*   message into it.  If the allocation fails, we leave a null     */
    /*   pointer in place to record the fact.               */

    *entry_ptr = (char *) malloc (msgsiz+1);
    /* Allocate room for the message */
    /* and store pointer in entry */
    if  (*entry_ptr != NULL)    /* If allocation succeeded */
    {
        (void) memcpy (*entry_ptr, msgbuf, msgsiz+1);
        /* Copy the message over */
        /* (including trailing null) */
        *flags_ptr = 0;     /* And flag it is dynamic */
    }

    /*   Lastly just return to the caller */

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_REPLACE_STATDIAG - Replace current diagnostic on        */
/*      diagnostic stack with a static message          */
/*                                  */
/*  This routine replaces the current diagnostic on the     */
/*  diagnostic stack with a different (static) one.         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*      PSL_REPLACE_STATDIAG (diag_ptr, msg_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  msg_ptr  = Pointer to Static Message string to add to stack.    */
/*                                  */
/************************************************************************/

void PSL_REPLACE_STATDIAG (struct diagstk *usrdiag_ptr, char *msg_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    char       **entry_ptr;    /* Local Entry Pointer */
    unsigned long       *flags_ptr;     /* Pointer to Diagnostic Entry Flags */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   First free the old entry on the stack if it was dynamic */

    entry_ptr = &(*diag_ptr).diag_stackpc100[(*diag_ptr).diag_cursubl - 1];
    /* Get pointer to entry */
    flags_ptr = &(*diag_ptr).diag_flagsl100[(*diag_ptr).diag_cursubl - 1];
    /* And pointer to flags for entry */
    if  ((*flags_ptr & DIAGFL_STAT) == 0)
        /* If message is dynamic */
    {
        free (*entry_ptr);      /* Free the message string */
    }

    /*   Now copy the message pointer into the entry and flag that it   */
    /*   is staic.                              */

    *entry_ptr = msg_ptr;       /* Store pointer to new message */
    *flags_ptr = DIAGFL_STAT;   /* And flag it is static */

    /*   Lastly just return to the caller */

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_RESIGNAL_ERR - Resignal the latest error            */
/*                                  */
/*  This routine is called by any code that has trapped an error    */
/*  and has decided to resignal it for someone else to look after.  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  errnum = PSL_RESIGNAL_ERR (diag_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*  errnum   = The error number of the latest error.        */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*                                  */
/************************************************************************/

unsigned long PSL_RESIGNAL_ERR (struct diagstk *usrdiag_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Just try to signal the latest error and, if that returns to    */
    /*   us, return the status to the caller.               */

    PSL_SIGNAL_ERROR (diag_ptr,     /* Signal an error */
                      (*diag_ptr).diag_sigclsl,
                      /* Specifying latest class */
                      (*diag_ptr).diag_signuml);
    /* and latest error */
    return (*diag_ptr).diag_signuml;    /* Return error number to caller */
}

/************************************************************************/
/*                                  */
/*  PSL_SET_ERRORCTX - Set context of start of last error       */
/*                                  */
/*  This routine is called at the start of a set of diagnostics */
/*  that relate to an error and set the current context as the  */
/*  context of the last error.                  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SET_ERRORCTX (diag_ptr);                    */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr    = Pointer to Diagnostics Stack.         */
/*                                  */
/************************************************************************/

void PSL_SET_ERRORCTX (struct diagstk *usrdiag_ptr)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Reset the last error context */

    (*diag_ptr).diag_lsterrl =
        (*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl;
    /* Set latest error context */
    return;
}

/************************************************************************/
/*                                  */
/*  PSL_SETUP_OUDUSR - Set up routine for PSL_OUTPUT_DIAG       */
/*                                  */
/*  This routine may be called to set up the User output type to    */
/*  be used by PSL_OUTPUT_DIAG                  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SETUP_OUDUSR (diag_ptr, output_ptr, outwid, actrtn_ptr) */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr    = Pointer to Diagnostics Stack.         */
/*  output_ptr  = Pointer to Output Type data structure.        */
/*            Passing a NULL pointer switches off the       */
/*            specified output type.                */
/*  outwid      = Width of Output Type.             */
/*  actrtn_ptr  = Pointer to User Action Routine.           */
/*                                  */
/************************************************************************/

void    PSL_SETUP_OUDUSR (struct diagstk *diag_ptr,
                          void *output_ptr,
                          unsigned long outwid,
                          unsigned long (*actrtn_ptr)
                          (void *act_output_ptr,
                           const char *errbuf_ptr,
                           unsigned long numlin,
                           unsigned long act_outwid))

{
    if ( output_ptr == NULL )       /* If Clearing Information */
    {
        (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_USER;
        /* Clear Output Type */
        (*diag_ptr).diag_outtypl &= ~DIAG_OUTTYP_WRAP_USER;
        /* Clear Wrap Flag */
        (*diag_ptr).diag_user_ptrp = NULL;
        /* Clear Pointer to User */
        (*diag_ptr).diag_user_actrtnpf = NULL;
        /* Clear Pointer to Action Routine */
        (*diag_ptr).diag_user_outwidl = 0;
        /* Clear Output Width */
    }
    else                /* Setting Information */
    {
        (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_USER;
        /* Store Output Type */
        (*diag_ptr).diag_outtypl |= DIAG_OUTTYP_WRAP_USER;
        /* Store Wrap Flag */
        (*diag_ptr).diag_user_ptrp = output_ptr;
        /* Store Pointer to User */
        (*diag_ptr).diag_user_actrtnpf = actrtn_ptr;
        /* Store Pointer to Action Routine */
        (*diag_ptr).diag_user_outwidl = outwid;
        /* Store Output Width */
    }
    return;
}

/************************************************************************/
/*                                  */
/*  PSL_SIGDIAG - Put a standard message onto the diagnostics stack */
/*          and signal it                   */
/*                                  */
/*  This routine adds the text associated with a standard message   */
/*  message number to the diagnostics stack, substituting user  */
/*  parameters as it goes. It then calls PSL_SIGNAL_ERROR to signal */
/*  the error.                          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SIGDIAG (diag_ptr, numseg, msg1, msg1prm1, msg1prm2, ...,   */
/*                     msg2, msg3, ...);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     msgi     = Message number for ith message.           */
/*     msg1prmj = jth Parameter for 1st message.            */
/*                                  */
/*  The number and type of the parameters for each string are   */
/*  determined by the insertion directives in the string itself.    */
/*  These directives are as defined for PSL_FAO.            */
/*                                  */
/************************************************************************/

void PSL_SIGDIAG (struct diagstk *usrdiag_ptr, unsigned long numseg, ...)
{
    /*
     *    Local Data   
     */
    va_list    args;       /* Argument List Pointer */
    unsigned long       msgnum;     /* Current Message Number */

    /*  Set up the Variable Argument List pointer and call PSL_PUTDIAG_L */

    va_start (args, numseg);        /* Initialize Pointer */

    PUTDIAG_L (usrdiag_ptr, numseg, &args, 0);
    /* Call PUTDIAG_L */
    va_end (args);          /* Tidy up argument list */

    /*  Set up the Variable Argument List pointer and call PSL_SIGNAL_ERROR */

    va_start (args, numseg);        /* Initialize Pointer */

    msgnum = va_arg (args, unsigned long);
    /* Get associated message number */

    PSL_SIGNAL_ERROR (usrdiag_ptr, PSLERRCLS_USER1, msgnum);
    /* Call PSL_SIGNAL_ERROR */

    /*  All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_SIGNAL_ERROR - Signal an error to the Error Registry.   */
/*                                  */
/*  This routine is called by any code that detects an error and    */
/*  wishes it to be handled by the Error Trapping Routines.  Prior  */
/*  to calling this routine the code should have added details of   */
/*  the error to the diagnostics stack.             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SIGNAL_ERROR (diag_ptr, errcls, errnum);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  errcls   = The error class (one of PSLERRCLS_*).        */
/*  errnum   = The error number.                    */
/*                                  */
/************************************************************************/

void PSL_SIGNAL_ERROR (struct diagstk *usrdiag_ptr,
                       unsigned long errcls,
                       unsigned long errnum)
{
    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    struct   errtrap    *err_ptr;       /* Pointer to Error Trap */
    unsigned long       trpcls;     /* Error class in Trap */
    unsigned long       trpnum;     /* Error number in Trap */
    unsigned long       response;       /* Response required */
    unsigned long       (*usrrtn_pf)(); /* Pointer to user routine */
    jmp_buf    *dest_ptr;      /* longjmp destination */
    static   long       error_phase = 0;    /* Static flag to avoid loops */
    long       saved_phase;    /* Saved content of error_phase */

    /*   There are three types of loop associated with Error Signalling.*/
    /*   One would occur if we did a LONGJMP to a location that caused  */
    /*   an error to be signalled that took the same LONGJMP.  We       */
    /*   prevent this by cancelling any LONGJMP error traps prior to    */
    /*   taking the jump.  The second problem is if we are asked to     */
    /*   report the error and another error is signalled while we are   */
    /*   doing the reporting.  We handle this by setting a flag for the */
    /*   duration of the reporting and performing an error exit if we   */
    /*   encounted such a "nested" error.                   */
    /*                                  */
    /*   The third problem is much nastier and occurs if the        */
    /*   diagnostics stack becomes so corrupt such that whatever we do  */
    /*   we hit an access violation.  We handle that by splitting the   */
    /*   error handling into phases and trying to cope with recursion   */
    /*   at each point.                         */

    if  ((error_phase == 1) || (error_phase == 3))
        /* If in phase 1 or phase 3 */
    {
        (void) fprintf (stderr,
                        "Fatal Recursive Error in PSL_SIGNAL_ERR\n");
        /* Output a nasty error message */
        PSL_EXIT();         /* and exit at once */
    }
    saved_phase = error_phase;  /* Save current Phase */
    error_phase = 1;        /* Show we're doing Phase 1 */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Save latest error details for use by PSL_GET_LSTERROR */

    (*diag_ptr).diag_signuml = errnum;
    /* Save error number */
    (*diag_ptr).diag_sigclsl = errcls;
    /* and class */

    /*   Check to see if the error class if zero, and, if so, just exit. */

    if (errcls == 0)        /* If no error class */
    {
        error_phase = 0;        /* Show process complete */
        return;         /* And exit */
    }

    /*    If we were in the middle of phase 2 error reporting then */
    /*    skip immediately to phase 3 */

    if  (saved_phase == 2)      /* Were we reporting an error already? */
    {
        ERROR_EXIT (diag_ptr,   /* Yes - exit with error */
                    "PSL_SIGNAL_ERROR called while reporting previous error");
    }

    /*   The error registry is kept with the most recent at the front.  */
    /*   Search it, looking for the first (most recent) entry that can  */
    /*   handle the error.                          */
    /*                                  */
    /*  If there isn't one, report the error and exit.          */

    err_ptr = (*diag_ptr).diag_errtrpps;
    /* Get most recent error trap */
    while (err_ptr != NULL)     /* While not at end of list */
    {
        trpcls = (*err_ptr).err_classl;
        /* Extract current class */
        trpnum = (*err_ptr).err_msgnuml;
        /* and current number */
        response = (*err_ptr).err_responsel;
        /* and required response */

        if (((trpcls == PSLERRCLS_NONE) || (trpcls == errcls))
                /* If class is wild or matches */
                && ((trpnum == 0) || (trpnum == errnum)))
            /* and number is wild or matches */
        {
            goto handle_error;  /* Go and handle the error */
        }
        err_ptr = (*err_ptr).err_prev_ptrps;
        /* Reset to previous handler */
    }

    /*   Error - no entry to handle the error being signalled */

    error_phase = 3;        /* Move to phase 3 */
    ERROR_EXIT (diag_ptr,
                "PSL_SIGNAL_ERROR called with no appropriate error handler");

    /*   We found someone with an interest in the error - see what they */
    /*   want us to do with it.                     */

handle_error:

    /*   If the response was "Call user action routine" then we must do */
    /*   that first as it might return to us to ask us to do something  */
    /*   else (typically PSLERRACT_RETURN).                 */

    if (response == PSLERRACT_CALL) /* If "Call User Routine" */
    {
        usrrtn_pf = (*err_ptr).err_handlerpf;
        /* Get address of user routine */
        response = (*usrrtn_pf)(    /* And call it */
                       diag_ptr,   /* passing diagnostics pointer */
                       errcls, /* and error class */
                       errnum);    /* and error number */
    }

    /*   Next check for the special case of PSLERRCLS_EXCEPT with a     */
    /*   response of either PSLERRACT_REPORT or PSLERRACT_RETURN.  If   */
    /*   found, reset the response to PSLERRACT_EXIT.           */

    if  ((errcls == PSLERRCLS_EXCEPT) /* If an Exception Error */
            &&  ((response == PSLERRACT_REPORT)
                 /* and either Report */
                 ||   (response == PSLERRACT_RETURN)))
        /* or Return */
    {
        response = PSLERRACT_EXIT;  /* Reset to Report & Exit */
    }

    /*   Now see if the response is one that requires the error to be   */
    /*   reported and, if so, report it.  Note that we normally just    */
    /*   output the last error, but we might output the whole stack if  */
    /*   the appropriate trace flag is set.                 */

    if  ((response == PSLERRACT_REPORT)
            /* If Report */
            ||  (response == PSLERRACT_REPJMP)
            /* or Report and Jump */
            ||  (response == PSLERRACT_EXIT))
        /* or Report and Exit */
    {
        error_phase = 2;        /* Say we're reporting error */
        if  (PSL_TEST_TRACE(PTRC_GEN_ERRORS))
            /* If we're tracing errors */
        {
            PSL_OUTPUT_DIAG (diag_ptr, 0, DIOUFL_ALL);
            /* Output the Entire Stack */
        }
        else            /* Else (only want last error) */
        {
            PSL_OUTPUT_DIAG (diag_ptr, 0, DIOUFL_LSTERR);
            /* Just output last error */
        }
        error_phase = 3;        /* Say we've finished reporting error */
    }

    /*   Now case on the response to do the requested action */

    switch (response)
    {
    case PSLERRACT_RETURN:  /* If Return */
    case PSLERRACT_REPORT:  /* or Report (done above) */
    case PSLERRACT_CALL:    /* or Call (done above) */
        {
            error_phase = 0;    /* Show process complete */
            return;         /* and return */
        }
    case PSLERRACT_EXIT:    /* If Exit */
        {
            error_phase = 0;    /* Show process complete */
            PSL_EXIT();     /* and exit at once */
            break;
        }
    case PSLERRACT_JUMP:    /* If Jump */
    case PSLERRACT_REPJMP:  /* or Report and Jump */
        {
            dest_ptr = (*err_ptr).err_jmp_ptrpj;
            /* Get destination for longjmp */
            PSL_FLUSH_ERRORS (diag_ptr, (*err_ptr).err_numberl);
            /* Flush the intervening error traps */
            /* as well as ourselves */
            error_phase = 0;    /* Show process complete */
            longjmp(*dest_ptr, errcls);
            /* And do the jump */
        }
    default:            /* Owt else is bad */
        {
            error_phase = 3;    /* Move to Phase 3 */
            (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "During PSL_SIGNAL_ERROR");
            /* Add a diagnostic */
            (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad response value %d",
                                       response);  /* describing problem */
            ERROR_EXIT(diag_ptr, "PSL-F-BADPARAM, Bad Parameter Value");
            /* And exit with it */
        }
    }
}

/************************************************************************/
/*                                  */
/*  PSL_SUPDIAG - Add supplementary diagnostics to the last error   */
/*                                  */
/*  This routine adds suplementary diagnostics to the last error.   */
/*  It is needed because secondary messages cannot have     */
/*  substitution parameters.                    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SUPDIAG (diag_ptr, numseg, msg1, msg1prm1, msg1prm2, ...,   */
/*                     msg2, msg3, ...);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     msgi     = Message number for ith message.           */
/*     msg1prmj = jth Parameter for 1st message.            */
/*                                  */
/*  The number and type of the parameters for each string are   */
/*  determined by the insertion directives in the string itself.    */
/*  These directives are as defined for PSL_FAO.            */
/*                                  */
/*    This routine forms a shell around PSL_SUPDIAG_L which does all of */
/*    the work.                             */
/*                                  */
/************************************************************************/

void PSL_SUPDIAG (struct diagstk *usrdiag_ptr, unsigned long numseg, ...)

{
    /*
     *    Local Data   
     */
    va_list    args;            /* Argument List Pointer */

    /*
     *    Set up the Variable Argument List pointer and call PSL_SUPDIAG_L
     */
    va_start (args, numseg);        /* Initialize Pointer */

    PSL_SUPDIAG_L (usrdiag_ptr, numseg, &args);
    /* Call PSL_SUPDIAG_L */

    /*   All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_SUPDIAG_L - Add supplementary diagnostics to the last error */
/*                                  */
/*  This routine adds suplementary diagnostics to the last error.   */
/*  It is needed because secondary messages cannot have     */
/*  substitution parameters.                    */
/*                                  */
/*  It is similar to PSL_SUPDIAG but uses a different calling   */
/*  format for instances when an argument list is required to be    */
/*  passed.                             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PSL_SUPDIAG_L (diag_ptr, numseg, va_ptr)            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     va_ptr  = Pointer to va_list structure for extra parameters  */
/*                                  */
/*  The routine works as follows:                   */
/*                                  */
/*  1. Save context of last error                   */
/*  2. Invert last error on diagnostics stack           */
/*  3. Call PUTDIAG_L to add supplementary diagnostics      */
/*  4. Invert last error on diagnostics stack           */
/*  5. Fix the last error context to the saved value        */
/*  6. Invert last error on diagnostics stack           */
/*                                  */
/*  Assume that the last error on the stack consists of 3 lines:    */
/*                                  */
/*      L3                          */
/*      L2                          */
/*      L1                          */
/*                                  */
/*  (reading the stack from the bottom).                */
/*                                  */
/*  After step 2 we have:                       */
/*                                  */
/*      L1                          */
/*      L2                          */
/*      L3                          */
/*                                  */
/*  Assume that step 3 adds another 3 line diagnostic. We then have:*/
/*                                  */
/*      L1                          */
/*      L2                          */
/*      L3                          */
/*      L6                          */
/*      L5                          */
/*      L4                          */
/*                                  */
/*  After step 4 we have:                       */
/*                                  */
/*      L1                          */
/*      L2                          */
/*      L3                          */
/*      L4                          */
/*      L5                          */
/*      L6                          */
/*                                  */
/*  Finally, after step 6 we have:                  */
/*                                  */
/*      L6                          */
/*      L5                          */
/*      L4                          */
/*      L3                          */
/*      L2                          */
/*      L1                          */
/*                                  */
/************************************************************************/

void PSL_SUPDIAG_L (struct diagstk *usrdiag_ptr, unsigned long numseg
                    ,va_list *va_ptr)

{
    /*
     *    Local Data   
     */
    unsigned long       context;        /* Saved context */

    /*  Save the current last error context */

    context = PSL_GET_ERRORCTX (usrdiag_ptr);

    /*  Invert the last error on the stack */

    PSL_DIAG_INVERT (usrdiag_ptr);

    /*    Pass the argument list straight through to PUTDIAG_L */

    PUTDIAG_L (usrdiag_ptr, numseg, va_ptr, PUTDIAG_SUPP);
    /* Call PUTDIAG_L */

    /*  Invert the last error on the stack */

    PSL_DIAG_INVERT (usrdiag_ptr);

    /*  Fix the context of the last error to the saved value */

    PSL_FIX_ERRORCTX (usrdiag_ptr, context);

    /*  Invert the last error on the stack */

    PSL_DIAG_INVERT (usrdiag_ptr);

    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_TRAP_ERRORS - Add an item to the Error Registry.        */
/*                                  */
/*  This routine registers the caller's interest in handling the    */
/*  the specified error(s).  Various different responses to any */
/*  particular error can be active at any time, but the most    */
/*  recent call to this routine is the one that would be triggered. */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  context = PSL_TRAP_ERRORS (diag_ptr, errcls, errnum,        */
/*                 response, jmp_ptr, handler);     */
/*                                  */
/*  Where:                              */
/*                                  */
/*  context  = Unsigned Long Context of error trap item.        */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  errcls   = The class of errors to be trapped (0 => All)     */
/*         (Should be one of the PSLERRCLS_*.)          */
/*  errnum   = The number of the error to be trapped (0 => All) */
/*  response = The response to take on an error.            */
/*         (Should be one of the PSLERRACT_*.)          */
/*  jmp_ptr  = A pointer to a jmp_buf set up by setjmp, or NULL.    */
/*  handler  = A pointer to an error handling function, or NULL.    */
/*                                  */
/************************************************************************/

unsigned long PSL_TRAP_ERRORS (struct diagstk  *usrdiag_ptr,
                               unsigned long   errcls,
                               unsigned long   errnum,
                               unsigned long   response,
                               jmp_buf         *jmp_ptr,
                               unsigned long (*handler) (
                                   struct diagstk *,
                                   unsigned long,
                                   unsigned long))
{
    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    struct   errtrap    *err_ptr;       /* Pointer to Error Trap */
    struct   errtrap    *prev_ptr;      /* and to previous one */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   If PSL_TRACING is enabled, perform some sanity checking on the */
    /*   argument list.                         */

#ifdef  PSL_TRACING
    CHECK_TRAP_ERRORS (diag_ptr, errcls, errnum,
                       response, jmp_ptr, handler);
#endif  /* PSL_TRACING */

    /*   Allocate another errtrap structure and link it in at the front */
    /*   of the current entries (if any).  This way, searching for an   */
    /*   appropriate entry in PSL_SIGNAL_ERROR() is easy.           */

    err_ptr = (struct errtrap *) malloc(sizeof(struct errtrap));
    /* Allocate new errtrap structure */
    if (err_ptr == NULL)        /* If allocation failed */
    {
        ERROR_EXIT (diag_ptr, "Memory Allocation for Error Trap Failed");
        /* report error and exit */
    }

    /*   Fill in the information that other routines may need */

    (*err_ptr).err_classl = errcls; /* Insert error class */
    (*err_ptr).err_msgnuml = errnum;
    /* and number */
    (*err_ptr).err_responsel = response;
    /* and response */
    (*err_ptr).err_jmp_ptrpj = jmp_ptr;
    /* and jmpbuf pointer */
    (*err_ptr).err_handlerpf = handler;
    /* and handler routine address */
    prev_ptr = (*diag_ptr).diag_errtrpps;
    /* Get previous error trap pointer */
    (*err_ptr).err_prev_ptrps = prev_ptr;
    /* and store it */
    (*err_ptr).err_contextl = (*diag_ptr).diag_cursubl
                              + (*diag_ptr).diag_biasl;
    /* and current diagnostics context */
    if  (prev_ptr == NULL)      /* If no previous error trap */
    {
        (*err_ptr).err_numberl = 1; /* this is number 1 */
    }
    else                /* Else (previous trap) */
    {
        (*err_ptr).err_numberl = (*prev_ptr).err_numberl + 1;
        /* this is one higher than that */
    }

    /*   Link into the diagnostics stack */

    (*diag_ptr).diag_errtrpps = err_ptr;
    /* Set error trap pointer */

    /*   All done - return context to the caller */

    return (*err_ptr).err_numberl;
}

/************************************************************************/
/*                                  */
/*  RESET_STATUS - Reset the Worst Error Status         */
/*                                  */
/*      This routine resets the 'Worst Error' status maintained by  */
/*      PSL_PUTDIAG and used by PSL_EXIT.               */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  RESET_STATUS (new_status_ptr, flag_ptr)             */
/*                                  */
/*  Where:                              */
/*                                  */
/*  new_status_ptr  = Pointer to Unsigned Long new status       */
/*  flag_ptr    = Pointer to flag:              */
/*          0 - reset WORST_ERROR only if new_status    */
/*              is worse than current value (default)   */
/*          1 - reset WORST_ERROR regardless        */
/*                                  */
/************************************************************************/
void    RESET_STATUS  (unsigned long *new_status_ptr,
                       unsigned char *flag_ptr)
{
    /*   Check to see if the status represents an error and, if so, set */
    /*   the global flag shared with PSL_EXIT.  Otherwise, check to see */
    /*   if the flag is set and, if so, clear the global flag.      */

    if  (!(psl_test_status(*new_status_ptr)))
        /* Is this status an error? */
    {
        error_reported = PTRUE; /* Yes - always set the flag */
    }
    else if ((*flag_ptr) == 1)  /* else if regardless flag is set */
    {
        error_reported = PFALSE;    /* Clear the flag */
    }
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  ADD_DIAGNOSTIC - Add diagnostic to diagnostic stack     */
/*                                  */
/*  This routine adds a diagnostic to a diagnostic stack;       */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*  ival = ADD_DIAGNOSTIC (diag_ptr, msg_ptr, flags);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*  ival     = Unsigned Long Context prior to added diagnostic. */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  msg_ptr  = Pointer to Message Text to add to stack.     */
/*  flags    = Unsigned Long Message Flags:             */
/*           DIAGFL_STAT => Message is a Static Message     */
/*                                  */
/************************************************************************/

static unsigned long ADD_DIAGNOSTIC (struct diagstk *usrdiag_ptr,
                                     char *msg_ptr,
                                     unsigned long flags)
{

    /*   Local Data   */

    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    long       msgsiz;     /* Formatted Message Size */
    char       **entry_ptr;    /* Local Entry Pointer */
    unsigned long       *flags_ptr;     /* And entry flags pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   First see if there is room on the stack for another entry and, */
    /*   if not, shuffle the entries to make some room.         */

    if  ((*diag_ptr).diag_cursubl == DIAG_STKSIZ)
        /* If already at first entry */
    {
        if  (((*diag_ptr).diag_flagsl100[0] & DIAGFL_STAT) == 0)
            /* If oldest message is dynamic */
        {
            free ((*diag_ptr).diag_stackpc100[0]);
            /* Free it */
        }
        (void) memmove (&(*diag_ptr).diag_flagsl100[0],
                        &(*diag_ptr).diag_flagsl100[1],
                        (DIAG_STKSIZ-1)*sizeof(unsigned long));
        /* Shuffle the flags up by one */
        (void) memmove (&(*diag_ptr).diag_stackpc100[0],
                        &(*diag_ptr).diag_stackpc100[1],
                        (DIAG_STKSIZ-1)*sizeof(char *));
        /* Shuffle the stack up by one */
        (*diag_ptr).diag_biasl += 1; /* And increment the bias */
    }
    else                /* Else (room for another) */
    {
        (*diag_ptr).diag_cursubl += 1;
        /* Just decrement subscript */
    }

    /*   If the message is static, then we just store its pointer in    */
    /*   the stack.  Otherwise we try to allocate a buffer to hold the  */
    /*   message and copy the message into it.  If the allocation       */
    /*   fails, we leave a null pointer in place to record the fact.    */

    entry_ptr = &(*diag_ptr).diag_stackpc100[(*diag_ptr).diag_cursubl - 1];
    /* Get pointer to entry */
    flags_ptr = &(*diag_ptr).diag_flagsl100[(*diag_ptr).diag_cursubl - 1];
    /* And to flags for entry */
    *flags_ptr = flags;     /* Store message flags */
    if  ((flags & DIAGFL_STAT) != 0) /* If message is static */
    {
        *entry_ptr = msg_ptr;   /* Just store message pointer */
    }
    else                /* Else (message is dynamic) */
    {
        msgsiz = strlen (msg_ptr);  /* Get the message size */
        *entry_ptr = (char *) malloc (msgsiz+1);
        /* Allocate room for the message */
        /* and store pointer to it */
        if  (*entry_ptr != NULL)    /* If allocation succeeded */
        {
            (void) memcpy (*entry_ptr, msg_ptr, msgsiz+1);
            /* Copy the message over */
            /* (including trailing null) */
        }
    }               /* End of (if message static) */

    /*   Lastly return the subscript (including bias) of the entry      */
    /*   prior to this one.                         */

    return ((*diag_ptr).diag_cursubl + (*diag_ptr).diag_biasl - 1);
    /* Return previous subscript */
}

/************************************************************************/
/*                                  */
/*  CHECK_TRAP_ERRORS - Check params to PSL_TRAP_ERRORS     */
/*                                  */
/*  This routine checks the parameters for plausibility.  Be aware  */
/*  that many values and combinations are hard or impossible to     */
/*  check.                              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  CHECK_TRAP_ERRORS (diag_ptr, errcls, errnum,            */
/*             response, jmp_ptr, handler);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  errcls   = The error class (one of PSLERRCLS_*).        */
/*  errnum   = The error number.                    */
/*  response = The response to take on an error.            */
/*         (Should be one of the PSLERRACT_*.)          */
/*  jmp_ptr  = A pointer to a jmp_buf set up by setjmp, or NULL.    */
/*  handler  = A pointer to an error handling function, or NULL.    */
/*                                  */
/************************************************************************/

#ifdef  PSL_TRACING

static void CHECK_TRAP_ERRORS (struct diagstk  *diag_ptr,
                               unsigned long   errcls,
                               unsigned long   errnum,
                               unsigned long   response,
                               jmp_buf         *jmp_ptr,
                               unsigned long   (*handler)())
{
    /*   Local Data   */

    unsigned long       diag_context;   /* Diagnostics Context */
    Boolean    error_found;    /* Error found flag */

    /*   Initialize our "error found" flag and add a diagnostic to the  */
    /*   diagnostics stack to say what we're doing              */

    error_found = PFALSE;       /* No errors found yet */
    diag_context = PSL_ADD_DIAGNOSTIC (diag_ptr,
                                       "During Error Trapping Routines");
    /* Say what we're doing */

    /*   First check class is valid */

    if  (errcls > PSLERRCLS_MAX)    /* If an invalid value */
    {
        error_found = PTRUE;    /* show we have an error */
        (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad Class %d", errcls);
        /* and add a diagnostic */
    }

    /*   Next check response is valid and, for each response, check     */
    /*   that the parameters look valid as well.                */

    switch (response)
    {
    case PSLERRACT_RETURN:  /* If return */
    case PSLERRACT_REPORT:  /* or report */
    case PSLERRACT_EXIT:    /* or exit */
        {
            if (jmp_ptr != NULL)    /* If a jmp_buf specified */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter jmp_ptr Not NULL");
                /* and add a diagnostic */
            }
            if (handler != NULL)    /* If a handler specified */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter handler Not NULL");
                /* and add a diagnostic */
            }
            break;          /* And break out of switch */
        }
    case PSLERRACT_JUMP:    /* If a jump */
    case PSLERRACT_REPJMP:  /* or a report and jump */
        {
            if (jmp_ptr == NULL)    /* If no jmp_buf structure */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter jmp_ptr Is NULL");
                /* and add a diagnostic */
            }
            if (handler != NULL)    /* If a handler specified */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter handler Not NULL");
                /* and add a diagnostic */
            }
            break;          /* And break out of switch */
        }
    case PSLERRACT_CALL:    /* If a call */
        {
            if (jmp_ptr != NULL)    /* If a jmp_buf specified */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter jmp_ptr Not NULL");
                /* and add a diagnostic */
            }
            if (handler == NULL)    /* If a handler not specified */
            {
                error_found = PTRUE; /* flag an error */
                (void) PSL_ADD_DIAGNOSTIC (diag_ptr,
                                           "Parameter handler Is NULL");
                /* and add a diagnostic */
            }
            break;          /* And break out of switch */
        }
    default:            /* Anything else is bad */
        {
            error_found = PTRUE;    /* flag an error */
            (void) PSL_ADD_DIAGNOSTIC (diag_ptr, "Bad Response %d",
                                       response);  /* and add a diagnostic */
        }
    }               /* End of (switch) */

    /*   If we detected any errors then we will log them via ERROR_EXIT */
    /*   and exit.  This is reasonable as this is PSL_TRACING code.     */
    /*   Otherwise, we should call PSL_SIGNAL_ERROR instead.        */

    if (error_found)        /* If we had an error */
    {
        ERROR_EXIT (diag_ptr, "PSL-F-BADPARAM, Bad Parameter Value");
        /* Exit with it */
    }

    /*   All OK - restore the diagnostics stack and exit */

    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_context);
    return;
}

#endif  /* PSL_TRACING */

/************************************************************************/
/*                                  */
/*  CREATE_DIAG_BUFFER - Create a buffer of diagnostics     */
/*                                  */
/*  This routine buffers some or all of the diagnostics in the  */
/*  stack into a contiguous character buffer.           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     CREATE_DIAG_BUFFER (diag_ptr, context, width, wrap_flag, */
/*                 bufptr_ptr, count_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr   = Pointer to Diagnostics Stack.          */
/*  context    = Unsigned Long Context of first diagnostic NOT to   */
/*           be included in buffer.             */
/*  width      = Unsigned long width at which to wrap/truncate. */
/*  wrap_flag  = Non-zero if wrapping at margin required.       */
/*      bufptr_ptr = Address of (char *) variable to hold pointer to    */
/*           buffer containing diagnostics.  Set to NULL if */
/*           buffer allocation failed.              */
/*  count_ptr  = Address of unsigned long to hold number of lines   */
/*           in formatted buffer.               */
/*                                  */
/************************************************************************/

static void CREATE_DIAG_BUFFER (struct diagstk *diag_ptr,
                                unsigned long context,
                                unsigned long width,
                                unsigned long wrap_flag,
                                char **bufptr_ptr,
                                unsigned long *count_ptr)
{

    /*   Local Data   */

    char       *item_ptr;      /* Local Item Pointer */
    unsigned long       itmsubl;        /* Local Item Subscript */
    char       *buf_ptr;       /* Pointer to buffer */
    unsigned long       diag_size;      /* Size of diagnostic */

    /*   First calculate the number of lines we need in the output      */
    /*   buffer.  If long messages are to be truncated then this is     */
    /*   just a count of the number of messages to be output.       */
    /*   Otherwise we must scan through the user and count the number   */
    /*   of segments needed for each one.                   */

    if  (wrap_flag == 0)        /* If truncating */
    {
        (*count_ptr) = (*diag_ptr).diag_cursubl +
                       (*diag_ptr).diag_biasl - context;
        /* Calculate number of items */
    }
    else                /* Else (wrapping) */
    {
        itmsubl = (*diag_ptr).diag_cursubl;
        /* Initialize subscript */
        (*count_ptr) = 0;       /* and count */
        while ((itmsubl != 0)   /* While there is more on the stack */
                &&   (context < (itmsubl + (*diag_ptr).diag_biasl)))
            /* And we haven't reached our target */
        {
            item_ptr = (*diag_ptr).diag_stackpc100[itmsubl - 1];
            /* Get pointer to item for current entry */
            if (item_ptr == NULL)   /* If pointer is null */
            {
                item_ptr = MSG_ALLOCDIAGFAIL;
                /* Log "Allocation Failed" instead */
            }
            diag_size = strlen (item_ptr);
            /* Get length of diagnostic */
            (*count_ptr) += ((diag_size + width - 1) / width);
            /* and hence number of lines needed */
            itmsubl -= 1;       /* And decrement current */
        }               /* End of (while more on stack) */
    }               /* End of (if wrapping messages) */

    /*   Now allocate the amount of space needed for the buffer and     */
    /*   initialize it to nulls.                        */

    buf_ptr = (char *) calloc ((*count_ptr), width);
    /* Allocate/initialize Space Required */
    *bufptr_ptr = buf_ptr;      /* And return start pointer to caller */

    /*   If the allocation succeeded - copy in the diagnostics */

    if  (buf_ptr != NULL)       /* If allocation succeeded */
    {
        itmsubl = (*diag_ptr).diag_cursubl;
        /* Initialize subscript */
        while ((itmsubl != 0)   /* While there is more on the stack */
                &&   (context < (itmsubl + (*diag_ptr).diag_biasl)))
            /* And we haven't reached our target */
        {
            item_ptr = (*diag_ptr).diag_stackpc100[itmsubl - 1];
            /* Get pointer to item for current entry */
            if (item_ptr == NULL)   /* If pointer is null */
            {
                item_ptr = MSG_ALLOCDIAGFAIL;
                /* Log "Allocation Failed" instead */
            }
            diag_size = strlen (item_ptr);
            /* Get length of diagnostic */
            while (diag_size != 0)  /* While more to output */
            {
                diag_size = strlen (item_ptr);
                /* Get length of remaining bit */
                if  (diag_size > width)
                    /* If too large for buffer */
                {
                    diag_size = width;
                    /* Truncate at right margin */
                }
                (void) memcpy (buf_ptr, item_ptr, diag_size);
                /* Copy it to save buffer */
                buf_ptr += width;   /* Point to next line in buffer */
                if  (wrap_flag == 0) /* If truncating */
                {
                    diag_size = 0;  /* Just flag no more to do */
                }
                else        /* Else (wrapping) */
                {
                    item_ptr += diag_size;
                    /* Update diagnostic pointer */
                    diag_size = strlen (item_ptr);
                    /* Get length of what's left */
                }
            }           /* End of (while more to do) */
            itmsubl -= 1;       /* And decrement current */
        }               /* End of (while more on stack) */
    }               /* End of (if allocation failed) */

    /*   All done - return to caller */

    return;
}

/************************************************************************/
/*                                  */
/*  ERROR_EXIT - Report error(s) and exit.              */
/*                                  */
/*  This routine outputs all diagnostics and then exits.        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  ERROR_EXIT (diag_ptr, errmsg);                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  errmsg   = A pointer to an error message, or NULL.      */
/*                                  */
/************************************************************************/

static void ERROR_EXIT (struct diagstk *diag_ptr, char *errmsg)
{
    if (errmsg != NULL)     /* If a message was specified */
    {
        (void) PSL_ADD_DIAGNOSTIC (diag_ptr, errmsg);
        /* add it to the stack */
    }

    /*   If we're on UNIX and we're outputting to the terminal, then    */
    /*   try to close down CURSES first.  In all cases, we then inhibit */
    /*   all interception and output the whole stack.           */

#ifdef  UNIX
    if  (((*diag_ptr).diag_outtypl & DIAG_OUTTYP_TCB) != 0 )
        /* If Output to TCB requested */
    {
        endwin();           /* Close down CURSES */
    }
#endif  /* UNIX */

    (*diag_ptr).diag_outtypl = 0;   /* Inhibit all interception */
    PSL_OUTPUT_DIAG (diag_ptr, 0, DIOUFL_ALL);
    /* Output all diagnostics */
    errexit_flag = PTRUE;       /* Remember we're exiting */
    PSL_EXIT();         /* And exit */
}

/************************************************************************/
/*                                  */
/*  FLUSH_ERDIAG - Flush items from the Error Registry (MK II). */
/*                                  */
/*  This routine removes all items in the error registry that were  */
/*  recorded after a particular diagnostic message.  It is called   */
/*  by PSL_REMOVE_DIAGNOSTICS and is not available for general use. */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  FLUSH_ERDIAG (diag_ptr, context);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context of diagnostic message defining */
/*          limit of items to be deleted.           */
/*         0 => clear error registry.               */
/*                                  */
/************************************************************************/

static void FLUSH_ERDIAG (struct diagstk *diag_ptr, unsigned long context)
{
    /*   Local Data   */

    struct   errtrap    *err_ptr;       /* Pointer to Error Trap */
    struct   errtrap    *prev_ptr;      /* and to previous one */

    /*   Free errtrap structures until the context is found. */

    err_ptr = (*diag_ptr).diag_errtrpps;
    /* Get latest error trap */
    while ((err_ptr != NULL)    /* While more levels to check */
            &&  (context < (*err_ptr).err_contextl))
        /* and we've not reached specified level*/
    {
        prev_ptr = (*err_ptr).err_prev_ptrps;
        /* Get pointer to previous level */
        free(err_ptr);      /* Release this level */
        err_ptr = prev_ptr;     /* And check previous one */
    }

    /*   We've found the latest acceptable one - store it as the current one */

    (*diag_ptr).diag_errtrpps = err_ptr;
    /* Set error trap pointer */

    /*   All done - return to the caller */

    return;
}

/************************************************************************/
/*                                  */
/*  OUTPUT_TO_STDERR - Output diagnostics to stderr         */
/*                                  */
/*  This routine outputs some or all of the diagnostics in the  */
/*  stack to the standard error unit.               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     OUTPUT_TO_STDERR (diag_ptr, context);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context of first diagnostic NOT to be  */
/*         output.                      */
/*                                  */
/************************************************************************/

static void OUTPUT_TO_STDERR (struct diagstk *diag_ptr,
                              unsigned long context)
{

    /*   Local Data   */

    char       *item_ptr;      /* Local Item Pointer */
    unsigned long       itmsubl;        /* Local Item Subscript */

    /*   Just loop through the stack, outputting each message to        */
    /*   stderr.                                */

    itmsubl = (*diag_ptr).diag_cursubl;
    /* Initialize subscript */
    while ((itmsubl != 0)       /* While there is more on the stack */
            &&   (context < (itmsubl + (*diag_ptr).diag_biasl)))
        /* And we haven't reached our target */
    {
        item_ptr = (*diag_ptr).diag_stackpc100[itmsubl - 1];
        /* Get pointer to item for current entry */
        if (item_ptr == NULL)   /* If pointer is null */
        {
            item_ptr = MSG_ALLOCDIAGFAIL;
            /* Log "Allocation Failed" instead */
        }
        (void) fprintf (stderr, "%s\n", item_ptr);
        /* Output diagnostic to stderr */
        itmsubl -= 1;       /* And decrement current */
    }               /* End of (while more on stack) */

    /*   All done - return to caller */

    return;
}

/************************************************************************/
/*                                  */
/*  PUTDIAG_L - Put a standard message onto the diagnostics stack.  */
/*                                  */
/*  This routine does all the hard work of PSL_PUTDIAG and is   */
/*  called thereby.                         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  PUTDIAG_L (diag_ptr, numseg, va_ptr, flags);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack.         */
/*     numseg   = Number of message segments to output.     */
/*     va_ptr   = Pointer to va_list structure for extra parameters */
/*     flags    = Unsigned long flags:              */
/*          PUTDIAG_SUPP -> This is supplementary message   */
/*                                  */
/************************************************************************/

static void PUTDIAG_L (struct diagstk *usrdiag_ptr,
                       unsigned long numseg,
                       va_list *va_ptr,
                       unsigned long flags)
{
    /*
     *    Local Data
     */
    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    static   char       msgbuf[256];    /* Local Message Buffer */
    static   char       fmtbuf[256];    /* Local Format Buffer */
    unsigned long       curseg;     /* Current Segment Counter */
    unsigned long       msgnum;     /* Current Message Number */
    unsigned long       msglen;     /* Length of message text */
    char       *msg_ptr;       /* Pointer to next character in message */
    unsigned char       resetflg=0;     /* Reset Status only if Worse */
    char       *fmt_ptr;       /* Format Buffer Pointer */
    char       *nxt_ptr;       /* Next Segment Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*   Flag that we are starting an error sequence. */

    PSL_SET_ERRORCTX (diag_ptr);    /* Flag this is start of error */

    /*   Loop through each of the message segments, initially setting   */
    /*   up the message prefix and a pointer to the format string.      */

    for (curseg = 0; curseg < numseg; curseg++)
        /* For each message segment */
    {
        msgnum = va_arg (*va_ptr, unsigned long);
        /* Get associated message number */

        /*   If the first message segment then we want to format parameters */
        /*   into it via PSL_FAO_L.  This is potentially a problem if the   */
        /*   message string contains one or more !/ directives as PSL_FAO   */
        /*   explicitly disallows these.                    */

        if  (curseg == 0)       /* If first segment */
        {

            if  ((flags & PUTDIAG_SUPP) != 0)
                /* If a supplementary message */
            {
                (void) PSL_GET_MESSAGE (msgnum, msgbuf, sizeof(msgbuf),
                                        PSLMSG_PREFIX);
                /* Set up message prefix only */
            }
            else            /* Else (first message) */
            {
                (void) PSL_GET_MESSAGE (msgnum, msgbuf, sizeof(msgbuf),
                                        (PSLMSG_PREFIX | PSLMSG_UTLNAM));
                /* Set up prefix & utility name */
            }

            (void) PSL_GET_MESSAGE (msgnum, fmtbuf,
                                    sizeof(fmtbuf), PSLMSG_TEXT);
            /* Set up skeleton message */
            msglen = strlen(msgbuf); /* And length of prefix */
            msg_ptr = &msgbuf[msglen];
            /* And pointer to next character */
            *(msg_ptr++) = ','; /* And add ", " */
            *(msg_ptr++) = ' ';
            msglen += 2;        /* Increment message length */
            fmt_ptr = fmtbuf;   /* Initialize Format Pointer */
            nxt_ptr = strstr (fmt_ptr, "!/");
            /* Look for !/ directive */
            while (nxt_ptr != NULL) /* While there is one */
            {
                *nxt_ptr = '\0';    /* Replace with null-termination */
                PSL_FAO_L (msg_ptr, sizeof(msgbuf) - msglen,
                           fmt_ptr, va_ptr);
                /* Format the String */
                (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
                /* Add to diagnostics stack */
                msglen = 0;     /* Reset the message length */
                msg_ptr = msgbuf;   /* and pointer */
                fmt_ptr = nxt_ptr + 2;
                /* Reset to next bit of format */
                nxt_ptr = strstr (fmt_ptr, "!/");
                /* And look for next !/ directive */
            }
            PSL_FAO_L (msg_ptr, sizeof(msgbuf) - msglen, fmt_ptr, va_ptr);
            /* Format the final segment */
        }

        /*   If not the first segment, then just get the whole message */

        else            /* Else (not first segment) */
        {
            (void) PSL_GET_MESSAGE (msgnum, msgbuf, sizeof(msgbuf),
                                    (PSLMSG_PREFIX | PSLMSG_TEXT));
            /* Set up whole message */
        }



        /*   If not the first segment or a supplementary message, replace the leading */
        /*   character with a '-'..                               */
#ifdef VMS_MSG_FORMAT
        if  ((curseg != 0)      /* If not first segment */
                ||  ((flags & PUTDIAG_SUPP) != 0))
            /* or a supplementary message */
        {
            msgbuf[0] = '-';    /* Reset character at start */
        }
#endif


        /*   In all cases just add the (formatted) message to the stack. */

        (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
        /* Add to diagnostics stack */
    }               /* End of (for each message) */

    /*   All segments done - Invert the  diagnostics. */

    PSL_DIAG_INVERT (diag_ptr); /* Invert the diagnostics */

    /*   Call RESET_STATUS to keep the "Worst Error" up to date, and    */
    /*   return to the caller                       */

    RESET_STATUS (&msgnum, &resetflg);
    /* Reset 'Worst Error' */
    return;             /* Return to caller */
}

/************************************************************************/
/*                                  */
/*  VALIDATE_DIAGPTR - Validate specified Diagnostics Pointer   */
/*                                  */
/*  This routine checks that the diagnostics pointer supplied by    */
/*  the user is valid and, if not, replaces it with a pointer to    */
/*  the static diagnostics stack we keep in place for emergencies.  */
/*  Note that this routine should always be converted to in-line    */
/*  code by the compiler, but is kept separate for simplicity.  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr    = Returned pointer to valid Diagnostics Stack.  */
/*  usrdiag_ptr = User's Diagnostics Stack pointer.         */
/*                                  */
/************************************************************************/

static struct diagstk *VALIDATE_DIAGPTR (struct diagstk *usrdiag_ptr)
{
    /*   First check to see if the user's diagnostic pointer is NULL,   */
    /*   and, if so, substitute our own.  Note that this may never have */
    /*   been allocated so we need to set up the structure type just in */
    /*   case we need it.                           */

    if  (usrdiag_ptr == NULL)   /* If user's pointer is NULL */
    {
        fixed_diagstk.diag_strtypl = DIAG_STRNAM;
        /* Ensure Structure Name is set up */
        return &fixed_diagstk;  /* And return pointer to it */
    }

    /*   If the user's diagnostic pointer is not NULL, and we have      */
    /*   error tracing switched on, then we'll also check the structure */
    /*   type and exit with an error if it's dud.               */

#ifdef  PSL_TRACING
    else if ((*usrdiag_ptr).diag_strtypl != DIAG_STRNAM)
        /* Else, if it is invalid */
    {
        ERROR_EXIT (&fixed_diagstk,
                    "PSL-F-BADPARAM, Invalid Diagnostics Stack Pointer");
        /* Exit with an error */
        return NULL;        /* Keep Lint happy */
    }
#endif  /* PSL_TRACING */

    /*   Otherwise it must be OK, so we just return it to the caller */

    else                /* Else (exists & OK) */
    {
        return usrdiag_ptr;     /* Just return it */
    }
}

/************************************************************************/
/*                                  */
/*          OBSOLETE ROUTINES               */
/*                                  */
/************************************************************************/

/*   The PSL_ADD/REPLACE_DIAGNOSTIC_L routines originally existed   */
/*   as part of an earlier approach whereby the TRM routines had    */
/*   their own private interface to these routines.  This was       */
/*   withdrawn during V2.1, but we cannot remove the routines       */
/*   without breaking the upwards compatibility of SYSSHR.      */
/*   However, we can turn them into stubs as the systems that did   */
/*   use them never really did anything critical with them anyway.  */

void    PSL_ADD_DIAGNOSTIC_L (void)
{}


void    PSL_REPLACE_DIAGNOSTIC_L (void)
{}

/************************************************************************/
/*                                  */
/*  PSL_OUTPUT_DIAGNOSTICS - Output diagnostics from stack (old)    */
/*                                  */
/*  This routine outputs some or all of the diagnostics in the  */
/*  stack until it is returned to a specified previous position (or */
/*  as close as we can get.                     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_OUTPUT_DIAGNOSTICS (diag_ptr, context);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*  diag_ptr = Pointer to Diagnostics Stack.            */
/*  context  = Unsigned Long Context of first diagnostic NOT to be  */
/*         output. A context of 0 indicates that all        */
/*         diagnostics are to be output.            */
/*                                  */
/************************************************************************/

void PSL_OUTPUT_DIAGNOSTICS (struct diagstk *diag_ptr,
                             unsigned long context)
{
    if (context == 0)
    {
        PSL_OUTPUT_DIAG (diag_ptr, context, DIOUFL_ALL);
    }
    else
    {
        PSL_OUTPUT_DIAG (diag_ptr, context, 0);
    }

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_LOG_CIOERR - Log C File I/O error               */
/*                                  */
/*  This routine adds appropriate diagnostics to the Diagnostic */
/*  Stack after a failed C I/O operation.               */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*     PSL_LOG_CIOERR (diag_ptr, msgnum, ...);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack          */
/*     msgnum   = Message Number                    */
/*     ...      = Arguments required by message associated with */
/*            Message Number                    */
/*                                  */
/************************************************************************/

void    PSL_LOG_CIOERR (struct diagstk *usrdiag_ptr,
                        unsigned long msgnum,
                        ...)
{
    /*
     *    Local Data   
     */
    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    va_list    args;            /* Argument List Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*
     *    Set up the Variable Argument List pointer and call PSL_LOG_CIOERR_L
     */
    va_start (args, msgnum);        /* Initialize Pointer */

    PSL_LOG_CIOERR_L (diag_ptr, msgnum, &args);
    /* Call PSL_LOG_CIOERR_L */

    /*   All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_LOG_CIOERR_L - Log C File I/O error             */
/*                                  */
/*  This routine does all the hard work of PSL_LOG_CIOERR and is    */
/*  called thereby.                         */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*     PSL_LOG_CIOERR_L (diag_ptr, msgnum, va_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack          */
/*     msgnum   = Message Number                    */
/*     va_ptr   = Pointer to va_list structure for extra parameters */
/*                                  */
/************************************************************************/

static void PSL_LOG_CIOERR_L (struct diagstk *diag_ptr,
                              unsigned long msgnum,
                              va_list *va_ptr)
{
    /*
     *  Local Data.
     */
    static   char       unknown_err_msg [] = "Unknown File I/O Error Encountered";
    /* Message for Unknown Error */
    static   char       msgbuf [256];   /* Local Message Buffer */
    static   char       fmtbuf [256];   /* Local Format Buffer */
    unsigned long       msglen;     /* Length of message text */
    char       *msg_ptr;       /* Pointer to next character */
    /* ... in message */
    unsigned char       resetflg = 0;   /* Reset Status only if Worse */
    char       *fmt_ptr;       /* Format Buffer Pointer */
    char       *nxt_ptr;       /* Next Segment Pointer */

    int        saved_errno;    /* Saved copy of errno */

    /*
     *  On NTI (at least) PSL_GET_MESSAGE does things that can affect errno.
     *  Therefore we save it here and use the saved copy later.
     */
    saved_errno = errno;

    /*
     *  Flag that we are starting an error sequence.
     */
    PSL_SET_ERRORCTX (diag_ptr);    /* Flag this is start of error */

    /*
     *  Insert a "%" as the first character in the buffer so that we actually
     *  get one in the message - vsprintf (in PSL_ADD_DIAGNOSTIC) will throw
     *  away a single "%" as it indicates a formatting directive.
     */
#ifdef VMS_MSG_FORMAT
    msgbuf [0] = '%';
#endif

    /*
     *  We want to format parameters via PSL_FAO_L. This is potentially a
     *  problem if the message string contains one or more !/ directives
     *  as PSL_FAO explicitly disallows these.
     */
#ifdef VMS_MSG_FORMAT
    (void) PSL_GET_MESSAGE (msgnum, &msgbuf [1], sizeof (msgbuf) - 1,
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */
#else
    (void) PSL_GET_MESSAGE (msgnum, msgbuf, sizeof (msgbuf),
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */
#endif
    (void) PSL_GET_MESSAGE (msgnum, fmtbuf, sizeof (fmtbuf), PSLMSG_TEXT);
    /* Set up skeleton message */
    msglen = strlen (msgbuf);       /* And length of prefix */

    msgbuf [msglen] = ',';      /* And add ", " */
    msgbuf [msglen+1] = ' ';
    msglen += 2;            /* Increment message length */
    msg_ptr = &msgbuf [msglen];     /* Set pointer to next character */

    fmt_ptr = fmtbuf;           /* Initialize Format Pointer */

    /*
     *  Handle !\ directives.
     */
    nxt_ptr = strstr (fmt_ptr, "!/");   /* Look for !/ directive */

    while (nxt_ptr != NULL)     /* While there is one */
    {
        *nxt_ptr = '\0';        /* Replace with null-termination */
        PSL_FAO_L (msg_ptr, sizeof (msgbuf) - msglen, fmt_ptr, va_ptr);
        /* Format the String */
        (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
        /* Add to diagnostics stack */
        msglen = 0;         /* Reset the message length */
        msg_ptr = msgbuf;       /* and pointer */
        fmt_ptr = nxt_ptr + 2;      /* Reset to next bit of format */
        nxt_ptr = strstr (fmt_ptr, "!/");
        /* And look for next !/ directive */
    }
    PSL_FAO_L (msg_ptr, sizeof (msgbuf) - msglen, fmt_ptr, va_ptr);
    (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add to diagnostics stack */

    /*
     *  Try the various possible error numbers in turn to get a reason
     *  for the error.
     */
    if (saved_errno != 0)       /* If standard C error */
    {
        msg_ptr = strerror (saved_errno);
        /* Get pointer to standard message */
        (void) strncpy (&msgbuf [1], msg_ptr, sizeof (msgbuf) - 2);
        /* Copy it to Output Buffer */
    }
    else                /* Getting desparate! */
    {
        (void) strncpy (&msgbuf [1], unknown_err_msg, sizeof (msgbuf) - 2);
        /* Use fallback message */
    }
    msgbuf [0] = '-';           /* Insert leading "-" */
    (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add to diagnostics stack */

    /*
     *  Invert the Diagnostic Stack.
     */
    PSL_DIAG_INVERT (diag_ptr);     /* Invert Stack */

    /*
     *  Call RESET_STATUS to keep the "Worst Error" up to date.
     */
    RESET_STATUS (&msgnum, &resetflg);

    /*
     *  Reset the various File I/O error numbers.
     */
    PSL_CLEARERR ();

    return;
}

/************************************************************************/
/*                                  */
/*  PSL_SIG_CIOERR - Log C File I/O error and signal it     */
/*                                  */
/*  This routine adds appropriate diagnostics to the Diagnostic */
/*  Stack after a failed C I/O operation. It then calls     */
/*  PSL_SIGNAL_ERROR to signal the error.               */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*     PSL_SIG_CIOERR (diag_ptr, msgnum, ...);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack          */
/*     msgnum   = Message Number                    */
/*     ...      = Arguments required by message associated with */
/*            Message Number                    */
/*                                  */
/************************************************************************/

void    PSL_SIG_CIOERR (struct diagstk *usrdiag_ptr,
                        unsigned long msgnum,
                        ...)
{
    /*
     *    Local Data   
     */
    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    va_list    args;            /* Argument List Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*
     *    Set up the Variable Argument List pointer and call PSL_LOG_CIOERR_L
     */
    va_start (args, msgnum);        /* Initialize Pointer */

    PSL_LOG_CIOERR_L (diag_ptr, msgnum, &args);
    /* Call PSL_LOG_CIOERR_L */

    PSL_SIGNAL_ERROR (diag_ptr, PSLERRCLS_USER1, msgnum);
    /* Call PSL_SIGNAL_ERROR */

    /*  All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

#ifdef WIN32
/************************************************************************/
/*                                  */
/*  PSL_LOG_LASTERROR - Log last error              */
/*                                  */
/*  This routine adds appropriate diagnostics to the Diagnostic */
/*  Stack after a failed WIN32 operation.               */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*     PSL_LOG_LASTERROR (diag_ptr, msgnum, ...);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack          */
/*     msgnum   = Message Number                    */
/*     ...      = Arguments required by message associated with */
/*            Message Number                    */
/*                                  */
/************************************************************************/

void    PSL_LOG_LASTERROR (struct diagstk *usrdiag_ptr,
                           unsigned long msgnum,
                           ...)
{
    /*
     *    Local Data   
     */
    struct   diagstk    *diag_ptr;      /* Diagnostics Pointer */
    va_list    args;            /* Argument List Pointer */

    /*   Get a valid diagnostics pointer */

    diag_ptr = VALIDATE_DIAGPTR (usrdiag_ptr);
    /* Ensure we have valid pointer */

    /*
     *    Set up the Variable Argument List pointer and call PSL_LOG_LASTERROR_L
     */
    va_start (args, msgnum);        /* Initialize Pointer */

    PSL_LOG_LASTERROR_L (diag_ptr, msgnum, &args);
    /* Call PSL_LOG_LASTERROR_L */

    /*   All done - tidy up the Argument List and exit */

    va_end (args);          /* Tidy up argument list */
    return;             /* And return */
}

/************************************************************************/
/*                                  */
/*  PSL_LOG_LASTERROR_L - Log last WIN32 File I/O error     */
/*                                  */
/*  This routine does all the hard work of PSL_LOG_LASTERROR and is */
/*  called thereby.                         */
/*                                  */
/*      Calling Convention:                     */
/*                                  */
/*     PSL_LOG_LASTERROR_L (diag_ptr, msgnum, va_ptr);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     diag_ptr = Pointer to Diagnostics Stack          */
/*     msgnum   = Message Number                    */
/*     va_ptr   = Pointer to va_list structure for extra parameters */
/*                                  */
/************************************************************************/

static void PSL_LOG_LASTERROR_L (struct diagstk *diag_ptr,
                                 unsigned long msgnum,
                                 va_list *va_ptr)
{
    /*
     *  Local Data.
     */
    static   char       unknown_err_msg [] = "Unknown File I/O Error Encountered";
    /* Message for Unknown Error */
    static   char       msgbuf [256];   /* Local Message Buffer */
    static   char       fmtbuf [256];   /* Local Format Buffer */
    unsigned long       msglen;     /* Length of message text */
    char       *msg_ptr;       /* Pointer to next character */
    /* ... in message */
    unsigned char       resetflg = 0;   /* Reset Status only if Worse */
    char       *fmt_ptr;       /* Format Buffer Pointer */
    char       *nxt_ptr;       /* Next Segment Pointer */
    DWORD  dwError = GetLastError();    /* Get last error */
    LPVOID lpMessageBuffer = NULL;      /* Pointer to error text buffer */

    /*
     *  Flag that we are starting an error sequence.
     */
    PSL_SET_ERRORCTX (diag_ptr);    /* Flag this is start of error */

    /*
     *  Insert a "%" as the first character in the buffer so that we actually
     *  get one in the message - vsprintf (in PSL_ADD_DIAGNOSTIC) will throw
     *  away a single "%" as it indicates a formatting directive.
     */
#ifdef VMS_MSG_FORMAT
    msgbuf [0] = '%';
#endif

    /*
     *  We want to format parameters via PSL_FAO_L. This is potentially a
     *  problem if the message string contains one or more !/ directives
     *  as PSL_FAO explicitly disallows these.
     */
#ifdef VMS_MSG_FORMAT
    (void) PSL_GET_MESSAGE (msgnum, &msgbuf [1], sizeof (msgbuf) - 1,
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */
#else
    (void) PSL_GET_MESSAGE (msgnum, msgbuf, sizeof (msgbuf),
                            (PSLMSG_PREFIX | PSLMSG_UTLNAM));
    /* Set up message prefix */
#endif
    (void) PSL_GET_MESSAGE (msgnum, fmtbuf, sizeof (fmtbuf), PSLMSG_TEXT);
    /* Set up skeleton message */
    msglen = strlen (msgbuf);       /* And length of prefix */

    msgbuf [msglen] = ',';      /* And add ", " */
    msgbuf [msglen+1] = ' ';
    msglen += 2;            /* Increment message length */
    msg_ptr = &msgbuf [msglen];     /* Set pointer to next character */

    fmt_ptr = fmtbuf;           /* Initialize Format Pointer */

    /*
     *  Handle !\ directives.
     */
    nxt_ptr = strstr (fmt_ptr, "!/");   /* Look for !/ directive */

    while (nxt_ptr != NULL)     /* While there is one */
    {
        *nxt_ptr = '\0';        /* Replace with null-termination */
        PSL_FAO_L (msg_ptr, sizeof (msgbuf) - msglen, fmt_ptr, va_ptr);
        /* Format the String */
        (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
        /* Add to diagnostics stack */
        msglen = 0;         /* Reset the message length */
        msg_ptr = msgbuf;       /* and pointer */
        fmt_ptr = nxt_ptr + 2;      /* Reset to next bit of format */
        nxt_ptr = strstr (fmt_ptr, "!/");
        /* And look for next !/ directive */
    }
    PSL_FAO_L (msg_ptr, sizeof (msgbuf) - msglen, fmt_ptr, va_ptr);
    /* Format the final segment */
    (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add to diagnostics stack */

    /*
     * Format the LastError value
     */
    if (FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
                       FORMAT_MESSAGE_FROM_SYSTEM,
                       NULL,
                       dwError,
                       LANG_NEUTRAL,
                       (LPSTR) &lpMessageBuffer,
                       0,
                       NULL))       /* Get system message text */
    {
        (void) strncpy (&msgbuf [1], lpMessageBuffer, sizeof (msgbuf) - 2);
        /* Copy it to Output Buffer */
    }
    else                    /* It failed */
    {
        (void) strncpy (&msgbuf [1], unknown_err_msg, sizeof (msgbuf) - 2);
        /* Use fallback message */
    }
    msgbuf [0] = '-';           /* Insert leading "-" */
    (void) ADD_DIAGNOSTIC (diag_ptr, msgbuf, 0);
    /* Add to diagnostics stack */
    /*
     *  If a buffer was allocated to store the system message text
     *  we need to free it up.
     */
    if (lpMessageBuffer != NULL)        /* If buffer was allocated */
    {
        LocalFree ((HLOCAL) lpMessageBuffer);
        /* Free it */
    }

    /*
     *  Invert the Diagnostic Stack.
     */
    PSL_DIAG_INVERT (diag_ptr);     /* Invert Stack */

    /*
     *  Call RESET_STATUS to keep the "Worst Error" up to date.
     */
    RESET_STATUS (&msgnum, &resetflg);

    return;
}
#endif
