/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*            **** V - T E S T ****                  */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: VTEST_RTNS_WIN32                */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*    This module contains miscellaneous platform specific routines used     */
/*    by more than one V-TEST component.                     */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

#ifdef  WIN32               /* WIN32 SPECIFIC MODULE        */

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>

/*
 *  Standard WIN32 header files.
 */
#include <windows.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */
#include <trc_rtndef.h>         /* Trace Routines       */

static unsigned long VTEST_IS_DIR   (struct diagstk *diagstk_ptr,
                                     char    *filnam_ptr,
                                     Boolean *flag_ptr);
/*
 *  Structure Definitions.
 */
#include <vtest_srvinfdef.h>        /* VTEST Server information */

/*
 *  ...and general parameter files.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <pslerr_prmdef.h>      /* Error Constants      */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols   */
#include <psl_trc_prmdef.h>     /* Trace Routine Parameters     */
#include <psl_vm_prmdef.h>      /* LL Routine Parameters    */
#include <pslerr_prmdef.h>      /* Error Routine Symbols    */
#include <vtest_prmdef.h>       /* V-TEST Parameters        */

/*
 *  Define the two pipe functions
 */
#define READPIPE    0
#define WRITEPIPE   1




/************************************************************************/
/*                                  */
/*    VTEST_CREATE_PROCESS - Create a TEST Server on the local node */
/*    ===================                       */
/*                                  */
/*    Calling Format:                           */
/*  iret = VTEST_CREATE_PROCESS (diagstk_ptr            */
/*                 ,image_ptr               */
/*                 ,username                */
/*                 ,priority                */
/*                 ,spawned_process         */
/*                 ,&srvpid);               */
/*                                  */
/*    Where:                                */
/*  iret = PSL Status return                    */
/*  diagstk_ptr - Pointer to the diagnostic stack to use        */
/*  image_ptr   - Pointer to name of image to run.          */
/*  username    - Username under which the process is to be created */
/*  priority    - Priority for new process              */
/*  spawned_process - PTRUE if we are to spawn the process, PFALSE  */
/*            if a detached process is required.        */
/*  srvpid - Address of a long to return the PID of the server  */
/*                                  */
/*    This routine is called in order to create a Test Server process   */
/*    running on the local node                     */
/*                                  */
/************************************************************************/

unsigned long   VTEST_CREATE_PROCESS    (struct   diagstk *diagstk_ptr
        ,         char    *image_ptr
        ,         char    *username_ptr
        ,unsigned long    priority
        ,         Boolean spawned_process
        ,unsigned long    *srvpid_ptr)
{
    /*
     *  Local Data
     */
    unsigned long       iret;       // PSL return status
    STARTUPINFO     startupInfo = {0};  // Startup information
    PROCESS_INFORMATION processInformation; // Process information
    char            msgbuf[300];

    /*
     *  Assume success
     */
    iret = PSL_NORMAL;

    /*
     *  Output Trace Information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_MISC))
    {
        PSL_OUTPUT_TRACE ("Attempting to start a local process\n");
    }

    // Set up structures

    startupInfo.cb = sizeof (STARTUPINFO);

    // Create the process

    if (!CreateProcess (NULL,       // pointer to name of executable module
                        image_ptr,  // pointer to command line string
                        NULL,       // pointer to process security attributes
                        NULL,       // pointer to thread security attributes
                        FALSE,      // handle inheritance flag
                        DETACHED_PROCESS, // creation flags
                        NULL,       // pointer to new environment block
                        NULL,       // pointer to current directory name
                        &startupInfo,   // pointer to STARTUPINFO
                        &processInformation))
        // pointer to PROCESS_INFORMATION
    {
        (void)sprintf (msgbuf, "CreateProcess() [%s]", image_ptr);
        iret = PSL_DIAG_LASTERROR (diagstk_ptr, msgbuf);
    }
    else
    {
        *srvpid_ptr = processInformation.dwProcessId;
        // Return the pid
    }

    return iret;
}

/************************************************************************/
/*                                  */
/*    VTEST_LOCAL_SERVERS - Create a linked list of LOCAL TEST SERVERS  */

/*    Calling Convention:                       */
/*                                  */
/*  iret = VTEST_LOCAL_SERVERS (diagstk_ptr ,ll_ptr);       */
/*                                  */
/*    Where:                                */
/*  iret = PSL routine return status                */
/*  diagstk_ptr - Pointer to diagnostics stack to use.      */
/*  ll_ptr - Pointer to the LL to hold the records.         */
/*                                  */
/*    This routine is called to construct a linked list of all of the   */
/*    V-TEST Server processes which are present on the local machine    */
/*                                  */
/*    No notice is taken as to whether or not the current user will be  */
/*    able to access the server or not.                 */
/*                                  */
/*    After the list has been used, FREE_LOCAL_SERVERS should be called */
/*    to deallocated the linked list created.               */
/*                                  */
/************************************************************************/

unsigned long   VTEST_LOCAL_SERVERS  (struct   diagstk *diagstk_ptr
                                      ,struct   psl_vm_llb *srv_ll_ptr)

{
    unsigned long    iret = PSL_NORMAL; /* PSL return status value  */
    unsigned long    saved_diagctx;     /* Saved diagnostics context    */
    struct   srvinf  srvinf_rec;        /* Server list record       */
    long    keyfile_fd;        /* FD for keyfiles      */
    int     status;                /* Status from FILE functions */
    int     closests;              /* Status from close() */
    Boolean  param_error = PFALSE; /* Error in parameters flag */
    HANDLE  hFile;         // File handle
    HANDLE  hProcess;      // Process handle
    char    filepath[255];     /* Path of file         */
    char    filename[255];     /* Name of file         */
    WIN32_FIND_DATA FileData;  // File information
    Boolean foundFile;

    /*
     *    First things first, add a diagnostic message to the stack.
     */
    saved_diagctx = PSL_ADD_STATDIAG (diagstk_ptr
                                      ,"During VTEST_LOCAL_SERVERS");

    /*
     *    Validate the input parameters
     *
     *    The linked list pointer must not be zero.
     */
    if (srv_ll_ptr == 0)
    {
        param_error = PTRUE;
        (void) PSL_ADD_STATDIAG (diagstk_ptr ,"Invalid Linked List Pointer");
    }

    if (param_error)
    {
        iret = PSL_BADPARAM;
        PSL_PUTDIAG (diagstk_ptr ,2 ,PSL_SYSERVERR,
                     "VTEST_LOCAL_SERVERS",iret);
        goto end;
    }

    /*
     *  Locate the <psl_root>/<node> directory
     */
    PSL_GET_NODEDIR (filepath, sizeof (filename));
    strcat (filepath, "\\");

    /*
     *    Scan the directory for files named "VTEST_KEY_nnnnnnnn"
     */
    strcpy (filename, filepath);
    strcat (filename, "VTEST_KEY_????????");
    hFile = FindFirstFile(filename, &FileData);
    if (hFile == INVALID_HANDLE_VALUE) goto end;
    foundFile = PTRUE;

    /*
     *    This is a valid keyfile but may not represent a Test Executer or Server
     *    which is active at this time. Check for this by stripping the PID off the
     *    end of the end of the file name and checking for a process with this PID
     *
     *    If we cannot open the file for any reason we simply ignore it, we do not
     *    report an error as the problem is not really that important.
     */
    while (foundFile)
    {
        strcpy (filename, filepath);
        strcat (filename, FileData.cFileName);
        keyfile_fd = open (filename ,O_RDONLY ,0);
        if (keyfile_fd == -1) goto next_file;

        /*
         *    Read the keyfile record from the file. This will contain information which
         *    will identify the Test Server/Executer for which this keyfile was created.
         */
        do
        {
            status = read (keyfile_fd ,&srvinf_rec, sizeof (struct srvinf));
        }
        while ((status < 0) && (errno == EINTR));

        /*
         *    close the KEY file NOW so this file descriptor is not leaked.
         */
        closests = close (keyfile_fd);
        if (closests == -1)
        {
            iret = PSL_CLOSEIN;
            PSL_LOG_CIOERR (diagstk_ptr, iret, filename);
            goto end;
        }
        /*
         *    if we failed to read data from the KEY file...
         */
        if (status < 0)
        {           /* If we had an error reading   */
            iret = PSL_READERR; /* ...then report and return it */
            PSL_LOG_CIOERR (diagstk_ptr, PSL_READERR, filename);
            goto end;
        }
        else if (status > 0)
        {           /* ...else if we read a record  */

            /*
             *    We have read a record from a V-TEST Keyfile. From this we can recover
             *    much information relating to the Test Executer and should check to see
             *    firstly if the file is valid and secondly ensure that we have access to
             *    the Test Executer to which the file pertains.  Note that we assume a file
             *    with a PID of FFFFFFFF is valid, as this is used for debugging.
             */
            hProcess = OpenProcess(SYNCHRONIZE, FALSE, srvinf_rec.srvinf_pidl);
            /* ...check for process still exists */

            /*
             *    If the process to which this file pertains is no longer active then 
             *    delete the file and skip to the next one.
             */
            if ((hProcess == NULL)           /* ...if process not there */
                    && (srvinf_rec.srvinf_pidl != 0xFFFFFFFF))
                /* And not debug process */
            {
                PSL_CLEARERR();
                status = remove (filename);
                /* ...delete the keyfile    */
                if (status != 0)
                {               /* ...if we have an error   */
                    iret = PSL_FILNOTDEL;
                    /* ...then report and return it */
                    PSL_LOG_CIOERR (diagstk_ptr, iret ,filename);
                    goto end;
                }
            }
            else
            {
                if (hProcess != NULL) CloseHandle(hProcess);
                /* Close the process handle */

                /*
                 *    All o.k. - Add this keyfile record to the menu linked list.
                 */
                iret = PSL_VM_INSERT (srv_ll_ptr ,(char *) &srvinf_rec
                                      ,NULL ,NULL ,PSL_VM_POS_END);
                /* ...and it to end of LL   */
                if (!(psl_test_status (iret)))
                {               /* ...if we failed to add it    */
                    PSL_PUTDIAG (diagstk_ptr ,2 ,PSL_SYSERVERR
                                 ,"PSL_VM_INSERT (VTEST_LOCAL_SERVERS)"
                                 ,iret);
                    goto end;           /* ...signal the error and exit */
                }

            }                   /* ...end of PID check block    */
        }                   /* ...record read from keyfile  */
next_file:
        ;
        foundFile = FindNextFile (hFile, &FileData);
    }

    /*
     *    END
     *    ---
     *
     *    Close the search handle
     */
end:
    ;
    if (hFile != INVALID_HANDLE_VALUE) FindClose (hFile);


    /*
     *    ...and clean up the diagnostics stack if required.
     */
    if (psl_test_status (iret))     /* If routine successful      */
    {
        PSL_REMOVE_DIAGNOSTICS (diagstk_ptr ,saved_diagctx);
    }                   /* Remove diagnostics         */
    return iret;            /* ...and return with status      */
}


/************************************************************************/
/*                                  */
/*    VTEST_IS_DIR - Is the specified file a directory file     */
/*    ============                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = VTEST_IS_DIR (diagstk_ptr, filnam_ptr, flag_ptr);    */
/*                                  */
/*    Inputs:                               */
/*  diagstk_ptr    - Pointer to the diagnostic stack        */
/*  filnam_ptr  - Pointer to null terminated filename       */
/*                                  */
/*    Outputs:                              */
/*  flag_ptr    - Pointer to Boolean for file is{not} directory */
/*                                  */
/*    Return Status                             */
/*  PSL_NORMAL if no errors                     */
/*  error stat otherwise.                       */
/*                                  */
/*    This routine is called to determine whether the specified     */
/*    filespec is a directory or not.                   */
/*                                  */
/************************************************************************/
static unsigned long VTEST_IS_DIR   (struct diagstk *diagstk_ptr,
                                     char    *filnam_ptr,
                                     Boolean *flag_ptr)
{
    /*
     *    Local data
     */
    unsigned long   iret;           /* Return status */

    /*
     *  Assume success and file is not a directory.
     */
    iret = PSL_NORMAL;
    *flag_ptr = PFALSE;

    return iret;
}


/************************************************************************/
/*                                  */
/*    VTEST_CHK_SRVACC - Check current user has access to server    */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = VTEST_CHK_SRVACC (scope, groupid, userid);       */
/*                                  */
/*    Inputs:                               */
/*  scope       -   The scope to check against          */
/*  groupid     -   Group ID of current user            */
/*  userid      -   User ID of current user             */
/*                                  */
/*    Return Status                             */
/*  PSL_NORMAL if no errors                     */
/*  PSL_NOPRIV if user has no access to the runtime component   */
/*                                  */
/*    This routine is called to determine whether we may access the */
/*    specified V-TEST environment.                 */
/*                                  */
/*    This is determined by the scope of that environment and our UID.  */
/*                                  */
/*    The rules are:                            */
/*                                  */
/*    a) If the scope is WORLD then access is granted           */
/*                                  */
/*    b) If the scope is GROUP AND the user is in the same group as the */
/*   supervisor, OR is the superuser then access is granted.    */
/*                                  */
/*    c) If the scope is USER AND this user is the same OR he is the    */
/*   superuser access is granted.                   */
/*                                  */
/************************************************************************/

unsigned long VTEST_CHK_SRVACC (unsigned long scope,
                                unsigned long groupid,
                                unsigned long userid)
{
    /*
     *  Local Data,
     */

    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    VTEST_GET_USERDATA - Get the GID/UID for the specified user   */
/*    ==================                        */
/*                                  */
/*    Calling Format:                           */
/*  iret = VTEST_GET_USERDATA (diag_ptr                 */
/*                                ,username_ptr             */
/*                                ,gid_ptr              */
/*                                ,uid_ptr);                */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret         = PSL routine return status            */
/*             PSL_NORMAL if username found/details returned    */
/*             PSL_USRNOTFOU if username not found      */
/*                     other error from underlying routines     */
/*  diag_ptr     = Pointer to the diagnostics stack to use      */
/*  username_ptr = Pointer to null terminated user name.        */
/*  gid_ptr      = Pointer to long group ID returned        */
/*  uid_ptr      = Pointer to long process ID returned      */
/*                                  */
/*    This routine is called to translate the specified username and    */
/*    return the group and member ID for it. If the username does not   */
/*    exist on this node then a status of PSL_USRNOTFOU is returned */
/*                                  */
/*    Error Handling:                           */
/*  All errors are returned to the caller with appropriate diags    */
/*      added to the stack.                     */
/*                                  */
/************************************************************************/

unsigned long   VTEST_GET_USERDATA  (struct   diagstk *diag_ptr
                                     ,         char    *username_ptr
                                     ,         long    *gid_ptr
                                     ,         long    *uid_ptr)
{
    /*
     *    Local data
     */
    unsigned long    iret = PSL_NORMAL; /* PSL return status value  */

    return iret;            /* ...and return with status      */
}
#endif  /*** WIN32 ONLY MODULE ***/
