/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// GraphTypes.cpp: implementation of the CTimerGraph class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "GraphTypes.h"
#include "CECGlobalSettings.h"
#include "CYREventSink.h"
#include "CECVirtGridView.h"
#include "CECExcelExport.h"
#include "FilterURLDlg.h"
#include "SelectNodeDlg.h"
#include "ComDef.h"
#include "CustomOtaDlg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

extern CECGlobalSettings* g_pGlobalSettings;
extern CYREventSink* pEventSink;
//////////////////////////////////////////////////////////////////////
// CGraphTypeBase Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CGraphTypeBase::CGraphTypeBase(CGraphView *pView,const CString& strTitle)
{
    m_strTitle = strTitle;
    m_nNextGroup = 1;
    m_nDataRun = -1;

    m_pView = pView;
    m_pView->SetGraph(this);
}

CGraphTypeBase::~CGraphTypeBase()
{}

/////////////////////////////////////////////////////////////////////////////
// @mfunc   Initialize the title component of the graph

void CGraphTypeBase::SetTitleComponent()
{
    if(!g_pGlobalSettings->m_bShowTitle)
    {
        m_pView->m_ctlChart.SetTitleText("");
    }
    else
    {
        m_pView->m_ctlChart.SetTitleText(m_strTitle);
        m_pView->m_ctlChart.SetTitleTextColor(g_pGlobalSettings->m_TitleTextCol);
        m_pView->m_ctlChart.SetTitleFont(g_pGlobalSettings->m_TitleFont);
        m_pView->m_ctlChart.SetTitleBkColor(g_pGlobalSettings->m_TitleBGCol);
    }
    //  g_pGlobalSettings->m_TitleFontSize; //UNUSED
}

/////////////////////////////////////////////////////////////////////////////
// @mfunc   Initialize the display component

void CGraphTypeBase::SetDisplayComponent()
{
    // Set the grid and text colours/styles
    m_pView->m_ctlChart.SetPlotBkColor(g_pGlobalSettings->m_BGCol);
    m_pView->m_ctlChart.SetLabelFont(g_pGlobalSettings->m_LabelFont);
    m_pView->m_ctlChart.SetGridColor(g_pGlobalSettings->m_GridCol);
    m_pView->m_ctlChart.SetLabelColor(g_pGlobalSettings->m_LabelCol);
    m_pView->m_ctlChart.SetAxisTextColor(g_pGlobalSettings->m_ATCol);
}

void CGraphTypeBase::SetGraphType(int nType)
{
    for(int nGroup=1;nGroup<=m_nNextGroup;nGroup++)
        m_pView->m_ctlChart.SetSeriesStyle(nGroup,nType);
}

int CGraphTypeBase::GetGraphType()
{
    return m_pView->m_ctlChart.GetSeriesStyle(1);
}

/////////////////////////////////////////////////////////////////////////////
// @mfunc   Add a legend to the graph
void CGraphTypeBase::DisplayLegend()
{
    //  pLegend->GetStyle()->SetColor(RGB(192,192,192));
}


void CGraphTypeBase::SetDataColours()
{
    //
    // For each data group added to the graph...
    //
    COleSafeArray saData;
    saData.CreateOneDim(VT_I4,MAX_DEF_COLOURS,g_pGlobalSettings->m_nDefLineColours);
    m_pView->m_ctlChart.SetDataColors(saData);
}


BOOL CGraphTypeBase::RemoveTestRun(int nRunIndex)
{
    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);

    m_pView->m_ctlChart.RemoveSeries(strTestRun);
    m_nNextGroup--;
    return true;
}

BOOL CGraphTypeBase::ExportToExcel()
{
    CECExcelExport ExcelExport;
    return ExcelExport.SetGraph(&m_pView->m_ctlChart);
}

void CGraphTypeBase::GetMenuItems(MenuItemArray& arMenuItems)
{
    arMenuItems.RemoveAll();
}

void CGraphTypeBase::OnCommand(UINT uiCommandID,CWnd *pParent)
{}


void CGraphTypeBase::AddDataFromGraph(CGraphView *pView)
{
    VARIANT &var = pView->m_ctlChart.GetData();
    m_pView->m_ctlChart.SetData(var);

    COleSafeArray arData(var);
    long lNumGroups;
    arData.GetUBound(2,&lNumGroups);

    int nFirstGroup = m_nNextGroup;
    m_nNextGroup += lNumGroups+1;

    for(int nSeries=1;nSeries<=lNumGroups+1;nSeries++)
    {
        int nStyle = pView->m_ctlChart.GetSeriesStyle(nSeries);
        m_pView->m_ctlChart.SetSeriesStyle(nSeries+nFirstGroup-1,nStyle);

        CString str = pView->m_ctlChart.GetSeriesAnnotation(nSeries);
        m_pView->m_ctlChart.SetSeriesAnnotation(nSeries+nFirstGroup-1,str);

        BOOL b = pView->m_ctlChart.GetSeriesOnSecondaryAxis(nSeries);
       m_pView->m_ctlChart.SetSeriesOnSecondaryAxis(nSeries+nFirstGroup-1,b);
    }

    SetDataColours();
}

void CGraphTypeBase::SetValue(int nIndex,int nGroup,double dValue,const CString& strAnnotation)
{
}
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CTimerGraph::CTimerGraph(CGraphView *pView,const CString& strTitle):
        CGraphTypeBase(pView,strTitle)
{
    SetTitleComponent();
    SetDisplayComponent();
}

CTimerGraph::~CTimerGraph()
{
}


BOOL CTimerGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);

    TimerData *pTimerData = pGrid->GetTimerData(strTestRun);
    CArray <TimerInterval*,TimerInterval*>& arTimerIntervals = pTimerData->arTimerIntervals;

    DWORD dwNumIntervals = arTimerIntervals.GetSize();
    if(!dwNumIntervals)
        return false;

    COleSafeArray saData;
    DWORD dwNumElements[]= {dwNumIntervals , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    double dLastInterval=0;
    for(long nInterval=0;nInterval<(long)dwNumIntervals;nInterval++)
    {
        // Format the Label for the data value
        double dEnd = arTimerIntervals[nInterval]->dInterval;

        CString strLabel;
        strLabel.Format("%.1f - %.1f", dLastInterval, dEnd);
        dLastInterval = dEnd;

        // Get the data value off the grid and add to the graph
        pVar[nInterval+dwNumIntervals].vt = VT_R8;
        pVar[nInterval+dwNumIntervals].dblVal = GetDataPoint(arTimerIntervals[nInterval]);

        // Label the data value
        pVar[nInterval].vt = VT_BSTR;
        pVar[nInterval].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}

//////////////////////////////////////////////////////////////////////
// CCumFreqGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCumFreqGraph::CCumFreqGraph(CGraphView *pView) :
        CTimerGraph(pView,"Cumulative Frequency Distribution")
{
    // Set Axis/Graph styles for an area chart
    m_pView->m_ctlChart.SetXAxisLabel("Response Time (seconds)");
    m_pView->m_ctlChart.SetYAxisLabel("Frequency");
}

CCumFreqGraph::~CCumFreqGraph()
{
}


const double &CCumFreqGraph::GetDataPoint(TimerInterval* pInterval)
{
    return pInterval->dCumFreq;
}

//////////////////////////////////////////////////////////////////////
// CStdFreqGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CStdFreqGraph::CStdFreqGraph(CGraphView *pView):
        CTimerGraph(pView,"Standard Frequency Distribution")
{
    m_pView->m_ctlChart.SetXAxisLabel("Response Time (seconds)");
    m_pView->m_ctlChart.SetYAxisLabel("Frequency");
}

CStdFreqGraph::~CStdFreqGraph()
{
}


const double &CStdFreqGraph::GetDataPoint(TimerInterval* pInterval)
{
    return pInterval->dFrequency;
}

//////////////////////////////////////////////////////////////////////
// CPercentFreqGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPercentFreqGraph::CPercentFreqGraph(CGraphView *pView):
        CTimerGraph(pView,"Percentage Frequency Distribution")
{
    m_pView->m_ctlChart.SetXAxisLabel("Response Time (seconds)");
    m_pView->m_ctlChart.SetYAxisLabel("Frequency");
}

CPercentFreqGraph::~CPercentFreqGraph()
{
}


const double &CPercentFreqGraph::GetDataPoint(TimerInterval* pInterval)
{
    return pInterval->dPercentFreq;
}

//////////////////////////////////////////////////////////////////////
// CRTATGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRTATGraph::CRTATGraph(CGraphView *pView):
        CTimerGraph(pView,"Response Times / Active Threads")
{
    m_pView->m_ctlChart.SetXAxisLabel("No. Active Threads");
    m_pView->m_ctlChart.SetYAxisLabel("Response Time (seconds)");
}

CRTATGraph::~CRTATGraph()
{
}


BOOL CRTATGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    struct RTAT_DATA_t
    {
        int nNumTimers;
        double dTotalTime;
    }
    *pRtatData;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);

    TimerData *pTimerData = pGrid->GetTimerData(strTestRun);

    // Calculations for RTAT graph
    CMap <DWORD,DWORD,void*,void *> mapThreadsToData;
    CDWordArray arNumThreads;

    POSITION pos = pTimerData->listTimers.GetHeadPosition();
    while(pos)
    {
        Timer *pTimer = pTimerData->listTimers.GetNext(pos);
        if(!pTimer->bInFilter)
            continue;

        DWORD dwNumThreads = pTimer->dwActiveThreads;

        if(mapThreadsToData.Lookup(dwNumThreads,(void*&)pRtatData))
        {
            pRtatData->nNumTimers++;
            pRtatData->dTotalTime += pTimer->dValue;
        }
        else
        {
            pRtatData = new RTAT_DATA_t;
            pRtatData->nNumTimers = 1;
            pRtatData->dTotalTime = pTimer->dValue;
            mapThreadsToData.SetAt(dwNumThreads,pRtatData);

            for(int nNdx=0;nNdx<arNumThreads.GetSize();nNdx++)
                if(arNumThreads[nNdx]>dwNumThreads)
                    break;

            //  if(!nNdx)nNdx++;

            arNumThreads.InsertAt(nNdx/*-1*/,dwNumThreads);
        }
    }

    COleSafeArray saData;

    DWORD dwMaxNumThreads = 1;
	VARIANT *pVar = NULL;
	if(arNumThreads.GetSize())
		dwMaxNumThreads = arNumThreads[arNumThreads.GetSize()-1];

	DWORD dwNumElements[]= {dwMaxNumThreads, 2};
	saData.Create(VT_VARIANT,2,dwNumElements);

	saData.AccessData((void**)&pVar);
	pVar[0].vt = VT_EMPTY;

	CString strLabel;
	int nLastTerm = 1;
	for(int nNdx=0;nNdx<arNumThreads.GetSize();nNdx++)
	{
		DWORD dwNumThreads = arNumThreads[nNdx];
		mapThreadsToData.Lookup(dwNumThreads,(void*&)pRtatData);
		double dTime = pRtatData->dTotalTime / pRtatData->nNumTimers;

		// Add null data for all no of terminals between the last value and this one
		for (int nNdx2=nLastTerm;nNdx2<(int)dwNumThreads; nNdx2++)
		{
			VariantInit(&pVar[nNdx2-1+dwMaxNumThreads]);

			strLabel.Format("%d", nNdx2);
			pVar[nNdx2-1].vt = VT_BSTR;
			pVar[nNdx2-1].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
		}

		nLastTerm = dwNumThreads+1;

		// Add the data
		pVar[dwNumThreads-1+dwMaxNumThreads].vt = VT_R8;
		pVar[dwNumThreads-1+dwMaxNumThreads].dblVal = dTime;

		strLabel.Format("%d", (int)dwNumThreads);
		pVar[dwNumThreads-1].vt = VT_BSTR;
		pVar[dwNumThreads-1].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
		delete pRtatData;
	}

	saData.UnaccessData();
    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);
    return true;
}

CUsersTimeGraph::CUsersTimeGraph(CGraphView *pView):
        CTimerGraph(pView,"Active Users / Elapsed Time")
{
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Active Users");
}

CUsersTimeGraph::~CUsersTimeGraph()
{}



BOOL CUsersTimeGraph::AddTestRun(int nRunIndex)
{
    /*  m_nDataRun = nRunIndex;
     
        // Gain access to the grid for retrieval of data
        CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;
        
        // Get information about the data set to display
        CString strTestRun;
        g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
        
        TimerData *pTimerData = pGrid->GetTimerData(strTestRun);
        
        int nLastRow = pTimerData->listTimers.GetCount();
        if(!nLastRow)
            return true;
     
        DWORD dwStartTime = pTimerData->listTimers.GetHead()->dwTimeStamp;
        double dStartTime = pGrid->GetStartTime(nRunIndex);
        double dEndTime = pGrid->GetStopTime(nRunIndex);
     
        //For elapsed time graphs the data may need to be offset
        //so zero is the same for all graphs
        int nTimeOffset = dwStartTime - dStartTime;
     
        double dTimeStamp=0;
        int nTotalUsers=0;
        int nNumUsers=0;
        int nCurrentRow=0;
        int nElapsedTime =0;
        double dValue=0;
        int nCountAtTime=-1;
        CString strLabel;
        Timer *pTimer = NULL;
     
        POSITION pos = pTimerData->listTimers.GetHeadPosition();
        for(double dTime=dwStartTime;dTime<=dEndTime;dTime++)
        {
            while(dTimeStamp<=dTime)
            {
                nTotalUsers += nNumUsers;
                nCountAtTime++;
     
                if(pos)
                {
                    pTimer = pTimerData->listTimers.GetNext(pos);
                    while(pos && !pTimer->bInFilter)
                        pTimer = pTimerData->listTimers.GetNext(pos);
                }
                
                if(pos && pTimer)
                {               
                    dTimeStamp = pTimer->dwTimeStamp;
                    nNumUsers = pTimer->dwActiveThreads;
                }
                else
                    break;
            }
     
            if(nCountAtTime)
            {
                dValue = (double)nTotalUsers / nCountAtTime;
                nTotalUsers = 0;
                nCountAtTime=0;
     
                SetValue(nElapsedTime,m_nNextGroup, dValue);
            }
            else
            {
                SetValue(nElapsedTime,m_nNextGroup,0);
                SRGraphData *pData = GetSafeData(nElapsedTime, m_nNextGroup);
                pData->SetNull(TRUE);
            }
     
            nElapsedTime++;
            
            strLabel.Format("%d", nElapsedTime+nTimeOffset);
            GetSafeData(nElapsedTime-1,m_nNextGroup)->SetAnnotation(strLabel);
        }
     
     
        GetGroup(m_nNextGroup++)->SetAnnotation(strTestRun);
     
        SetDataColours();
    */
    return true;
}

CRTTimeGraph::CRTTimeGraph(CGraphView *pView):
        CTimerGraph(pView,"Response Time / Elapsed Time")
{
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Response Time (seconds)");
}

CRTTimeGraph::~CRTTimeGraph()
{}


BOOL CRTTimeGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    DWORD dwStartTime = pGrid->GetStartTime(nRunIndex);

    TimerData *pTimerData = pGrid->GetTimerData(strTestRun);

    int nLastRow = pTimerData->listTimers.GetCount();
    if(!nLastRow)
        return true;

    double dStartTime = pTimerData->listTimers.GetHead()->dwTimeStamp;
    double dEndTime = pTimerData->listTimers.GetTail()->dwTimeStamp;


    //For elapsed time graphs the data may need to be offset
    //so zero is the same for all graphs
    int nTimeOffset = (int)(dStartTime - dwStartTime);

    double dTimeStamp=0;
    double dTotalRespTime=0;
    double dRespTime=0;
    int nCurrentRow=0;
    int nElapsedTime =0;
    double dValue=0;
    int nCountAtTime=-1;
    CString strLabel;
    Timer *pTimer = NULL;

    COleSafeArray saData;

    int nNumSeconds = (int)(dEndTime-dStartTime+1);
    DWORD dwNumElements[]= {nNumSeconds, 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    POSITION pos = pTimerData->listTimers.GetHeadPosition();
    for(double dTime=dStartTime;dTime<=dEndTime;dTime++)
    {
        while(dTimeStamp<=dTime)
        {
            dTotalRespTime += dRespTime;
            nCountAtTime++;

            if(pos)
            {
                pTimer = pTimerData->listTimers.GetNext(pos);
                while(pos && !pTimer->bInFilter)
                    pTimer = pTimerData->listTimers.GetNext(pos);
            }

            if(pos && pTimer)
            {
                dTimeStamp = pTimer->dwTimeStamp;
                dRespTime = pTimer->dValue;
            }
            else
                break;
        }

        if(nCountAtTime)
        {
            dValue = dTotalRespTime / nCountAtTime;
            dTotalRespTime = 0;
            nCountAtTime=0;

            pVar[nElapsedTime+nNumSeconds].vt = VT_R8;
            pVar[nElapsedTime+nNumSeconds].dblVal = dValue;
        }
        else
        {
            pVar[nElapsedTime+nNumSeconds].vt = VT_EMPTY;
        }

        nElapsedTime++;

        strLabel.Format("%d", nElapsedTime+nTimeOffset);
        pVar[nElapsedTime-1].vt = VT_BSTR;
        pVar[nElapsedTime-1].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();
    return true;
}

//////////////////////////////////////////////////////////////////////
// COTAGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COTAGraph::COTAGraph(CGraphView *pView,const CString& strTitle):
        CGraphTypeBase(pView,strTitle)
{
    m_nNumTimers=0;
    m_nNextGroup=1;
    SetTitleComponent();
    SetDisplayComponent();

    //  DisplayLegend();
}


COTAGraph::COTAGraph(const CString& strQueryName,CGraphView *pView):
        CGraphTypeBase(pView,strQueryName)
{
    m_nNextGroup=1;
    m_strQueryName = strQueryName;

    SetTitleComponent();
    SetDisplayComponent();

    //    DisplayLegend();

    // Set Axis/Graph styles for an area chart
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel(m_strQueryName);
}


COTAGraph::~COTAGraph()
{
}


BOOL COTAGraph::AddQuery(int nRunIndex,const CString& strQueryName,const CString& strNode)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasOTA())
        return true;

    OtaQueryArray *parOtaData = pGrid->GetOtaData(strTestRun);
    if(parOtaData == NULL)
        return false;

    CString strLabel;
    //Search for the desired query
    for(int nNdx=0;nNdx<parOtaData->GetSize();nNdx++)
    {
        OtaQuery *pQuery = parOtaData->GetAt(nNdx);
        if((pQuery->strQuery != strQueryName) || (strNode != pQuery->strNode))
            continue;

        OtaDataList &OtaData = pQuery->listQueryData;

		DWORD dwFirstTime = 0;
        DWORD dwLastTime = 0;

        COleSafeArray saData;

        if(OtaData.GetCount())
        {
	        dwFirstTime = pGrid->GetStartTime(nRunIndex);
			dwLastTime = OtaData.GetTail()->dwStartTime;
		}

        DWORD dwNextTime = dwFirstTime;

        DWORD dwDuration = dwLastTime-dwFirstTime+1;
        DWORD dwNumElements[]= {dwDuration , 2};
        saData.Create(VT_VARIANT,2,dwNumElements);

        VARIANT *pVar;
        saData.AccessData((void**)&pVar);

        DWORD dwTime;
        POSITION pos = OtaData.GetHeadPosition();
		if(!pos)
		{
            pVar[0].vt = VT_BSTR;
            pVar[0].bstrVal = _com_util::ConvertStringToBSTR("");
            pVar[1].vt = VT_R8;
            pVar[1].bstrVal = 0;
		}

        while(pos)
        {
            OtaQueryData *pData = OtaData.GetNext(pos);

            for(;dwNextTime<pData->dwStartTime;dwNextTime++)
            {
                dwTime = dwNextTime-dwFirstTime;
                VariantInit(&pVar[dwTime+dwDuration]);

                strLabel.Format("%d",dwTime);
                pVar[dwTime].vt = VT_BSTR;
                pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            }

            dwTime = dwNextTime-dwFirstTime;
            pVar[dwTime+dwDuration].vt = VT_R8;
            pVar[dwTime+dwDuration].dblVal = pData->nValue;

            strLabel.Format("%d",dwTime);
            pVar[dwTime].vt = VT_BSTR;
            pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            dwNextTime++;
        }

        saData.UnaccessData();

        m_pView->m_ctlChart.SetData(saData);
        m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);
		m_pView->m_ctlChart.SetLegendText(m_nNextGroup++,strQueryName + " (" + strNode + ")");
        m_pView->m_ctlChart.ShowLegend(TRUE);
        return true;
    }

    return false;
}

BOOL COTAGraph::AddTimer(int nRunIndex,const CString& strTimer)
{
    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasTimers())
        return true;

    TimerData *pTimerData = pGrid->GetTimerData(strTestRun);
    if(pTimerData == NULL)
        return false;

    DWORD dwStartTime = pGrid->GetStartTime(nRunIndex);
    DWORD dwLastTime = pGrid->GetStopTime(nRunIndex);
    DWORD dwNextTime = dwStartTime;

    COleSafeArray saData;
    DWORD dwDuration = dwLastTime-dwStartTime+1;
    DWORD dwNumElements[]= {dwDuration , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    int nCount=0;
    double dTotalTime = 0;
    DWORD dwPrevTime=0;

    DWORD dwTime;
    CString strLabel;
    POSITION pos = pTimerData->listTimers.GetHeadPosition();
    while(pos)
    {
        Timer *pTimer = pTimerData->listTimers.GetNext(pos);
        if(pTimer->bInFilter && (pTimer->strTimerName == strTimer))
        {

            if(!dwPrevTime || (dwPrevTime == pTimer->dwTimeStamp))
            {
                nCount++;
                dTotalTime += pTimer->dValue;
            }
            else
            {
                for(;dwNextTime<dwPrevTime;dwNextTime++)
                {
                    dwTime = dwNextTime-dwStartTime;
                    VariantInit(&pVar[dwTime+dwDuration]);

                    strLabel.Format("%d",dwTime);
                    pVar[dwTime].vt = VT_BSTR;
                    pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
                }

                dwTime = dwNextTime-dwStartTime;
                pVar[dwTime+dwDuration].vt = VT_R8;
                pVar[dwTime+dwDuration].dblVal = dTotalTime / nCount;

                strLabel.Format("%d",dwTime);
                pVar[dwTime].vt = VT_BSTR;
                pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);

                dTotalTime = pTimer->dValue;
                nCount=1;
                dwNextTime++;
            }
            dwPrevTime = pTimer->dwTimeStamp;
        }

        if(!pos)
        {
            for(;dwNextTime<dwPrevTime;dwNextTime++)
            {
                dwTime = dwNextTime-dwStartTime;
                VariantInit(&pVar[dwTime+dwDuration]);

                strLabel.Format("%d",dwTime);
                pVar[dwTime].vt = VT_BSTR;
                pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            }

            dwTime = dwNextTime-dwStartTime;
            pVar[dwTime+dwDuration].vt = VT_R8;
            pVar[dwTime+dwDuration].dblVal = dTotalTime / nCount;

            strLabel.Format("%d",dwTime);
            pVar[dwTime].vt = VT_BSTR;
            pVar[dwTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
        }
    }

    saData.UnaccessData();
    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);
    m_pView->m_ctlChart.SetSeriesOnSecondaryAxis(m_nNextGroup,true);
    m_pView->m_ctlChart.SetLegendText(m_nNextGroup++,"TIMER:" + strTimer);

    m_nNumTimers++;
    return true;
}

BOOL COTAGraph::AddTestRun(int nRunIndex,const CString& strQueryName)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasOTA())
        return true;

    DWORD dwStartTime = pGrid->GetStartTime(nRunIndex);
    DWORD dwEndTime = pGrid->GetStopTime(nRunIndex);

    OtaQueryArray *parOtaData = pGrid->GetOtaData(strTestRun);
    if(parOtaData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }

    if(m_strCurrentNode.IsEmpty())
        m_strCurrentNode = parOtaData->GetAt(0)->strNode;

    // Find out how many queries we have
    CString strQName,strNode;
    for(int nNdx=0;nNdx<parOtaData->GetSize();nNdx++)
    {
        OtaQuery *pQuery = parOtaData->GetAt(nNdx);

        strNode = pQuery->strNode;
        strQName = pQuery->strQuery;

        if((strQName == strQueryName) && (m_strCurrentNode == strNode))
        {
            int nStartGroup = m_nNextGroup;

            DWORD dwTime=dwStartTime,dwVal=0;

            OtaDataList &OtaData = pQuery->listQueryData;
  
			if(OtaData.GetCount())
            {
	            dwTime = OtaData.GetHead()->dwStartTime;
		        dwVal = OtaData.GetHead()->nValue;
			}

            int nTimeOffset = dwTime - dwStartTime;

            DWORD dwMinTime = dwTime;
            DWORD dwPrevTime=dwStartTime;

            COleSafeArray saData;
            DWORD dwNumPoints = dwEndTime-dwStartTime+1;
            DWORD dwNumElements[]= {dwNumPoints , 2};
            saData.Create(VT_VARIANT,2,dwNumElements);

            VARIANT *pVar;
            saData.AccessData((void**)&pVar);

            CString strLabel;
            POSITION pos = OtaData.GetHeadPosition();
            BOOL bContinue = pos ? TRUE : FALSE;
            while(bContinue)
            {
                if(!pos)
                    bContinue = false;

                for(;dwPrevTime<dwTime;dwPrevTime++)
                {
                    VariantInit(&pVar[dwPrevTime-dwMinTime+nTimeOffset+dwNumPoints]);

                    strLabel.Format("%d", (int)(dwPrevTime-dwMinTime+nTimeOffset));
                    pVar[dwPrevTime-dwMinTime+nTimeOffset].vt = VT_BSTR;
                    pVar[dwPrevTime-dwMinTime+nTimeOffset].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
                }

                // Add the data
                pVar[dwTime-dwMinTime+nTimeOffset+dwNumPoints].vt = VT_R8;
                pVar[dwTime-dwMinTime+nTimeOffset+dwNumPoints].dblVal = dwVal;

                strLabel.Format("%d", (int)(dwTime-dwMinTime+nTimeOffset));
                pVar[dwTime-dwMinTime+nTimeOffset].vt = VT_BSTR;
                pVar[dwTime-dwMinTime+nTimeOffset].bstrVal = _com_util::ConvertStringToBSTR(strLabel);

                dwPrevTime = dwTime+1;

                if(pos)
                {
                    OtaQueryData *pQuery = OtaData.GetNext(pos);
                    dwTime = pQuery->dwStartTime;
                    dwVal = pQuery->nValue;
                }
            }

            saData.UnaccessData();

            m_pView->m_ctlChart.SetData(saData);
            m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

            // Set the group heading label
            m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);
        }
    }

    SetDataColours();

    return true;
}

void COTAGraph::GetMenuItems(MenuItemArray& arMenuItems)
{
    arMenuItems.RemoveAll();

    if(GetNumInjectors() >= 2)
    {
        MenuItems MI;

        MI.strItem = "Select Injector";
        MI.uiCommandID = enMenuItemNode;
        arMenuItems.Add(MI);
    }
}

int COTAGraph::GetGraphType()
{
    if(m_nNextGroup-1>m_nNumTimers)
        return m_pView->m_ctlChart.GetSeriesStyle(m_nNumTimers+1);
    else
        return CGraphTypeBase::GetGraphType();
}

int COTAGraph::GetNumInjectors()
{
    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(m_nDataRun, strTestRun);

    OtaQueryArray *parOtaData = pGrid->GetOtaData(strTestRun);

    CMap <CString,LPCTSTR,bool,bool> m_mapInjectors;
    for(int nNdx=0;nNdx<parOtaData->GetSize();nNdx++)
    {
        m_mapInjectors.SetAt(parOtaData->GetAt(nNdx)->strNode,false);
    }
    return m_mapInjectors.GetCount();
}

void COTAGraph::OnCommand(UINT uiCommandID,CWnd *pParent)
{
    switch(uiCommandID)
    {
    case enMenuItemNode:
        {
            CStringArray arstrNodes;

            // Gain access to the grid for retrieval of data
            CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

            // Get information about the data set to display
            CString strTestRun;
            g_pGlobalSettings->GetRunInfo(m_nDataRun, strTestRun);

            OtaQueryArray *parOtaData = pGrid->GetOtaData(strTestRun);

            int nQueryNdx=0;
            for(int nNdx=0;nNdx<parOtaData->GetSize();nNdx++)
                arstrNodes.Add(  parOtaData->GetAt(nNdx)->strNode  );

            //Put up select node dlg
            CSelectNodeDlg dlg(arstrNodes);
            dlg.m_strSelectedNode = m_strCurrentNode;
            if(dlg.DoModal()==IDOK)
            {
                m_strCurrentNode = dlg.m_strSelectedNode;
                //reload data
                RemoveTestRun(m_nDataRun);
                AddTestRun(m_nDataRun);
                ::PostMessage(pParent->GetSafeHwnd(),WMU_DATA_CHANGED,0,m_nDataRun);
            }
        }
        break;
    }
}

//////////////////////////////////////////////////////////////////////
// CPercentCPUGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPercentCPUGraph::CPercentCPUGraph(CGraphView *pView):
        COTAGraph(pView,"Percent of CPU")
{
    m_strQueryName = strPercentCPUHeading;

    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Percent of CPU");
}

CPercentCPUGraph::~CPercentCPUGraph()
{
}

//////////////////////////////////////////////////////////////////////
// CNetUseGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CNetUseGraph::CNetUseGraph(CGraphView *pView):
        COTAGraph(pView,"Percent Network Utilisation")
{
    m_strQueryName = strPercentNetUseHeading;

    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Percent Of Network Usage");
}

CNetUseGraph::~CNetUseGraph()
{
}

//////////////////////////////////////////////////////////////////////
// CPageSwapGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CPageSwapGraph::CPageSwapGraph(CGraphView *pView):
        COTAGraph(pView,"Page Swapping")
{
    m_strQueryName = strPageSwappingHeading;

    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Page Swapping per Second");
}

CPageSwapGraph::~CPageSwapGraph()
{
}

//////////////////////////////////////////////////////////////////////
// CCustomOTAGraph Class
////////////////////////////
///////////////////////////////////////////////////////////////////////
/////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCustomOTAGraph::CCustomOTAGraph(CGraphView *pView):
        COTAGraph(pView,"Custom Performance Monitor Graph")
{
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
}

CCustomOTAGraph::~CCustomOTAGraph()
{}


BOOL CCustomOTAGraph::AddTestRun(int nRun)
{
    m_nDataRun = nRun;
    CString strNode,strQuery;

    BOOL bAddRun = false;

    if(m_arstrTimers.GetSize() || m_arstrQueries.GetSize())
    {
        bAddRun = true;
    }
    else
    {
        CCustomOTADlg dlg(nRun);
        bAddRun = (dlg.DoModal()==IDOK);

        if(!bAddRun)
            return false;

        for(int n=0;n < dlg.m_arstrTimers.GetSize();n++)
            m_arstrTimers.Add(dlg.m_arstrTimers[n]);

        for(n=0;n < dlg.m_arstrQueries.GetSize();n++)
            m_arstrQueries.Add(dlg.m_arstrQueries[n]);
    }

    if(bAddRun)
    {
        ClearData();

        int nNumTimers = m_arstrTimers.GetSize();
        int nNumQueries = m_arstrQueries.GetSize();

        for(int nTimer=0;nTimer<nNumTimers;nTimer++)
            AddTimer(nRun,m_arstrTimers[nTimer]);

        for(int nQuery=0;nQuery<nNumQueries;nQuery++)
        {
            CString &strQueryNode = m_arstrQueries[nQuery];

            int nPos = strQueryNode.ReverseFind('(');
            if(nPos == -1)
                continue;

            strQuery = strQueryNode.Left(nPos-1);
            strNode = strQueryNode.Mid(nPos+1,strQueryNode.GetLength()-nPos-2);

            AddQuery(nRun,strQuery,strNode);
        }

        if(nNumTimers && !nNumQueries)
        {
            m_pView->m_ctlChart.SetYAxisLabel("Response Time");
            m_pView->m_ctlChart.SetY2AxisLabel("Response Time");
        }
        else
        {
            m_pView->m_ctlChart.SetYAxisLabel("Query Value");
            m_pView->m_ctlChart.SetY2AxisLabel("Response Time");
        }

        SetDataColours();
        SetGraphType(VtChSeriesType2dLine);
        return true;
    }
    else
        return false;
}

void CCustomOTAGraph::ClearData()
{
    COleSafeArray ar;
    DWORD dwNumElements[]= {0,0};
    ar.Create(VT_VARIANT,2,dwNumElements);

    m_pView->m_ctlChart.SetData(ar);

    m_nNextGroup=1;
    m_nNumTimers=0;
}

void CCustomOTAGraph::GetMenuItems(MenuItemArray& arMenuItems)
{
    arMenuItems.RemoveAll();

    MenuItems MI;

    if(GetNumInjectors() >= 2)
    {
        MI.strItem = "Select Injector";
        MI.uiCommandID = enMenuItemNode;

        arMenuItems.Add(MI);
    }

    MI.strItem = "Select &Queries";
    MI.uiCommandID = enMenuItemCustomise;

    arMenuItems.Add(MI);
}


void CCustomOTAGraph::OnCommand(UINT uiCommandID,CWnd *pParent)
{
    switch(uiCommandID)
    {
    case enMenuItemCustomise:
        {
			m_arstrTimers.RemoveAll();
			m_arstrQueries.RemoveAll();
            AddTestRun(m_nDataRun);
            if(pParent)
                pParent->Invalidate();
        }
        break;
    default:
        COTAGraph::OnCommand(uiCommandID,pParent);
        break;

    }
}


void CCustomOTAGraph::SetGraphType(int nType)
{
    if(m_nNextGroup-1>m_nNumTimers)
    {
        //If there are ota stats and timers only change ota stat type...
        for(int nGroup=m_nNumTimers+1;nGroup<m_nNextGroup;nGroup++)
            m_pView->m_ctlChart.SetSeriesStyle(nGroup,nType);
    }
    else
    {
        //otherwise just change timers
        for(int nGroup=1;nGroup<m_nNextGroup;nGroup++)
            m_pView->m_ctlChart.SetSeriesStyle(nGroup,nType);
    }
}

//////////////////////////////////////////////////////////////////////
// CHttpGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHttpGraph::CHttpGraph(CGraphView* pView,const CString& strTitle):
        CGraphTypeBase(pView,strTitle)
{
    m_bHasURLFilter = false;
    SetTitleComponent();
    SetDisplayComponent();
}

CHttpGraph::~CHttpGraph()
{}


BOOL CHttpGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;
    return false;
}

void CHttpGraph::GetMenuItems(MenuItemArray& arMenuItems)
{
    arMenuItems.RemoveAll();

    MenuItems MI;

    if(m_bHasURLFilter)
    {
        MI.strItem = "Filter URLs";
        MI.uiCommandID = enMenuItemFilter;
        arMenuItems.Add(MI);
    }
}

void CHttpGraph::OnCommand(UINT uiCommandID,CWnd *pParent)
{
    switch(uiCommandID)
    {
    case enMenuItemFilter:
        {
            CStringArray arstrURLs;

            // Gain access to the grid for retrieval of data
            CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

            // Get information about the data set to display
            CString strTestRun;
            g_pGlobalSettings->GetRunInfo(m_nDataRun, strTestRun);

            pGrid->GetURLs(strTestRun,arstrURLs);

            //Put up filter dialog
            //(should only show this if there is only one group in this graph)
            CFilterURLDlg dlg(arstrURLs,m_ardwURLFilters[0]);
            if(dlg.DoModal()==IDOK)
            {
                //reload data
                RemoveTestRun(m_nDataRun);
                AddTestRun(m_nDataRun);
                ::PostMessage(pParent->GetSafeHwnd(),WMU_DATA_CHANGED,0,m_nDataRun);
            }
        }
        break;
    }
}

BOOL CHttpGraph::IsURLInFilter(int nURLNdx,int nGroup)
{
    if(!m_ardwURLFilters[nGroup].GetSize())
        return true;

    for(int nURL=0;nURL<m_ardwURLFilters[nGroup].GetSize();nURL++)
    {
        if(nURLNdx == (int)m_ardwURLFilters[nGroup][nURL])
            return true;
    }

    return false;
}
//////////////////////////////////////////////////////////////////////
// CThroughputGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CThroughputGraph::CThroughputGraph(CGraphView* pView):
        CHttpGraph(pView,"Total Bytes Per Second")
{
    m_bHasURLFilter = true;
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Total Bytes / Second");
}

CThroughputGraph::~CThroughputGraph()
{
}


BOOL CThroughputGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }
    HttpDataArray &arHttpData = *parHttpData;

    int nNumStats = arHttpData.GetSize();
    if(!nNumStats)
        return true;

    DWORD dwStartTime = arHttpData[0]->dwTimeStamp;
    DWORD dwEndTime = arHttpData[nNumStats-1]->dwTimeStamp;

    int nTimeOffset = dwStartTime - pGrid->GetStartTime(nRunIndex);

    COleSafeArray saData;

    DWORD dwNumPoints = dwEndTime-dwStartTime+1;
    DWORD dwNumElements[]= {dwNumPoints , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    CString strLabel;
    int nElapsedTime=0;
    DWORD dwTotalBytes=0;
    DWORD dwCurrentTime=0;
    int nCurrentStat=0;

    DWORD dwTime = dwStartTime;

    while(1)
    {
        for(;nCurrentStat<nNumStats;nCurrentStat++)
            if(IsURLInFilter( arHttpData[nCurrentStat]->nURLNdx ,m_nNextGroup-1))
                break;

        if(nCurrentStat<nNumStats)
            dwCurrentTime = arHttpData[nCurrentStat]->dwTimeStamp;
        else
            dwCurrentTime = 0;

        if(dwTime == dwCurrentTime)
        {
            dwTotalBytes += arHttpData[nCurrentStat]->dwReplySize;
        }
        else
        {
            nElapsedTime = dwTime-dwStartTime;

            pVar[nElapsedTime + dwNumPoints].vt = VT_R8;
            pVar[nElapsedTime + dwNumPoints].dblVal = dwTotalBytes;

            strLabel.Format("%d", nElapsedTime+nTimeOffset);
            pVar[nElapsedTime].vt = VT_BSTR;
            pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);

            if(nCurrentStat>=nNumStats)
                break;

            dwTotalBytes = arHttpData[nCurrentStat]->dwReplySize;

            //Fill in any empty values
            while(++dwTime<dwCurrentTime)
            {
                nElapsedTime = dwTime-dwStartTime;

                pVar[nElapsedTime + dwNumPoints].vt = VT_EMPTY;
                pVar[nElapsedTime + dwNumPoints].dblVal = 0;

                strLabel.Format("%d", nElapsedTime+nTimeOffset);
                pVar[nElapsedTime].vt = VT_BSTR;
                pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            }
        }
        nCurrentStat++;
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();
    return true;
}


//////////////////////////////////////////////////////////////////////
// CHitsPerSecondGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHitsPerSecondGraph::CHitsPerSecondGraph(CGraphView* pView):
        CHttpGraph(pView,"Total Hits Per Second")
{
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Total Hits / Second");
}

CHitsPerSecondGraph::~CHitsPerSecondGraph()
{
}


BOOL CHitsPerSecondGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }

    HttpDataArray &arHttpData = *parHttpData;
    int nNumStats = arHttpData.GetSize();
    if(!nNumStats)
        return true;

    DWORD dwStartTime = arHttpData[0]->dwTimeStamp;
    DWORD dwEndTime= arHttpData[nNumStats-1]->dwTimeStamp;

    COleSafeArray saData;

    DWORD dwNumPoints = dwEndTime-dwStartTime+1;
    DWORD dwNumElements[]= {dwNumPoints ,2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    int nTimeOffset = dwStartTime - pGrid->GetStartTime(nRunIndex);

    CString strLabel;
    int nElapsedTime=0;
    DWORD dwCount = 0;
    DWORD dwCurrentTime=0;
    int nCurrentStat=0;

    DWORD dwTime = dwStartTime;

    while(1)
    {
        for(;nCurrentStat<nNumStats;nCurrentStat++)
            if(IsURLInFilter( arHttpData[nCurrentStat]->nURLNdx ,m_nNextGroup-1))
                break;

        if(nCurrentStat<nNumStats)
            dwCurrentTime = arHttpData[nCurrentStat]->dwTimeStamp;
        else
            dwCurrentTime = 0;

        if(dwTime == dwCurrentTime)
        {
            dwCount++;
        }
        else
        {
            nElapsedTime = dwTime-dwStartTime;

            pVar[nElapsedTime + dwNumPoints].vt = VT_R8;
            pVar[nElapsedTime + dwNumPoints].dblVal = dwCount;

            strLabel.Format("%d", nElapsedTime+nTimeOffset);
            pVar[nElapsedTime].vt = VT_BSTR;
            pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);

            if(nCurrentStat>=nNumStats)
                break;

            dwCount=1;

            //Fill in any empty values
            while(++dwTime<dwCurrentTime)
            {
                nElapsedTime = dwTime-dwStartTime;

                pVar[nElapsedTime + dwNumPoints].vt = VT_EMPTY;

                strLabel.Format("%d", nElapsedTime+nTimeOffset);
                pVar[nElapsedTime].vt = VT_BSTR;
                pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            }
        }
        nCurrentStat++;
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);
    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}


//////////////////////////////////////////////////////////////////////
// CAvPageRTPerUserGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAvPageRTPerUserGraph::CAvPageRTPerUserGraph(CGraphView* pView):
        CHttpGraph(pView,"Average Hit Response Time / Number Of Hits")
{
    m_bHasURLFilter = true;
    m_pView->m_ctlChart.SetXAxisLabel("Number Of Hits");
    m_pView->m_ctlChart.SetYAxisLabel("Average Hit Response Time");
}

CAvPageRTPerUserGraph::~CAvPageRTPerUserGraph()
{
}


BOOL CAvPageRTPerUserGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }
    HttpDataArray &arHttpData = *parHttpData;

    typedef struct
    {
        int nCount;     //Number of occurences of x VUs
        double dTotalTime;  //Total Response time for x VUs
    }

    RespTime;

    CMap <int,int&,LPVOID,LPVOID> mapVUsToRespTime;

    int nNumVUs=0;
    int nMaxVUs = 0;
    DWORD dwLastTimeStamp = 0;
    DWORD dwTimeStamp = 0;
    DWORD dwRespTime=0;
    double dAvRespTime=0;
    int nNumIncludedVUs=0;

    bool bURLFound=true;

    RespTime *pRespTime;

    int nLastRow = arHttpData.GetSize();

    for(int nRow=0;nRow<=nLastRow;nRow++)
    {
        if(nRow != nLastRow)
        {
            //Apply URL Filter
            if(m_ardwURLFilters[m_nNextGroup-1].GetSize())
            {
                int nURLNdx = arHttpData[nRow]->nURLNdx;

                bURLFound=false;
                for(int nURL=0;nURL<m_ardwURLFilters[m_nNextGroup-1].GetSize()&&!bURLFound;nURL++)
                    bURLFound = (nURLNdx == (int)m_ardwURLFilters[m_nNextGroup-1][nURL]);
            }

            dwLastTimeStamp = dwTimeStamp;

            dwTimeStamp = arHttpData[nRow]->dwTimeStamp;
            if(bURLFound)
                dwRespTime = arHttpData[nRow]->dwResponseTime;
            else
                dwRespTime=0;

            if(!dwLastTimeStamp || (dwTimeStamp == dwLastTimeStamp))
            {
                nNumVUs++;
                dAvRespTime += dwRespTime;
                if(bURLFound)
                    nNumIncludedVUs++;
                continue;
            }
        }

        ASSERT(nNumIncludedVUs || (!nNumIncludedVUs && !dAvRespTime));
        //Get average for this occurence of nNumVUs
        if(nNumIncludedVUs)
        {
            dAvRespTime /= nNumIncludedVUs;

            if(mapVUsToRespTime.Lookup(nNumVUs,(LPVOID&)pRespTime))
            {
                pRespTime->nCount++;
                pRespTime->dTotalTime += dAvRespTime;
            }
            else
            {
                pRespTime = new RespTime;
                pRespTime->nCount=1;
                pRespTime->dTotalTime = dAvRespTime;
                mapVUsToRespTime.SetAt(nNumVUs,pRespTime);
            }
        }

        if(nMaxVUs < nNumVUs) nMaxVUs = nNumVUs;
        nNumVUs=1;
        nNumIncludedVUs= (bURLFound ? 1 : 0);

        dAvRespTime = dwRespTime;
    }

    COleSafeArray saData;

    DWORD dwNumPoints = nMaxVUs;
    DWORD dwNumElements[]= {dwNumPoints , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    //Put values into the chart
    CString strLabel;
    for(nNumVUs = 1;nNumVUs<=nMaxVUs;nNumVUs++)
    {
        if(mapVUsToRespTime.Lookup(nNumVUs,(LPVOID&)pRespTime))
        {
            dAvRespTime = (pRespTime->dTotalTime / pRespTime->nCount);
            delete pRespTime;
        }
        else
        {
            dAvRespTime = 0;
            pRespTime=NULL;
        }

        pVar[nNumVUs-1 + dwNumPoints].vt = VT_R8;
        pVar[nNumVUs-1 + dwNumPoints].dblVal = dAvRespTime;

        strLabel.Format("%d", nNumVUs);
        pVar[nNumVUs-1].vt = VT_BSTR;
        pVar[nNumVUs-1].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}

//////////////////////////////////////////////////////////////////////
// CErrorsPerUserGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CErrorsPerUserGraph::CErrorsPerUserGraph(CGraphView* pView):
        CHttpGraph(pView,"Errors / Active Connections")
{
    m_bHasURLFilter = true;
    m_pView->m_ctlChart.SetXAxisLabel("Active Connections");
    m_pView->m_ctlChart.SetYAxisLabel("Errors");
}

CErrorsPerUserGraph::~CErrorsPerUserGraph()
{
}


BOOL CErrorsPerUserGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }
    HttpDataArray &arHttpData = *parHttpData;

    CMap <int,int&,int,int&> mapVUsToErrors;

    int nNumErrors=0;
    int nNumVUs=0;
    int nMaxVUs = 0;
    WORD wResponseCode=0;
    DWORD dwLastTimeStamp = 0;
    DWORD dwTimeStamp = 0;
    bool bURLFound=true;
    CString strURL;

    int nLastRow = arHttpData.GetSize();

    for(int nRow=0;nRow<=nLastRow;nRow++)
    {
        if(nRow!=nLastRow)
        {
            //Apply URL Filter
            if(m_ardwURLFilters[m_nNextGroup-1].GetSize())
            {
                int nURLNdx = arHttpData[nRow]->nURLNdx;

                bURLFound=false;
                for(int nURL=0;nURL<m_ardwURLFilters[m_nNextGroup-1].GetSize()&&!bURLFound;nURL++)
                    bURLFound = (nURLNdx == (int)m_ardwURLFilters[m_nNextGroup-1][nURL]);
            }

            dwLastTimeStamp = dwTimeStamp;

            dwTimeStamp = arHttpData[nRow]->dwTimeStamp;
            if(bURLFound)
                wResponseCode = arHttpData[nRow]->wResponseCode;
            else
                wResponseCode=0;

            if(!dwLastTimeStamp || (dwTimeStamp == dwLastTimeStamp))
            {
                nNumVUs++;

                //Only an error if above 400
                if(wResponseCode>=400)
                    nNumErrors++;

                continue;
            }
        }

        int nNumPrevErrors=0;
        if(mapVUsToErrors.Lookup(nNumVUs,nNumPrevErrors))
            nNumErrors+=nNumPrevErrors;

        mapVUsToErrors.SetAt(nNumVUs,nNumErrors);

        if(nMaxVUs < nNumVUs) nMaxVUs = nNumVUs;
        nNumVUs=1;

        //Only an error if above 400
        if(wResponseCode>=400)
            nNumErrors=1;
        else
            nNumErrors=0;
    }

    COleSafeArray saData;

    DWORD dwNumPoints = nMaxVUs;
    DWORD dwNumElements[]= {dwNumPoints  , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    //Put values into the chart
    CString strLabel;
    for(nNumVUs = 1;nNumVUs<=nMaxVUs;nNumVUs++)
    {
        if(!mapVUsToErrors.Lookup(nNumVUs,nNumErrors))
            nNumErrors=0;

        pVar[nNumVUs-1 + dwNumPoints].vt = VT_R8;
        pVar[nNumVUs-1 + dwNumPoints].dblVal = nNumErrors;

        strLabel.Format("%d", nNumVUs);
        pVar[nNumVUs-1].vt = VT_BSTR;
        pVar[nNumVUs-1].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}

//////////////////////////////////////////////////////////////////////
// CErrorsTimeGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CErrorsTimeGraph::CErrorsTimeGraph(CGraphView* pView):
        CHttpGraph(pView,"Errors / Elapsed Time")
{
    m_bHasURLFilter = true;
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Errors");
}

CErrorsTimeGraph::~CErrorsTimeGraph()
{
}


BOOL CErrorsTimeGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }
    HttpDataArray &arHttpData = *parHttpData;

    int nLastRow = arHttpData.GetSize()-1;



    DWORD dwStartTime = arHttpData[0]->dwTimeStamp;
    DWORD dwEndTime = arHttpData[nLastRow]->dwTimeStamp;

    int nTimeOffset = dwStartTime - pGrid->GetStartTime(nRunIndex);

    DWORD dwTimeStamp=0;
    CString strLabel;
    int nElapsedTime=0;
    WORD wResponseCode=0;
    int nTotalErrors=0;
    CString strURL;
    bool bFound=false;

    COleSafeArray saData;

    DWORD dwNumPoints = dwEndTime-dwStartTime+1;
    DWORD dwNumElements[]= {dwNumPoints  , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    int nCurrentRow=0;
    for(int nTime=dwStartTime;nTime<=(int)dwEndTime;nTime++)
    {
        while((dwTimeStamp<=(DWORD)nTime) && (nCurrentRow<=nLastRow+1))
        {
            //Only an error if above 400
            if(wResponseCode>=400)
            {
                nTotalErrors++;
                wResponseCode =0;
            }

            if(nCurrentRow<=nLastRow)
            {
                dwTimeStamp = arHttpData[nCurrentRow]->dwTimeStamp;
                wResponseCode = arHttpData[nCurrentRow]->wResponseCode;

                //Apply URL Filter
                if(m_ardwURLFilters[m_nNextGroup-1].GetSize())
                {
                    int nURLNdx = arHttpData[nCurrentRow]->nURLNdx;

                    bFound=false;
                    for(int nURL=0;nURL<m_ardwURLFilters[m_nNextGroup-1].GetSize()&&!bFound;nURL++)
                    {
                        bFound = (nURLNdx == (int)m_ardwURLFilters[m_nNextGroup-1][nURL]);
                    }

                    if(!bFound)
                        wResponseCode = 0;
                }
            }
            nCurrentRow++;
        }

        nElapsedTime = nTime-dwStartTime;

        pVar[nElapsedTime + dwNumPoints].vt = VT_R8;
        pVar[nElapsedTime + dwNumPoints].dblVal = nTotalErrors;

        strLabel.Format("%d", nElapsedTime+nTimeOffset+1);
        pVar[nElapsedTime].vt = VT_BSTR;
        pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}

//////////////////////////////////////////////////////////////////////
// CPagesPerSecondGraph Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPagesPerSecondGraph::CPagesPerSecondGraph(CGraphView* pView):
        CHttpGraph(pView,"Average Hit Response Time / Second")
{
    m_bHasURLFilter = true;
    m_pView->m_ctlChart.SetXAxisLabel("Elapsed Time");
    m_pView->m_ctlChart.SetYAxisLabel("Average Hit Response Time");
}

CPagesPerSecondGraph::~CPagesPerSecondGraph()
{
}


BOOL CPagesPerSecondGraph::AddTestRun(int nRunIndex)
{
    m_nDataRun = nRunIndex;

    // Gain access to the grid for retrieval of data
    CECVirtGridView *pGrid = pEventSink->m_pVirtGrid;

    // Get information about the data set to display
    CString strTestRun;
    g_pGlobalSettings->GetRunInfo(nRunIndex, strTestRun);
    if(!pEventSink->GetHasHttp())
        return true;

    HttpDataArray *parHttpData = pGrid->GetHttpData(strTestRun);
    if(parHttpData == NULL)
    {
        ASSERT(FALSE);
        return false;
    }
    HttpDataArray &arHttpData = *parHttpData;
    int nNumStats = arHttpData.GetSize();
    if(!nNumStats)
        return true;

    DWORD dwStartTime = arHttpData[0]->dwTimeStamp;
    DWORD dwEndTime = arHttpData[nNumStats-1]->dwTimeStamp;

    int nTimeOffset = dwStartTime - pGrid->GetStartTime(nRunIndex);

    COleSafeArray saData;

    DWORD dwNumPoints = dwEndTime-dwStartTime+1;
    DWORD dwNumElements[]= {dwNumPoints , 2};
    saData.Create(VT_VARIANT,2,dwNumElements);

    VARIANT *pVar;
    saData.AccessData((void**)&pVar);

    CString strLabel;
    int nElapsedTime=0;
    DWORD dwTotalRespTime=0;
    DWORD dwCount=0;
    DWORD dwCurrentTime=0;
    int nCurrentStat=0;

    DWORD dwTime = dwStartTime;

    while(1)
    {
        for(;nCurrentStat<nNumStats;nCurrentStat++)
            if(IsURLInFilter( arHttpData[nCurrentStat]->nURLNdx ,m_nNextGroup-1))
                break;

        if(nCurrentStat<nNumStats)
            dwCurrentTime = arHttpData[nCurrentStat]->dwTimeStamp;
        else
            dwCurrentTime = 0;

        if(dwTime == dwCurrentTime)
        {
            dwTotalRespTime += arHttpData[nCurrentStat]->dwResponseTime;
            dwCount++;
        }
        else
        {
            nElapsedTime = dwTime-dwStartTime;

            double dAvPageRespTime = (double)dwTotalRespTime / dwCount;
            pVar[nElapsedTime + dwNumPoints].vt = VT_R8;
            pVar[nElapsedTime + dwNumPoints].dblVal = dAvPageRespTime;

            strLabel.Format("%d", nElapsedTime+nTimeOffset);
            pVar[nElapsedTime].vt = VT_BSTR;
            pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);

            if(nCurrentStat>=nNumStats)
                break;

            dwTotalRespTime = arHttpData[nCurrentStat]->dwResponseTime;
            dwCount=1;

            //Fill in any empty values
            while(++dwTime<dwCurrentTime)
            {
                nElapsedTime = dwTime-dwStartTime;

                pVar[nElapsedTime + dwNumPoints].vt = VT_EMPTY;

                strLabel.Format("%d", nElapsedTime+nTimeOffset);
                pVar[nElapsedTime].vt = VT_BSTR;
                pVar[nElapsedTime].bstrVal = _com_util::ConvertStringToBSTR(strLabel);
            }
        }
        nCurrentStat++;
    }

    saData.UnaccessData();

    m_pView->m_ctlChart.SetData(saData);
    m_pView->m_ctlChart.SetSeriesStyle(m_nNextGroup,VtChSeriesType2dLine);

    // Set the data group name with the test run name.
    m_pView->m_ctlChart.SetSeriesAnnotation(m_nNextGroup++,strTestRun);

    SetDataColours();

    return true;
}

