/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// CECProjectCreator.cpp: implementation of the CECProjectCreator class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#pragma warning(disable:4786)
//
//    Product Includes
//
#include "..\..\Interfaces\CyrPlugIns.h"

//
//    Project Includes
//

#include "CECProjectCreator.h"
#include "..\..\Common\CECXServerError.h"
#include "..\..\Common\CECDefaultCodeWriter.h"
#include "..\..\Common\CECRegistryTools.h"
#include "..\..\Common\CYRTestProperties.h"
#include "resource.h"
#include "Server.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CECProjectCreator::CECProjectCreator()
{
    m_pServer = (CYRServer*) AfxGetMainWnd()->SendMessage (WMU_GET_SERVERPTR);
}

CECProjectCreator::~CECProjectCreator()
{
}


BOOL CECProjectCreator::Create(CString strName, CString strDesc)
{
    m_strName = strName;
    m_strDesc = strDesc;

    //
    // Create the project and set the description
    //
    CStringArray arstrProjects;
    m_pServer->GetProjectList(arstrProjects);
    for(int nProj=0;nProj<arstrProjects.GetSize();nProj++)
    {
        if(arstrProjects[nProj] == strName)
        {
            AfxMessageBox("A project with this name already exists.");
            return FALSE;
        }
    }

    if(!m_pServer->CreateProject(m_strName,m_strDesc))
    {
        CECXServerError error;
        error.DisplayErrorMessage("OpenSTA", m_pServer->GetLastError());
        return false;
    }
    m_pServer->OpenProject(m_strName);

    //Import the default Trace Settings into project root directory
    ImportTraceSettings(m_strName);

    //Import response_codes.inc into project source directory
    ImportResponseCodes(m_strName);

    // Create the default localhost Injector
    CreateLocalInjector(m_strName);

    m_pServer->CloseProject();

    //  Activate this new project by sending a message to the main window asking it to switch projects
    AfxGetMainWnd()->SendMessage (WMU_SET_PROJECT ,0 ,(LPARAM)(LPCTSTR)m_strName);

    return TRUE;
}

//Import response_codes.inc into project source directory
void CECProjectCreator::ImportResponseCodes(const CString& strProject)
{
    const CString strIncludeName("RESPONSE_CODES.INC");

    //Get the path for the include file
    char CyranoHome[256];
    if (GetEnvironmentVariable ("CYRANO_HOME" ,CyranoHome ,sizeof (CyranoHome)) <= 0)
    {
        AfxMessageBox("The CYRANO_HOME environment variable has not been set.");
        return;
    }

    CString strSourcePath(CyranoHome);
    strSourcePath += "\\Engines\\Web\\"+strIncludeName;

    //Get the path for the sources
    CString strTargetPath = m_pServer->GetFolderPath(CYRServer::enSourceFolder) + strIncludeName;

    if(CopyFile(strSourcePath,strTargetPath,FALSE))
    {
        //Import the file into the project
        m_pServer->CheckInFile(CYRServer::enSourceFolder,"",strIncludeName);
    }
}

//
// ImportTraceSettings
// ===================
//
// Import a default trace settings file into the project root directory to give us tracing
// before running a test. This is required to overcome the obvious need to turn on tracing
// the first time in to find out why the first created test that a user may create could fail.
void CECProjectCreator::ImportTraceSettings(const CString& strProject)
{
    const CString szTraceSettingsFile("TraceSettings");

    // Get the path for the default settings file - i.e. the <TARGETDIR>\Impact
    char CyranoHome[256];
    if (GetEnvironmentVariable ("CYRANO_HOME" ,CyranoHome ,sizeof (CyranoHome)) <= 0)
    {
        AfxMessageBox("The CYRANO_HOME environment variable has not been set.");
        return;
    }

    CString szSourcePath(CyranoHome);
    szSourcePath += "\\Engines\\" + szTraceSettingsFile;

    // Now get the path for the project path folder
    CString szTargetPath = m_pServer->GetFolderPath(CYRServer::enRootFolder) + szTraceSettingsFile;

    // And copy in the default settings to the project root folder
    if(CopyFile(szSourcePath, szTargetPath, FALSE))
    {
        m_pServer->CheckInFile(CYRServer::enRootFolder,"",szTraceSettingsFile,szTraceSettingsFile);
    }
}

//
// CreateLocalInjector
// ===================
//
// Create a local Injector for all new projects
//
void CECProjectCreator::CreateLocalInjector(const CString& strProject)
{
    // This is a three stage process

    // First we create the default node file

    // Now get the Active Computer name
    CECRegistryTools reg;
    CString szKeyName = "SYSTEM\\CurrentControlSet\\Control\\ComputerName\\ActiveComputerName";
    CString strInjectorName = reg.GetStringValue(HKEY_LOCAL_MACHINE ,szKeyName ,"ComputerName");

    // Get the path for the node file
    CString strTargetPath = m_pServer->GetFolderPath(CYRServer::enNodesFolder) + strInjectorName;

    // Create the File ...
    try
    {
        CStdioFile  File (strTargetPath, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive);
        File.WriteString (strInjectorName);
        File.WriteString ("\n");
        File.Close();
    }
    catch (CFileException e)
    {
        AfxMessageBox("Failed to create the default injector definition for this injector");
        return;
    }


    //
    //    Check in the Injector file.
    //
    m_pServer->CheckInFile(CYRServer::enNodesFolder,"",strInjectorName, strInjectorName);

    // And now we generate a tof for the ota module
    GenerateDefaultOTA(strProject);

    GenerateGlobalVars(strProject);
}

//
// GenerateDefaultOTA
// ==================
//
// Generate the default ota tof - should really use the code in Node Definition
// to do this.
//
void CECProjectCreator::GenerateDefaultOTA(const CString& strProject)
{
    CWnd* pMainWnd = AfxGetMainWnd();
    ASSERT (pMainWnd != NULL);

    CECRegistryTools reg;

    // Now get the Active Computer name
    CString szKeyName = "SYSTEM\\CurrentControlSet\\Control\\ComputerName\\ActiveComputerName";
    CString strInjectorName = reg.GetStringValue(HKEY_LOCAL_MACHINE ,szKeyName ,"ComputerName");
    CString strOTAFile = strInjectorName + "-MON.tof";

    CString strOTAPath = m_pServer->GetFolderPath(CYRServer::enObjectCodeFolder) + strOTAFile;

    //  Check if this file already under Source control...
    if(m_pServer->DoesFileExist(CYRServer::enObjectCodeFolder,"",strOTAFile))
        return;

    CECDefaultCodeWriter DCW;
    DCW.GenerateDefaultOtaTof(strOTAPath,strInjectorName);

    m_pServer->CheckInFile(CYRServer::enObjectCodeFolder,"",strOTAFile,strOTAFile);
}

void CECProjectCreator::GenerateGlobalVars(const CString& strProject)
{
    const CString strIncFile =  "GLOBAL_VARIABLES.INC";
    CString strFileName = m_pServer->GetFolderPath(CYRServer::enSourceFolder) + strIncFile;

    //  Check if this file already under Source control...
    if(m_pServer->DoesFileExist(CYRServer::enSourceFolder,"",strIncFile))
        return;

    //create an empty file
    CFile f(strFileName,CFile::modeCreate);
    f.Close();

    m_pServer->CheckInFile(CYRServer::enSourceFolder,"",strIncFile,strIncFile);
}
