/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TRecHolder.cpp
/**
 * Method definitions for the TRecHolder class
 **/

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
// Local includes
#include "TRecHolder.h"

/**
 * Default Constructor
 **/
TRecHolder::TRecHolder() :  _trecitArray(NULL), _reserved(0),
        _modified(true)
{}

/**
 * Private constructor used when creating part copies
 **/
TRecHolder::TRecHolder(treclist_t& trecs)
        :   _trecs(trecs),
        _trecitArray(NULL),
        _reserved(0),
        _modified(true)
{}

/**
 * Destructor, delete and free up each Transaction Record
 **/
TRecHolder::~TRecHolder()
{
    //
    // LINUX Port Changes
#if defined (LINUX)|| defined(HPUX)

    for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        delete *i;
    }

#else

    for(treclist_t::const_iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        delete *i;
    }

#endif

    _trecs.erase(_trecs.begin(), _trecs.end());
    delete []_trecitArray;
}

/**
 * This private method checks and fixs the array of iterators into the list
 **/
void TRecHolder::fix_trecits()
{
    iter_mutex.lock();
    if(_modified)
    {
        _length = 0;
        if(_reserved < _trecs.size())
        {
            delete []_trecitArray;
            _reserved = (_trecs.size()/RES_BLOCK + 1) * RES_BLOCK;
            _trecitArray = new treclist_t::iterator[_reserved];
        }
        for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
        {
            // Conditional needed if want to avoid invalid/unwanted trecs
            if((*i)->isKnown())
                _trecitArray[_length++] = i;
        }
        _modified = false;
    }
    iter_mutex.unlock();
}

/**
 * Accessor to save the records in this holder out, to the provided
 * stream
 **/
void TRecHolder::save(ostream& out) const
{
    TRecord* recP = NULL;
    for(treclist_t::const_iterator i = _trecs.begin();  i != _trecs.end(); i++)
    {
        if(recP != NULL)
            recP->saveDelim(out);

        recP = *i;
        recP->save(out);
    }
}

/**
 * Method to iterate through the records contained in this holder
 * finding limits of a given integer parameter
 **/
void TRecHolder::intPrmLimits(const string& key, int& min, int& max)
{

    bool firstTime = true;
    for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        const TParam* prmP = (*i)->param(key);
        // If this record has this parameter // which can be construde as an integer
        if((prmP != NULL) && prmP->rule()->intVal())
        {
            for(unsigned int idx = 0; idx < prmP->numVals(); idx++)
            {
                int val = prmP->intVal(idx);
                if(firstTime)
                {
                    firstTime = false;
                    min = val;
                    max = val;
                }
                if(val < min)
                {
                    min = val;
                }
                else if(val > max)
                {
                    max = val;
                }
            }
        }
    }
    return;
}

/**
 * Method to iterate through the records contained in this holder
 * limiting a given integer parameter to between certain levels
 **/
void TRecHolder::limitIntPrms(const string& key, int min, int max)
{

    for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        const TParam* prmP = (*i)->param(key);
        // If this record has this parameter which can be construde as an integer
        if((prmP != NULL) && prmP->rule()->intVal())
        {
            for(unsigned int idx = 0; idx < prmP->numVals(); idx++)
            {
                int val = prmP->intVal(idx);
                if(val < min)
                {
                    (*i)->setPrmVal(prmP->rule()->intVal(min), key, idx);
                }
                else if(val > max)
                {
                    (*i)->setPrmVal(prmP->rule()->intVal(max), key, idx);
                }
            }
        }
    }
    return;
}

/**
 * Method to iterate through the records contained in this holder
 * limiting a given integer parameter to between certain levels
 **/
void TRecHolder::copyPrms(const string& fromKey, const string& toKey)
{

    for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        const TParam* fPrmP = (*i)->param(fromKey);
        const TParam* tPrmP = (*i)->param(toKey);
        // If this record has this parameter // and the 'to' parameter
        if((fPrmP != NULL) && (tPrmP != NULL))
        {
            for(unsigned int idx = 0; idx < tPrmP->numVals(); idx++)
            {
                (*i)->setPrmVal(fPrmP->strVal(), toKey, idx);
            }
        }
    }
    return;
}

/**
 * Method to iterate through the records contained in this holder
 * setting those with the parameter described by the key's param
 * to the value provided.
 **/
void TRecHolder::setPrmVals(const string& key, const string& val)
{

    for(treclist_t::iterator i = _trecs.begin(); i != _trecs.end(); i++)
    {
        const TParam* prmP = (*i)->param(key);
        if(prmP != NULL)
        {   // If this record has this parameter
            for(unsigned int idx = 0; idx < prmP->numVals(); idx++)
            {
                (*i)->setPrmVal(val, key, idx);
            }
        }
    }
    return;
}

/**
 * Method to add another transaction record to the end of the list
 **/
bool TRecHolder::add(TRecord* recP)
{
    if(recP == NULL)
        return false;

    _trecs.push_back(recP);
    _modified = true;
    return true;
}

/**
 * Modifier to cut a single element out of this collection
 **/
TRecord* TRecHolder::cut(unsigned int idx)
{
    fix_trecits();
    if(idx >= _length)
        return NULL;
    TRecord* trecP = *(_trecitArray[idx]);
    _trecs.erase(_trecitArray[idx]);
    _modified = true;
    return trecP;
}

/**
 * Modifier to cut a section out of this collection
 **/
TRecHolder* TRecHolder::cut(unsigned int idx1, unsigned int idx2)
{
    if(idx1 >= idx2)
        return NULL;
    fix_trecits();
    treclist_t trecs;
    trecs.splice(trecs.end(), _trecs,
                 _trecitArray[idx1],
                 (idx2>=_length)?_trecs.end():_trecitArray[idx2]);

    _modified = true;
    return new TRecHolder(trecs);
}

/**
 * Modifier to insert a Transaction Record pointer into our container
 **/
bool TRecHolder::insert(unsigned int idx, TRecord* trecP)
{
    if(trecP == NULL)
        return false;
    fix_trecits();
    treclist_t::iterator it = (idx>=_length)?_trecs.end():_trecitArray[idx];
    _trecs.insert(it, trecP);
    _modified = true;
    return true;
}

/**
 * Modifier to insert the contents of another TRecHolder into our container
 * Requires deep copy of the contents.
 **/
bool TRecHolder::insert(unsigned int idx, TRecHolder* tRHP)
{
    if(tRHP == NULL)
        return false;
    fix_trecits();
    treclist_t::iterator it = (idx>=_length)?_trecs.end():_trecitArray[idx];
    for(treclist_t::reverse_iterator rit = tRHP->_trecs.rbegin();
            rit != tRHP->_trecs.rend();
            rit++)
    {
        it = _trecs.insert(it, (*rit)->duplicate());
    }
    _modified = true;
    return true;
}

/**
 * Method to find the first record, after or including the one specified 
 * at the index given in the parameter, that is identified by the
 * passed in key. True will be returned when found with the index set to
 * the index of the record of that type.  If none found false is returned
 * and the index zeroed.
 **/
bool TRecHolder::findRecType(const AtvfKey& key, unsigned int& idx)
{
    fix_trecits();
    while(idx < _length)
    {
        if((*(_trecitArray[idx]))->isType(key))
            return true;
        idx += 1;
    }
    idx = 0;
    return false;
}

/**
 * Method to find a piece of text within the records of this holder, 
 * starting at the record and parameter indexes given.
 * True will be returned when found with the indexes set to
 * that of the record and parameter containing the text.  If none found 
 * false is returned and the indexes zeroed.  When performing a reverse
 * search a negative index implies the end of its range and failed
 * searchs will set the indices to -1.
 **/
bool TRecHolder::find(const string& text,
                      int& recidx, int& prmidx,
                      bool matchCase,
                      bool forward)
{
    fix_trecits();
    if(forward)
    {
#ifndef __VAX
        while(recidx < static_cast<int>(_length))
#else
        while(recidx < (int)(_length))
#endif
        {
            if((recidx >= 0) && (*(_trecitArray[recidx]))->findInVal(text,
                    prmidx, matchCase,
                    forward))
                return true;
            recidx += 1;
        }
        recidx = 0;
    }
    else
    {   // reverse search
        if(recidx < 0)
            recidx = _length - 1;

        while(recidx >= 0)
        {
#ifndef __VAX
            if((recidx < static_cast<int>(_length)) &&
#else
            if((recidx < (int)(_length)) &&
#endif
                    (*(_trecitArray[recidx]))->findInVal(text, prmidx, matchCase, forward))
                return true;

            recidx -= 1;
        }
        recidx = -1;
    }
    return false;
}

/**
 * Accessor to an indexed transaction record, returned in simplest form
 * Return NULL on invalid index
 **/
TRecord* TRecHolder::operator[](unsigned int idx)
{
    fix_trecits();
    if(idx >= _length)
        return NULL;
    return *(_trecitArray[idx]);
}

/**
 * Accessor to an indexed transaction record, returned in simplest form
 * Return NULL on invalid index. This version does not rebuild the 
 * iterator array.
 **/
TRecord* TRecHolder::rec_at_index(unsigned int idx)
{
    if(idx >= _length)
        return NULL;
    return *(_trecitArray[idx]);
}

/**
 * Accessor to copy a section out of this collection
 * Copying creates a new Trecord and it is down to the caller to
 * look after this.
 **/
TRecord* TRecHolder::copy(unsigned int idx)
{
    fix_trecits();
    if(idx >= _length)
        return NULL;
    return (*(_trecitArray[idx]))->duplicate();
}

/**
 * Accessor to copy a section out of this collection
 * Copying requires a deep copy of the contents
 **/
TRecHolder* TRecHolder::copy(unsigned int idx1, unsigned int idx2)
{
    if(idx1 >= idx2)
        return NULL;
    treclist_t trecs;
    trecs.insert(trecs.end(), _trecitArray[idx1],
                 (idx2>=_length)?_trecs.end():_trecitArray[idx2]);
    // Make copy deep
    for(treclist_t::iterator i = trecs.begin(); i != trecs.end(); i++)
    {
        TRecord* trecP = (*i)->duplicate();
        *i = trecP;
    }
    return new TRecHolder(trecs);
}

/**
 * Global function to allow a record holder to be written out
 **/
ostream& operator<<(ostream& os, TRecHolder& h)
{
    os << h.length() << " Transaction Records";
    for(unsigned int i = 0; i < h.length(); i++)
    {
        os << endl << h[i];
    }
    return os;
}

