/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CECTofTxt.h
// @module  CECTofTxt and derived classes | This module defines the interface
//      to the CECTofTxt class and all classes derived from it. These
//      classes handle the parsing, encoding and evaluation of text
//      parameters (stxt and ltxt) within TOF records.
///////////////////////////////////////////////////////////////////////////////

#if !defined (CYR_CECTOFTXT)
#define CYR_CECTOFTXT

#include <string>
#include <list>
#include <stack>
#include <exception>
#include "CECOperand.h"

// Use the Standard C++ Library, by default.
#ifndef CYR_NO_STD_NAMESPACE
using namespace std;
#endif

// Forward declarations.
class TContext;
class TScript;

///////////////////////////////////////////////////////////////////////////////
// @class   ExprDsc | Embedded class to hold information about an expression
//      encoded within _exprEncoded.
class CECExprDsc
{
    // @access Public Members.
public:
    // @cmember Constructor..
    CECExprDsc(DataType dTyp, bool isVariable, string varNam,
               list<CECOperand>::const_iterator exprIt);
    // @cmember Destructor.
    virtual ~CECExprDsc();
    // @cmember Accessor function returning _dTyp.
    DataType DTyp() const
    {
        return _dTyp;
    };
    // @cmember Return 'true' if character operand, 'false' otherwise.
    bool IsChr() const
    {
        return (_dTyp==dtyp_character)?true:false;
    };
    // @cmember Return 'true' if integer operand, 'false' otherwise.
    bool IsInt() const
    {
        return (_dTyp==dtyp_integer)?true:false;
    };
    // @cmember Accessor function returning _isVariable.
    bool IsVariable() const
    {
        return _isVariable;
    };
    // @cmember Accessor function returning _varNam.
    string VarNam() const
    {
        return _varNam;
    };
    // @cmember Accessor function returning _exprIt.
    list<CECOperand>::const_iterator ExprIt() const
    {
        return _exprIt;
    };
    // @access Private Members.
private:
    // @cmember Data type.
    DataType _dTyp;
    // @cmember Flag indicating whether the expression being parsed is a variable
    // operand (possibly subscripted) or not (i.e. a literal or a 'true'
    // expression with multiple operands).
    bool _isVariable;
    // @cmember Name of variable (if _isVariable is true).
    string _varNam;
    // @cmember Iterator identifying the first operand in _exprEncoded for this expression.
    list<CECOperand>::const_iterator _exprIt;
}
; // CECExprDsc class

///////////////////////////////////////////////////////////////////////////////
// @class   CECTofTxt | This class provides generic functions for parsing text
//      parameters (stxt and ltxt) within TOF records and encoding them
//      into a more suitable and efficient form for run-time evaluation. It
//      also provides an ::Evaluate() function for performing that run-time
//      evaluation. The reason for any failure in parsing, encoding or
//      execution can be obtained using the ::Reason() function.
//
//      This is an abstract, base class, from which classes for parsing and
//      encoding specific text parameter types (e.g. those containing
//      integer expressions) are derived.
//
class CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
    CECTofTxt(const TScript* script);
    CECTofTxt(TContext* ctxt);
    CECTofTxt(const CECTofTxt& tftxt);
    // @cmember Destructor.
    virtual ~CECTofTxt();

    // @cmember Evaluate encoded text parameter value; return result.
    string Evaluate(TContext& ctxt) const;
    // @cmember Parse and encode text parameter value; return status.
    virtual bool Parse(const string& inStr, const int inPos = 0) = 0;
    // @cmember Accessor function returning reason for parsing, encoding or execution error.
    const string& Reason() const
    {
        return _reason;
    };
    // @cmember Accessor function returning number of values in value list.
    const long ValCnt () const
    {
        return _exprList.size();
    };
    // @cmember List describing the expressions encoded in exprEncoded.
    const list<CECExprDsc>* ExprListP () const
    {
        return &_exprList;
    };

    static const string stFALSE;
    static const string stTRUE;

    // @cmember Get 'raw' character value, i.e. with commands translated.
    string _getChrValue(const string& inStr, const int valStartPos, const int valEndPos);

    // @access Protected Members.
protected:
    // @cmember Check expression parsing complete.
    void _chkExprComplete();
    // @cmember Clear expression parsing data ready for more parsing.
    void _clrExpr();
    // @cmember Decrement expression level - ")" encountered.
    void _decExprLevel();
    // @cmember Return number of dimensions in supplied variable dimensions string.
    long _dimCnt (const string dimstr);
    // @cmember Encode function operand.
    void _encFnc (const FncCode fnccod);
    // @cmember Encode operator at current level and reset operator to none.
    void _encOpr();
    // @cmember Evaluate a ~EXTRACT function operand; return success/failure.
    bool _evaluateFncExtract(TContext& ctxt, list<CECOperand>::const_iterator* itP,
                             string* resstrP) const;
    // @cmember Evaluate a ~LOCATE function operand; return success/failure.
    bool _evaluateFncLocate(TContext& ctxt, list<CECOperand>::const_iterator* itP,
                            string* resstrP) const;
    // @cmember Evaluate a single operand; return success/failure.
    bool _evaluateOpd(TContext& ctxt, stack<CECOperand*>* operands,
                      list<CECOperand>::const_iterator* it) const;
    // @cmember Evaluate variable index; return success/failure.
    bool _evaluateVarIdx(TContext& ctxt, list<CECOperand>::const_iterator* itP,
                         string* idxstrP) const;
    // @cmember Get current token expanded to blank/end of string.
    const string _expandToBlank(const string inStr, const int pos);
    // @cmember Get current token expanded to end of string.
    const string _expandToEos(const string inStr, const int pos);
    // @cmember Increment expression level - "(" encountered.
    void _incExprLevel();
    // @cmember Parse and encode array indices; return array subscript flag.
    bool _psArrayIndices (const string& inStr, int* pos, int* idxcnt, const bool updpos = true);
    // @cmember Parse and encode character expression; return character expression flag.
    bool _psChrExpr (const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode character function; return character function flag.
    bool _psChrFnc (const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode character value; return character value flag.
    bool _psChrValue(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode character variable; return character variable flag.
    bool _psChrVariable(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode ~EXTRACT function reference; return ~EXTRACT function flag.
    bool _psFncExtract (const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode ~LOCATE function reference; return ~LOCATE function flag.
    bool _psFncLocate (const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode integer function; return integer function flag.
    bool _psIntFnc (const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode integer expression; return integer expression flag.
    bool _psIntExpr(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode integer value; return integer value flag.
    bool _psIntValue(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode integer variable; return integer variable flag.
    bool _psIntVariable(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode simple condition; return simple condition flag.
    bool _psSimpleCondition(const string& inStr, int* pos, const bool updpos = true);
    // @cmember Parse and encode variable; return variable flag.
    bool _psValList (const string& inStr, int* pos, long* valcnt, const bool updpos = true);
    // @cmember Parse and encode variable; return variable flag.
    bool _psVariable(const string& inStr, int* pos, const long reqopt,
                     const bool updpos = true);
    // @cmember Save operator on operand stack.
    void _saveOpr(const Operator opr);

    ///////////////////////////////////////////////////////////////////////////////
    // @class   ParseException | This embedded class is used to handle errors
    // within the CECTofTxt class or any of its derived classes. A Parse Exception
    // may be thrown by any function within the CECTofTxt class or any its derived
    // classs and should be trapped and processed in each of the top-level public
    // functions.
    //
    // @base    public | exception
class ParseException : public virtual exception
    {
        // @access Public Members.
    public:
        // @cmember Constructor.
    ParseException(const string& reason): _reason(reason)
        {}
        ;
        // @cmember Destructor.
        virtual ~ParseException()
        {}
        ;
        // @cmember Accessor function returning reason for exception.
        virtual string Reason()
        {
            return _reason;
        };
        // @access Private Members.
    private:
        // @cmember Reason for exception
        string _reason;
    }
    ; // ParseException class

    // @cmember Reason for error.
    string _reason;
    // @cmember Encoded expression to be evaluated at run-time.
    list<CECOperand> _exprEncoded;
    // @cmember List describing the expressions encoded in exprEncoded.
    list<CECExprDsc> _exprList;
    // @cmember Index level for current expression. Zero (0) if NOT parsing an
    // array subscript; > 0 if parsing an array subscript. May have nested
    // subscripts (increment on '[', decrement on ']').
    long _idxLevel;
    // @cmember Count of operands for expression being parsed (excludes array
    // subscript operands and operators).
    long _opdCnt;
    // @cmember Flag indicating whether the expression being parsed is a variable
    // operand (possibly subscripted) or not (i.e. a literal or a 'true'
    // expression with multiple operands).
    bool _isVariable;
    // @cmember Name of variable (if _isVariable is true).
    string _varNam;
    // @cmember Data type of expression being parsed.
    DataType _dTyp;
    // @cmember Iterator identifying the first operand in _exprEncoded for current expression.
    list<CECOperand>::const_iterator _exprIt;
    // @cmember Pointer to TScript record for current TOF file.
    const TScript* _scriptPtr;
    // @cmember Operator stack - used in parsing expressions.
    typedef stack<Operator> stkOperator;
    stkOperator _prsStack;
    // @cmember Flag indicating whether NOT operator specified; cleared on next '('.
    bool _notOprSpecified;
}
; // End of CECTofTxt class

///////////////////////////////////////////////////////////////////////////////
// @class   TtBoolExpr | This class handles the parsing, enoding and evaluation
//      of text parameters in TOF records containing boolean expressions.
//
// @base    public | CECTofTxt
//
class TtBoolExpr : public CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
TtBoolExpr(const TScript* script) : CECTofTxt(script)
    {}
    ;
    // @cmember Destructor.
    virtual ~TtBoolExpr()
    {}
    ;

    // @cmember Parse and encode TOF boolean expression string; return status.
    bool Parse(const string& inStr, const int inPos = 0);

    // @access Protected Members.
protected:
    // @cmember Parse and encode TOF boolean expression string.
    void _parse(const string& inStr, const int inPos);
}
; // End of TtBoolExpr class

///////////////////////////////////////////////////////////////////////////////
// @class   TtChrExpr | This class handles the parsing, enoding and evaluation
//      of text parameters in TOF records containing character expressions.
//
// @base    public | CECTofTxt
//
class TtChrExpr : public CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
TtChrExpr(const TScript* script) : CECTofTxt(script)
    {}
    ;
TtChrExpr(TContext* ctxt):CECTofTxt(ctxt)
    {}
    ;
    // @cmember Destructor.
    virtual ~TtChrExpr()
    {}
    ;

    // @cmember Parse and encode TOF character expression string; return status.
    bool Parse(const string& inStr, const int inPos = 0);

    // @access Protected Members.
protected:
    // @cmember Parse and encode TOF character expression string.
    void _parse(const string& inStr, const int inPos);
}
; // End of TtChrExpr class

///////////////////////////////////////////////////////////////////////////////
// @class   TtIntExpr | This class handles the parsing, enoding and evaluation
//      of text parameters in TOF records containing integer expressions.
//
// @base    public | CECTofTxt
//
class TtIntExpr : public CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
TtIntExpr(const TScript* script) : CECTofTxt(script)
    {}
    ;
TtIntExpr(TContext* ctxt):CECTofTxt(ctxt)
    {}
    ;
    // @cmember Destructor.
    virtual ~TtIntExpr()
    {}
    ;

    // @cmember Parse and encode TOF integer expression string; return status.
    bool Parse(const string& inStr, const int inPos = 0);

    // @access Protected Members.
protected:
    // @cmember Parse and encode TOF integer expression string.
    void _parse(const string& inStr, const int inPos);
}
; // End of TtIntExpr class

///////////////////////////////////////////////////////////////////////////////
// @class   TtValList | This class handles the parsing, encoding and evaluation
//      of text parameters in TOF records containing comma separated value
//      lists, where each value may be an integer or character expression.
//
// @base    public | CECTofTxt
//
class TtValList : public CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
TtValList(const TScript* script) : CECTofTxt(script)
    {}
    ;
TtValList(TContext* ctxt):CECTofTxt(ctxt)
    {}
    ;
TtValList(const TtValList& tt):CECTofTxt(tt),_valCnt(tt._valCnt)
    {}
    ;
    // @cmember Destructor.
    virtual ~TtValList()
    {}
    ;

    // @cmember Evaluate values in value list; return success/failure.
    bool EvaluateValues(TContext& ctxt, list<string>& retVals) const;
    // @cmember Evaluate value in value list; return success/failure.
    bool EvaluateValue(TContext& ctxt, list<CECOperand>::const_iterator& valItr,
                       string& retVal) const;
    // @cmember Evaluate array subscript, if any; return success/failure.
    bool EvaluateVarIdx(TContext& ctxt, list<CECOperand>::const_iterator* valItrP,
                        string* retValP) const;
    // @cmember Parse and encode value list string; return status.
    bool Parse(const string& inStr, const int inPos = 0);

    // @access Protected Members.
protected:
    // @cmember Evaluate a value in value list; return success/failure.
    bool _evaluateValue(TContext& ctxt, list<CECOperand>::const_iterator& itP,
                        string& retVal) const;
    // @cmember Evaluate values in value list; return success/failure.
    bool _evaluateValues(TContext& ctxt, list<CECOperand>::const_iterator* itP,
                         list<string>& retVals) const;
    // @cmember Parse and encode value list string.
    void _parse(const string& inStr, const int inPos);
    // @cmember Number of values in value list.
    long _valCnt;

}
; // End of TtValList class

///////////////////////////////////////////////////////////////////////////////
// @class   TtVarRef | This class handles the parsing, encoding and evaluation
//      of text parameters in TOF records containing variable (character or
//      integer) references.
//
// @base    public | CECTofTxt
//
class TtVarRef : public CECTofTxt
{
    // @access Public Members.
public:
    // @cmember Constructor.
TtVarRef(const TScript* script) : CECTofTxt(script)
    {}
    ;
TtVarRef(TContext* ctxt):CECTofTxt(ctxt)
    {}
    ;
    // @cmember Destructor.
    virtual ~TtVarRef()
    {}
    ;

    // @cmember Evaluate expressions in indices, if any; return variable subscript string.
    string Evaluate(TContext& ctxt) const;
    // @cmember Parse and encode TOF variable reference string; return status.
    bool Parse(const string& inStr, const int inPos = 0);
    // @cmember Accessor function returning variable name - excluding indices.
    const string& Varnam() const
    {
        return _varnam;
    };
    // @access Protected Members.
protected:
    // @cmember Parse and encode TOF variable reference string.
    void _parse(const string& inStr, const int inPos);
    // @cmember Variable name - excluding indices.
    string _varnam;

}
; // End of TtVarRef class

#endif // !defined(CYR_CECTOFTXT)
