/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// @doc     TInterRecords.cpp
// @module  TInterRecords |TRecords that require full integration with Architecture
///////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "TInterRecords.inl"


// *****************************************************************************
// * TIMERS
// *****************************************************************************

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Start Timer Record
// =====================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdStimer::execute(TContext& ctxt)
{

    ctxt.set_phase_text(PHASE_STIMER);

    string  l_name = strPrmVal(TParam::stName, ctxt);

    ctxt.log("START TIMER: " + l_name, TRACEMSG);

    ctxt.StartTimer(l_name);

    ctxt.set_phase_text(TP_PHASE_NONE);
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for End Timer Record
// ===================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdEtimer::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ETIMER);

    string  l_name = strPrmVal(TParam::stName, ctxt);

    ctxt.log("STOP TIMER: " + l_name, TRACEMSG);

    ctxt.StopTimer(l_name);
    ctxt.set_phase_text(TP_PHASE_NONE);
    return 0;
}


#ifndef __DLLSA__
///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Acquire Mutex Record
// =======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdAcmutex::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Acquire Mutex");
    unsigned    l_scope     = intPrmVal("scope", ctxt);
    string  s_name      = strPrmVal("name", ctxt);
    string  s_timeout   = strPrmVal("timeout", ctxt);
    float i_timeout=atoi(s_timeout.c_str())/1000;//Granularity of seconds
    bool timed_out=false;
    bool error=false;
    bool can_time_out=(s_timeout=="")?false:true;
    time_t stime;//start time
    time_t ctime;//current time
    time(&stime);
    bool have_acquired=false;
    try
    {
        while(!have_acquired && !timed_out )//loop until acquired or timed out
        {
            have_acquired=ctxt.LockMutex(s_name,l_scope);
            time(&ctime);
            if(((ctime-stime)>i_timeout)&& can_time_out)timed_out=true;//timeout
            ctxt.delay(500);//wait a half second
        }

    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    int dest=0;
    if(timed_out)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto then try errlab else Execute the Goto
        {
            labelIndex=PrepareGoto(this,TParam::stTmolab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            error=true;
        }
        if(!error)dest=ExecuteGoto(this,ctxt,TParam::stTmolab,&labelIndex);
    }
    if(error)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Acquire Mutex");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

}// end of TRcrdAcmutex::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Release Mutex Record
// =======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdRlmutex::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Release Mutex");
    unsigned int    l_scope = intPrmVal("scope", ctxt);
    string      s_name  = strPrmVal("name", ctxt);
    string      s_errlab    = strPrmVal("errlab", ctxt);
    bool is_unlocked=false;
    try
    {
        is_unlocked=ctxt.UnLockMutex(s_name,l_scope);
    }
    catch(...)
    {
        //treat all errors as cannot unlock
        is_unlocked=false;
    }
    int dest=0;
    if(is_unlocked==false)
    {
        if(s_errlab.empty())
            throw mdbParseException("Unable to unlock Mutex");
        int _labelIndex = ctxt.GetLabelIndex(s_errlab);
        dest=goto_destination(_labelIndex,ctxt);//jump to error label
    }

    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

}// end of TRcrdRlmutex::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Set Semaphore
// ================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdStsemaphore::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Set Sempahore");
    unsigned int    l_scope     = intPrmVal("scope", ctxt);
    string      s_name      = strPrmVal("name", ctxt);
    CECSemaphore* semaphore=new CECSemaphore(s_name,l_scope);
    bool is_set=false;
    try
    {
        is_set=semaphore->release();
    }
    catch(...)
    {
        is_set=false;
    }
    delete semaphore;
    int dest=0;
    if(is_set==false)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Set sempahore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

}// end of TRcrdStsemaphore::execute


///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Wtsemaphore Record
// =====================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdWtsemaphore::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Wait for semaphore");

    unsigned int    l_scope     = intPrmVal("scope", ctxt);
    string      s_name      = strPrmVal("name", ctxt);
    string      s_timeout   = strPrmVal("timeout", ctxt);
    float i_timeout=atoi(s_timeout.c_str())/1000;//Granularity of seconds
    bool timed_out=false;
    bool error=false;
    bool can_time_out=(s_timeout=="")?false:true;
    time_t stime;//start time
    time_t ctime;//current time
    time(&stime);
    CECSemaphore* semaphore=new CECSemaphore(s_name,l_scope);
    bool have_waited=false;
    try
    {
        while(!have_waited && !timed_out )//loop until acquired or timed out
        {
            have_waited=semaphore->tryacquire();//try and acquire semaphore, doesn't actually acquire it
            time(&ctime);
            if(((ctime-stime)>i_timeout)&& can_time_out)timed_out=true;//timeout
            ctxt.delay(500);//wait a half second
        }

    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    delete semaphore;
    int dest=0;
    if(timed_out)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto then try errlab else Execute the Goto
        {
            labelIndex=PrepareGoto(this,TParam::stTmolab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            error=true;
        }
        if(!error)dest=ExecuteGoto(this,ctxt,TParam::stTmolab,&labelIndex);
    }
    if(error)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Timed out Waiting for Sempahore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;



}// end of TRcrdWtsemaphore::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Clsemaphore Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdClsemaphore::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Clear Sempahore");
    unsigned int        l_scope     = intPrmVal("scope", ctxt);
    string      s_name      = strPrmVal("name", ctxt);
    CECSemaphore* semaphore=new CECSemaphore(s_name,l_scope);
    bool is_cleared=false;
    try
    {
        is_cleared=semaphore->acquire();
    }
    catch(...)
    {
        is_cleared=false;
    }
    delete semaphore;
    int dest=0;
    if(is_cleared==false)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Clear Semaphore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

}// end of TRcrdClsemaphore::execute

#endif __DLLSA__


