/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// Utility functions
//
#include "cxxutils.h"
#include "RegExp.h"
#include <CECInetInfo.h>

#ifdef WIN32
#include <process.h>
#elif defined (UNIX)
#include <unistd.h>
#else
#error "process functions not defined for this platform"
#endif

// Uppercase lookup table
char uppercaseTable[255];

// Lowercase lookup table
char lowercaseTable[255];

// flag to avoid recalculating our lookup tables
static bool bInitLookups;

// Prepare any lookup tables that we may wish to use
void InitLookupTables()
{
    // Prepare a lookup table for upper and lower case characters to avoid using
    // a slightly more expensive toupper/tolower.
    for (int k=0; k<256; k++)
    {
        uppercaseTable[k] = toupper(k);
        lowercaseTable[k] = tolower(k);
    }

    // And set up a static to indicate that we do not need to do this
    // ever again. Now we have done this in advance we no longer need to do
    // another toupper call again
    bInitLookups = true;
}

// Utility function, convert characters in string to upper case
void string_toupper(string & str)
{
    // See if we need to initialize the lookups

    // This is only ever done ONCE per process which gives us the performance boost,
    // OK we still have to check to see if it's been initialized but it's a small
    // overhead for the saving we gain.
    if (!bInitLookups)
        InitLookupTables();

    // And convert each character in turn by looking up its entry
    // in the lookup table
    for (char* l = (char*) str.c_str(); *l; l++)
        *l = uppercaseTable[*l];
}

// Utility function, convert characters in string to lower case
void string_tolower(string & str)
{
    // See if we need to initialize the lookups

    // This is only ever done ONCE per process which gives us the performance boost,
    // OK we still have to check to see if it's been initialized but it's a small
    // overhead for the saving we gain.
    if (!bInitLookups)
        InitLookupTables();

    // And convert each character in turn by looking up its entry
    // in the lookup table
    for (char* u = (char*) str.c_str(); *u; u++)
        *u = lowercaseTable[*u];
}

// Strip any leading directory from path return file name only,
// something like basename().
string filename_base(const string & path)
{
    string::size_type p = path.find_last_of(FN_PATHSEPC);
    if(p == path.npos)
        return path;
    return path.substr(++p);
}

// Strip any file name from path return directory only,
// something like dirname().
string filename_dir(const string & path)
{
    string::size_type p = path.find_last_of(FN_PATHSEPC);
    if(p == path.npos)
        return ".";
    return path.substr(0, p);
}

// Strip any file extension from path (anything after a dot).
string strip_extension(const string & path)
{
    string::size_type p = path.find_last_of('.');
    if(p == path.npos)
        return path;
    return path.substr(0, p);
}

// Utility function, convert integer to a string.
string int_to_string(int i)
{
    char   buf[20];

#ifdef WIN32

    //
    // If we are on win32 then use itoa for faster conversion
    return itoa(i, buf, 10);
#else

    sprintf(buf, "%d", i);
    string s(buf);
    return s;

#endif
}

// Utility function, convert unsigned integer to a string.
string int_to_string(unsigned i)
{
    char   buf[20];

#ifdef WIN32

    //
    // If we are on win32 then use itoa for faster conversion
    return itoa(i, buf, 10);

#else

sprintf(buf, "%u", i);
    string s(buf);
    return s;

#endif
}

string int_to_string(long i)
{
    char   buf[20];

#ifdef WIN32

    //
    // If we are on win32 then use ltoa for faster conversion
    return ltoa(i, buf, 10);

#else

sprintf(buf, "%ld", i);
    string s(buf);
    return s;
#endif
}

// Check if file in path can be opened for write
bool file_is_writable(string& path, string& whynot)
{
    if((access(path.c_str(), W_OK)) == 0)
        return true;
    whynot = strerror(errno);
    return false;
}

// Check if file in path can be opened for read
bool file_is_readable(string& path, string& whynot)
{
    if((access(path.c_str(), R_OK)) == 0)
        return true;
    whynot = strerror(errno);
    return false;
}

// Check if file in path exists
bool file_exists(string& path)
{
    if((access(path.c_str(), F_OK)) == 0)
        return true;
    return false;
}
// Replace text in string
bool replace_in_string(string& lookin, string& txtold, string& txtnew,
                       bool matchcase, unsigned startpos, bool replaceall)
{
    // Found flag
    bool ff = false;

    for(int i = startpos ; i != -1 ;)
    {
        i = find_in_string(lookin, txtold, matchcase, i);
        if(i == -1)
            return ff;
        ff = true;
        lookin.replace((string::size_type)i, txtold.size(), txtnew.c_str());
        i = i + txtnew.size();
        if (!replaceall)
            break;
    }
    return ff;
}

bool replace_in_string(string& lookin, CRegExp& regex, string& txtnew,
                       bool matchcase, unsigned startpos, bool replaceall)
{
    // Found flag
    bool ff = false;
    string localCopy;
    // If case insensitve, convert to uppercase
    string& lookIn = lookin;

    for(int i = startpos ; i != -1 ;)
    {
        //i = find_in_string(lookin, txtold, matchcase, i);
        int searchStart = i;
        i = regex.RegFind(lookIn.c_str()+i, (matchcase?TRUE:FALSE));
        if(i == -1)
            break;
        ff = true;
        char* replaceStr;
        replaceStr = regex.GetReplaceString(txtnew.c_str());

        i += searchStart;
        lookIn.replace((string::size_type)i, regex.GetFindLen(), replaceStr);
        i = i + strlen(replaceStr);
        delete replaceStr;
        if (!replaceall)
            break;
    }

    return ff;
}

// Optimized find_in_string function used when needing to know what a string begins
// with as opposed to scanning the entire string.
//
// We assume that the lookfor is passed in uppercase
// Find text in string, return index if found or -1 if not.
bool begins_with(string lookin, string lookfor)
{
    // See if we need to initialize the lookups

    // This is only ever done ONCE per process which gives us the performance boost,
    // OK we still have to check to see if it's been initialized but it's a small
    // overhead for the saving we gain.
    if (!bInitLookups)
        InitLookupTables();

    // It prevents corruption to the original string
    string tmplookin = lookin.c_str();

    string::size_type i=0;

    for (char* l = (char*) tmplookin.c_str(); i < lookfor.length(); l++)
    {
        // As HTTP headers can be up to 65535 in length we do not need to
        // uppercase the entire string - simply those we intend comparing
        *l = uppercaseTable[*l];

        i++;
    }

    // And directly compare the strings
    return (memcmp((char*) tmplookin.c_str(), (char*) lookfor.c_str(), lookfor.length()) == 0) ? true : false;
}

// Find text in string, return index if found or -1 if not.
int find_in_string(string lookin, string lookfor, bool matchcase, unsigned startpos)
{
    // Ensure not looking beyond end
    if(startpos >= lookin.size())
        return -1;
    // If case insensitve, convert to uppercase
    if(!matchcase)
    {
        // Temporary Hacking strings - needed to resolve possible STL bug
        // related to "copy on write" feature of strings (MSC 4.2).
        // For more information on this please refer to mdbvar.cxx

        // They prevent corruption to the original string
        string tmplookin = lookin.c_str();
        string tmplookfor = lookfor.c_str();

        string_toupper(tmplookin);
        string_toupper(tmplookfor);

        // Do search
        string::size_type p = tmplookin.find(tmplookfor, startpos);
        if(p == tmplookin.npos)
            return -1;
        return (int)p;
    }
    else
    {
        // Do search
        string::size_type p = lookin.find(lookfor, startpos);
        if(p == lookin.npos)
            return -1;
        return (int)p;
    }
}

int find_in_string(string lookin, CRegExp& regex, bool matchcase,
                   unsigned startpos)
{
    // Ensure not looking beyond end
    if(startpos >= lookin.size())
        return -1;
    // If case insensitve, convert to uppercase
    // Do search
    int p = regex.RegFind(lookin.c_str()+startpos, (matchcase?TRUE:FALSE));
    if(p != -1)
        p += startpos;
    return p;
}

// Compare two strings
bool string_compare(string string1, string string2, bool matchcase)
{
    // Compare fails if lengths differ
    if(string1.size() != string2.size())
        return false;
    for(string::size_type i = 0 ; i < string1.size() ; i++)
    {
        if(matchcase)
        {
            if(string1[i] != string2[i])
                return false;
        }
        else
        {
            if(toupper(string1[i]) != toupper(string2[i]))
                return false;
        }
    }
    return true;
}

// Remove specified characters from string
void remove_from_string(string & str, const char * remchars)
{
    string::size_type i;
    // Remove any unwanted chars
    for( ; (i = str.find_first_of(remchars)) != str.npos ; )
        str.erase(i, 1);
}

// Remove specified characters from end of string
void remove_trailing(string & str, const char * remchars)
{
    string::size_type i;
    // Remove any unwanted chars
    for( ; (i = str.find_last_of(remchars)) != str.npos ; )
    {
        if(i == (str.size() -1))
            str.erase(i, 1);
        else
            break;
    }
}

// Replace single newline with CR/LF pair
void string_lf2crlf(string & str)
{
    string::size_type i = 0;
    for( ; (i = str.find_first_of("\n", i)) != str.npos ; )
    {
        if(i == 0)
            str.replace(i, 1, "\r\n");

        else
        {
            if(str[i-1] != '\r')
                str.replace(i, 1, "\r\n");
        }
        i += 2;
    }
}
// Replace single CR with CR/LF pair
void string_cr2crlf(string & str)
{
    string::size_type i = 0;
    for( ; (i = str.find_first_of("\r", i)) != str.npos ; )
    {
        if(i == (str.size() -1))
            str.replace(i, 1, "\r\n");

        else
        {
            if(str[i+1] != '\n')
                str.replace(i, 1, "\r\n");
        }
        i += 2;
    }
}

string GetFUID()
{
    static unsigned int puid=0;
    string uident;
    // Use the nodeName, process ID and a count per use to create what should
    // be a network unique identifier for the duration of this test
    CECInetInfo info;
    string Node=info.GetIPAddressAsString();
    MDBPID pid=getpid();
    string strstr;
    strstr=Node+"-"+int_to_string(pid)+"-"+int_to_string(puid);
    puid++;
    return strstr;
}
