/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//  Cmdbgetopt implementation
//
#ifdef WIN32
#pragma warning(disable:4786)
#endif
#include "mdbgetopt.h"
#include "mdbstl.h"
#include "cxxutils.h"


// Constructor to build command line option/value list.
// Arguments are the standard argc and argv passed to main().
// The text in "facil" will be used to prefix error messages.
// A string exception is thrown on syntax error.
// **See important note for init() member.
Cmdbgetopt::Cmdbgetopt(int argc, char * argv[], string facil)
        : error(false)
{
    init(argc, argv, facil);
}

// Constructor overload to build command line option/value list.
// Argument "cmdline" is a single string containing options and values.
// The text in "facil" will be used to prefix error messages.
// A string exception is thrown on syntax error.
// This overload builds an argv[], argc from the command line string
Cmdbgetopt::Cmdbgetopt(string cmdline, string facil)
        : error(false)
{
    int     argc=0;
    ARGVEC  argvv;
    string::size_type f;
    string::size_type l = 0;

    // Loop through cmdline extracting words separated by whitespace.
    // Build temporary word list using a vector
    while(1)
    {
        // Find start of next word after position l
        f = cmdline.find_first_not_of(" \t", l);
        // If we hit end of cmdline, exit loop
        if(f == cmdline.npos)
            break;
        // Find end of word starting at f
        l = cmdline.find_first_of(" \t", f);
        // Stick word at end of vector
        argvv.push_back(cmdline.substr(f, l-f));
        // Bump count
        argc++;
    }
    // Allocate a char * array to act as argv
    char ** argv = new char *[argvv.size()+2];
    // Put pointer to words in vector in argv[].
    for(ARGVEC::size_type i = 0 ; i < argvv.size() ; i++)
        argv[i] = (char *)argvv[i].c_str();
    // Call our argc, argv[] processing member
    init(argc, argv, facil);
    // Done with argv
    delete [] argv;
}

// Destruction
Cmdbgetopt::~Cmdbgetopt()
{}

// Do actual argc, argv[] processing, called from constructors.
// NOTE: ***This function ignores argv[0] as, by convention,
// it is invokation name of the program. Use a dummy argv[0]
// if required.

void Cmdbgetopt::init(int argc, char * argv[], string facil)
{
    // Set error message prefix
    pfx = facil;

    // loop through arguments starting at argv[1]
    for(int i = 1 ; i < argc ; i++)
    {
        string opt;             // option
        string val;             // option value

        BSCLEAR(opt);           // Clear strings before each iteration
        BSCLEAR(val);           // to avoid STL bugs.

        // Expecting an option here (leading dash)
        if(*argv[i] == '-')
            opt = argv[i];
        else
        {
            // Not an option, report error
            error   = true;
            BSCLEAR(errtxt);
            errtxt   = pfx;
            errtxt  += "option syntax error <";
            errtxt  += argv[i];
            errtxt  += ">";
            // Throw string exception
            throw errtxt;
        }
        if(++i < argc)          // Check bounds
            if(*argv[i] != '-') // not another option
                val = argv[i];  // set value
            else                // no value
                --i;            // decrement argv index
        // Remove any trailing whitespace from option and value.
        remove_trailing(opt, " \t\v\r\n");
        remove_trailing(val, " \t\v\r\n");
        // Add option and (possibly empty) value to list.
        args[opt] = val;
    }
}

void Cmdbgetopt::define_option(string option, int valtype)
{
    // Insert option in list. No error check for duplicates so
    // first insertion will be used.
    valid_opts[option] = valtype;
}

void Cmdbgetopt::validate(void)
{
    // Clear error flag
    error = false;

    // Loop through options supplied
    for(CLARGLIST::iterator it = args.begin() ; it != args.end() ; it++)
    {
        // Ensure this option is in the valid list
        VOLIST::iterator i = valid_opts.find((*it).first);
        if(i == valid_opts.end())
        {
            // Not a valid option, report error
            error   =  true;
            BSCLEAR(errtxt);
            errtxt  =  pfx;
            errtxt  += "invalid option <";
            errtxt  += (*it).first;
            errtxt  += ">";
            // Throw string exception
            throw errtxt;
        }
        // Check option values match defined requirements
        switch((*i).second)
        {
            // No value allowed for option
        case CLO_NOVAL:
            // If value is not empty, report error
            if(!((*it).second).empty())
            {
                BSCLEAR(errtxt);
                errtxt  =  pfx;
                errtxt  += "value not allowed for option <";
                errtxt  += (*it).first;
                errtxt  += ">";
                throw errtxt;
            }
            break;
            // Integer value required for option
        case CLO_INTVAL:
            // Empty value is an error
            if(((*it).second).empty())
                error = true;
            else
            {
                // Allow leading sign charcter
                // NOTE: A leading minus will currently cause an error
                // in the constructor above because it is interpreted
                // as an option.
                string::size_type is = 0;
                if(((*it).second[is] == '-') || ((*it).second[is] == '+'))
                    is++;
                // Non numeric in rest of string is an error
                if((((*it).second).find_first_not_of("0123456789", is))
                        != ((*it).second).npos)
                    error = true;
            }
            // If error, report it
            if(error)
            {
                BSCLEAR(errtxt);
                errtxt  =  pfx;
                errtxt  += "integer value required for option <";
                errtxt  += (*it).first;
                errtxt  += ">";
                throw errtxt;
            }
            break;
            // Unsigned integer value required for option
        case CLO_UNSVAL:
            {
                // Check for non numeric digit in value
                string::size_type in =
                    ((*it).second).find_first_not_of("0123456789");
                // If non-numeric found or value is empty, error
                if( (in != ((*it).second).npos) || (((*it).second).empty()) )
                {
                    BSCLEAR(errtxt);
                    errtxt  =  pfx;
                    errtxt  += "unsigned integer value required for option <";
                    errtxt  += (*it).first;
                    errtxt  += ">";
                    throw errtxt;
                }
                break;
            }
            // A value required for option, don't care what it is
        case CLO_STRINGVAL:
            // If value is empty, report error
            if(((*it).second).empty())
            {
                BSCLEAR(errtxt);
                errtxt  =  pfx;
                errtxt  += "value required for option <";
                errtxt  += (*it).first;
                errtxt  += ">";
                throw errtxt;
            }
            break;
        default:
            break;
        }

    }
}

void Cmdbgetopt::print(void)
{
    for(CLARGLIST::iterator it = args.begin() ; it != args.end() ; it++)
        cout << (*it).first << " " << (*it).second << endl;
}

string Cmdbgetopt::get_value_for(string opt)
{
    // Return value for option or empty string if
    // not supplied on command line.
    CLARGLIST::const_iterator i = args.find(opt);
    if(i == args.end())
        return string("");
    return (*i).second;
}

string Cmdbgetopt::get_value_for(string opt, string dflt)
{
    // Return value for option or default string if
    // not supplied on command line.
    CLARGLIST::const_iterator i = args.find(opt);
    if(i == args.end())
        return dflt;
    return (*i).second;
}

int Cmdbgetopt::get_int_value_for(string opt)
{
    // Return value for option or zero if
    // not supplied on command line.
    CLARGLIST::const_iterator i = args.find(opt);
    if(i == args.end())
        return 0;
    return atoi((*i).second.c_str());
}

int Cmdbgetopt::get_int_value_for(string opt, int dflt)
{
    // Return value for option or default value if
    // not supplied on command line.
    CLARGLIST::const_iterator i = args.find(opt);
    if(i == args.end())
        return dflt;
    return atoi((*i).second.c_str());
}

bool Cmdbgetopt::option_supplied(string opt)
{
    // Return true if option "opt" was supplied on
    // command line.
    if((args.find(opt)) == args.end())
        return false;
    return true;
}
