/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// HTTPconn.cxx
// ------------
//=========================================================================

#include "HTTPconn.h"

#include "TPLog.h"
#include "cxxutils.h"
#include <CyrHelper.inl>
#include <RegExp.h>

// Initialise statics
CTMutex* CHTTPconn::statmtx;

// forward declaration
int Ascii2Binary(const string &inBuf, string &outBuf);


/////////////////////////////////////////////////////////////////////////////
// CHTTPconn

//
// Constructor
// =======
CHTTPconn::CHTTPconn(unsigned conid, string thread, string statLevel)
        :nID(conid)
{
    // Always set the error text to zero
    setError("");

    // Initialize member variables
    nConnectionFlags    = CF_DEFAULT;
    nPort               = 80;
    connection          = NULL;
    SSLStruct           = NULL;
    bProxy              = false;
    secureType          = HTTP_NOT_SECURE;

    // Initialize both stat file name and url file name
    int endTestId = thread.find_last_of("-");
    int threadId = atoi(thread.substr(endTestId + 1, thread.size() - endTestId).c_str());
    string tmpFileName = thread.substr(0, endTestId + 1) + int_to_string(threadId/10);

    statFileID = tmpFileName + ".stat";
    urlFileID = tmpFileName + ".urls";

    string filename = CyrGetModuleName().data();
    int index = find_in_string(filename,"TExecuter_web", false);
    statFileID = filename.substr(0, index) + "TEW_"  + statFileID;
    urlFileID = filename.substr(0, index) + "TEW_"  + urlFileID;


#ifndef __DLLSA__
    CECInetInfo inetInfo;
    string szHostName = inetInfo.GetHostName();
    userID.ip = inetInfo.GetIPAddress(szHostName);
    userID.pid = getpid();
    userID.user = threadId;

    if ( ((statLevel == "L") && (userID.user < 10)) || (statLevel == "H") )
        bCollectStat = true;
    else
        bCollectStat = false;
#endif
}


//
// Destructor
// =======
CHTTPconn::~CHTTPconn()
{
    // Write statistics in file
    printStatistics();

    if(nConnectionFlags & CF_CONNECTED)
    {
        // Clean up connection (logoff)
        disconnect();
    }
    else
        if (connection != NULL)
        {
            // If failed connection set to null
            nConnectionFlags &= ~CF_CONNECTED;
            delete connection ;
            connection = NULL;
        }

    if (SSLStruct)
    {
        // send SSL/TLS close_notify
        SSL_shutdown (SSLStruct->ssl);
        // Clean up.
        SSL_free (SSLStruct->ssl);
        SSL_CTX_free (SSLStruct->ctx);
        delete SSLStruct;
        SSLStruct = NULL;
    }
}


//
// addStatistic
// =======
//
// Put statistics list
//
void CHTTPconn::addStatistic(string pageName, time_t stamp, const CTTime& rqstTime)
{
#ifndef __DLLSA__
    if (bCollectStat)
        statisticList.push_back(CStatistics(pageName, stamp, rqstTime));
#endif
}


//
// printStatistics
// =======
//
// Put statistics in file
//
void CHTTPconn::printStatistics(void)
{
#ifndef __DLLSA__

    if ( ! bCollectStat )
        return;

    // Write statistics in file
    CTLockGuard<CTMutex> guard(statmtx[userID.user/10]);  // acquires mutex


    FILE *FIDstat;
    FILE *FIDurl;

#ifdef WIN32
    if ( ((FIDstat = fopen(statFileID.c_str(), "ab")) != NULL) && ((FIDurl = fopen(urlFileID.c_str(), "ab")) != NULL) )
#else
if ( ((FIDstat = fopen(statFileID.c_str(), "a")) != NULL) && ((FIDurl = fopen(urlFileID.c_str(), "a")) != NULL) )
#endif
    {
        // Write all statistics for this connection.
        while (!statisticList.empty())
        {
            const CStatistics &stats = statisticList.front();

            time_t timeStamp        = littlel(stats.getTimeStamp());
            unsigned long responseTime;
            if (stats.getResponseTime().isNull())
                responseTime = 0;
            else
            {
                responseTime = littlel((unsigned long)(stats.getResponseTime() - stats.getRequestTime()));
            }
            short replyCode         = littles(stats.getReplyCode());
            unsigned long replySize = littlel(stats.getReplySize());
            char zero = 0;

            // Write statistics in file
            fwrite(&timeStamp, sizeof(time_t), 1, FIDstat);

            fwrite(&userID.ip, sizeof(unsigned long), 1, FIDstat);
            fwrite(&userID.pid, sizeof(int), 1, FIDstat);
            fwrite(&userID.user, sizeof(short), 1, FIDstat);

            fwrite(&responseTime, sizeof(unsigned long), 1, FIDstat);

            fwrite(&replyCode, sizeof(short), 1, FIDstat);

            fwrite(&replySize, sizeof(unsigned long), 1, FIDstat);


            // Write URL in file
            fprintf(FIDurl, "%s", stats.getUrl().c_str());
            fwrite(&zero, sizeof(char), 1, FIDurl);

            statisticList.pop_front();
        }
        fclose(FIDstat);
        fclose(FIDurl);
    }
#endif
}



//
// printAllStatistics
// =======
//
// Put statistics in file for all connections
//
void CHTTPconn::printAllStatistics(TContext &ctxt)
{
    // Get connections map
    cons_map_t &conMap = ctxt.get_connections();

    for(cons_map_t::iterator j = conMap.begin(); j != conMap.end(); j++)
    {
        (static_cast<CHTTPconn*>((*j).second))->printStatistics();
    }
}




//
// connect
// =======
//
// Connect to host by port (default 80)
//
int CHTTPconn::connect(const string& host, int port/*=80*/)
{
    // Reset any error text we may of set
    setError("");

    // Store the connection details
    setHost(host);
    setPort(port);

    // And try for a connection
    try
    {
        if (SSLStruct)
        {
            delete SSLStruct;
            SSLStruct = NULL;
        }

        if (connection != NULL)
            delete connection ;

        connection = new CTSocket() ;

        // Set option to socket so that it will be possible to bind
        // to its port just after closesocket()
        struct linger my_linger;
        my_linger.l_onoff=1;
        my_linger.l_linger=0;

        connection->connect(getHost(), getPort());
        connection->setLinger((const char *)&my_linger, sizeof(my_linger));

        // Flag a valid connection
        nConnectionFlags |= CF_CONNECTED;
    }
    catch (const CTSocketException& e)
    {
        setError("Failed to connect : " + string(e.what()));
        return -1;
    }

    return 0;
}


//
// request
// =======
//
// send a http request through our connection either insecure mode or not
//
int CHTTPconn::request(int nType, const string &szHeader, const string &szBody, bool bSaveDoc /*= false*/)
{
    // We will send a new request, so cleanup the last document
    Document.cleanup();
	Document.setSaveDocument(bSaveDoc);

    if (secureType == HTTP_NOT_SECURE)
        return normalRequest(nType, szHeader, szBody);
    else
        return SSLRequest(nType, szHeader, szBody);
}



//
// normalRequest
// =======
//
// send a http request through our connection
//
int CHTTPconn::normalRequest(int nType, const string &szHeader, const string &szBody)
{
    // Reset any error text we may of set
    setError("");

    // Only do this if we have a valid connection
    if (isConnected())
    {
        // Set flag to CF_SEND (in case connection will close while sending the request)
        nConnectionFlags |= CF_SEND;
        nConnectionFlags &= ~CF_RECV;

        try
        {
            string szPacket;
            szPacket.append(szHeader);
            szPacket.append(szBody);

            connection->sendBuffer((void *)szPacket.c_str(), szPacket.size(), "");

            Document.setLastRequestType((enumRequestType)nType);
        }
        catch(CTSocketException &msg)
        {
            setError("Disconnection while sending a request: " + string(msg.what()));
            return -1;
        }
    }
    else
    {
        setError("Not connected to host.");
        return -1;
    }

    setInfo(szHeader.substr(0, szHeader.find_first_of("\r\n")));

    // If statistic list become large
    if (statisticList.size() >= MAX_STATLIST_SIZE)
        printStatistics();

    CTTime t;
    t.now();
    addStatistic(szHeader.substr(0, szHeader.find_first_of("\r\n")), time(NULL), t);

    // The request was send, we will have to read the response
    nConnectionFlags |= CF_RECV;
    nConnectionFlags &= ~CF_SEND;

    return 0;
}


//
// SSLrequest
// =======
//
// send a http request through a SSL connection
//
int CHTTPconn::SSLRequest(int nType, const string &szHeader, const string &szBody)
{
    // Reset any error text we may of set
    setError("");

    // Only do this if we have a valid connection
    if (isConnected())
    {
        // Set flag to CF_SEND (in case connection will close while sending the request)
        nConnectionFlags |= CF_SEND;
        nConnectionFlags &= ~CF_RECV;

        // ------------------------------
        string szPacket;
        szPacket.append(szHeader);
        // ------------------------------

        // ------------------------------
        // Do the SSL negotiation if not already done.
        if (SSLStruct == NULL)
        {
            // --- If the connection was made through a proxy server, send a CONNECT packet
            if (bProxy)
            {
                // ------------------------------
                // Extract server name from request line
                string server_name;
                int url_begin       = find_in_string(szHeader, "https://", false, 0);
                int server_begin    = url_begin + strlen("https://");
                int server_end      = szHeader.find_first_of("/ ", server_begin);
                if ( (server_begin != string::npos) && (server_end != string::npos) )
                    server_name = szHeader.substr(server_begin, server_end - server_begin);
                // ------------------------------

                // ------------------------------
                // Extract user-agent from request line
                string user_agent;
                int endField;
                int beginField;
                beginField = find_in_string(szHeader, "User-Agent", false, 0);
                if (beginField != string::npos)
                    endField = szHeader.find_first_of("\r\n", beginField);
                if (endField != string::npos)
                    user_agent = szHeader.substr(beginField, endField - beginField);
                // ------------------------------


                string szConnectData = "CONNECT ";

                // Extract version from request line
                int version_begin   = szHeader.find(" ", server_begin) + 1;
                int version_end     = szHeader.find_first_of(" \r\n", version_begin);
                string version      = szHeader.substr(version_begin, version_end - version_begin);

                szConnectData.append(server_name + " " + version + "\r\n");
                szConnectData.append(user_agent + "\r\n");
                szConnectData.append("Host: " + server_name + "\r\n");
                szConnectData.append("Content-Length: 0\r\n");
                szConnectData.append("Pragma: no-cache\r\n\r\n");

                try
                {
                    connection->sendBuffer((void *)szConnectData.c_str(), szConnectData.size(), "");

                    char recvBuf[512];
                    connection->recvAvailableData(recvBuf, sizeof(recvBuf));
                }
                catch(CTSocketException &msg)
                {
                    setError("Disconnection while sending a CONNECT: " + string(msg.what()));
                    return -1;
                }

                szPacket.erase(url_begin, server_end - url_begin);
            }

            // --- Start SSL negotiation
            if (SSLInit() == -1)
            {
                setError("SSL negotiation failed");
                return -1;
            }
        }
        // ------------------------------

        szPacket.append(szBody);

        if (SSL_write (SSLStruct->ssl, szPacket.data(), szPacket.length()) == -1)
        {
            setError("Error while sending request using SSL.");
            return -1;
        }
        // ------------------------------

        Document.setLastRequestType((enumRequestType)nType);
    }
    else
    {
        setError("Not connected to host.");
        return -1;
    }


    setInfo(szHeader.substr(0, szHeader.find_first_of("\r\n")));

    // If statistic list become large
    if (statisticList.size() >= MAX_STATLIST_SIZE)
        printStatistics();

    CTTime t;
    t.now();
    addStatistic(szHeader.substr(0, szHeader.find_first_of("\r\n")), time(NULL), t);

    // The request was send, we will have to read the response
    nConnectionFlags |= CF_RECV;
    nConnectionFlags &= ~CF_SEND;

    return 0;
}




//
// SSLInit
// =======
//
// Initialize SSL
//
int CHTTPconn::SSLInit(void)
{

    SSLStruct = new SSLStructure;

    SSLRand();
    SSLeay_add_ssl_algorithms();

    SSLStruct->meth = SSLv23_method();

    SSLStruct->ctx = SSL_CTX_new (SSLStruct->meth);
    if (SSLStruct->ctx == NULL)
    {
        setError("Error during SSL negotiation.");
        return -1;
    }

    SSLStruct->ssl = SSL_new (SSLStruct->ctx);
    if (SSLStruct->ssl == NULL)
    {
        setError("Error during SSL negotiation.");
        return -1;
    }

    SSL_set_fd (SSLStruct->ssl, getSocket()->getFileDescriptor());

    if (SSL_connect (SSLStruct->ssl) == -1)
    {
        setError("Error while an attempt to connect using SSL.");
        return -1;
    }

    return 0;
}


//
// SSLRand
// =======
//
//
//
int CHTTPconn::SSLRand()
{
    unsigned char md[MD5_DIGEST_LENGTH];
    clock_t now = (clock() * 1000) / CLOCKS_PER_SEC;
    for (long i = now; i < MD5_DIGEST_LENGTH + now; i++)
        md[i - now] = i%255 ;

    // Seed the random generator with the MD5 digest
    RAND_seed(md, MD5_DIGEST_LENGTH);

    return 0;
}



//
// receiveHeader
// =======
//
// Receive a http response's header
// This method is called unless the end of header was already recognized ("\r\n\r\n")
//
int CHTTPconn::receiveHeader(int size)
{

    int nData = 0;

    nData = receiveData(size);


    // if "\r\n\r\n" or "\n\n" (end of header) is present
    bool endOfHeader = (Document.find("\r\n\r\n") != string::npos) || (Document.find("\n\n") != string::npos) ;

    if ( (!Document.isHeaderLoaded()) && endOfHeader )
    {
        Document.setHeaderLoaded(true);

        if ( (Document.getReplyCode() >= 100) && (Document.getReplyCode() < 200) )
        {
            // This message is just informations.
            // The real response will follow.
            Document.remove1XXResponse();
            // If the real header is not here
            endOfHeader = (Document.find("\r\n\r\n") != string::npos) || (Document.find("\n\n") != string::npos) ;
            if (!endOfHeader)
            {
                Document.setHeaderLoaded(false);
                return nData;
            }
            else
                Document.setHeaderLoaded(true);
        }

        // Show reply code
        char buf[40];
        sprintf(buf, "Connection id %d, reply code %d", nID, Document.getReplyCode());
        logr->AddTrace(USERTRACE, TRACELEVEL, buf);

        // If the last request was a HEAD or the reply code is 1xx, 204 or 304,
        // the response is finished
        if (    (Document.getLastRequestType() == HTTP_REQUEST_HEAD)
                ||  (Document.getReplyCode() == 204)
                ||  (Document.getReplyCode() == 304)
           )
        {
            responseLoaded();
            return nData;
        }

        switch(Document.getTransferType())
        {
        case HTTP_TRANSFER_LENGTH_SPECIFIED:
            // Response size is known
            // Perhaps we have already received all the response
            if ( Document.getNbBytesLoaded() >= Document.getTotalLength() )
            {
                responseLoaded();
            }
            break;

        case HTTP_TRANSFER_CHUNK:
            // Response is chunked
            // Perhaps we have already read a part of a chunk
            {
                int chunkSize = 0;
                // Size field of the chunk is here
                while ( (chunkSize = Document.getChunkSize()) > 0 )
                {
                    // Do not include chunk header field in result document
                    if (Document.removeChunkInformation() == 0)
                        // if we successful remove the chunk informations
                        Document.incTotalLength(chunkSize + 2);
                }

                // If last chunk is here
                if ( (chunkSize == 0) && (!Document.isLastChunk()) )
                {
                    if (Document.removeChunkInformation() == 0)
                        Document.setLastChunk(true);
                }

                if ( (Document.isLastChunk())
                        && ( (Document.find("\r\n\r\n", Document.getTotalLength()) != string::npos) // trailers followed by an empty line
                             || (Document.find("\r\n", Document.getTotalLength()) == Document.getTotalLength()) ) // no trailers, just an empty line
                   )
                {
                    responseLoaded();
                }

                break;
            }

        case HTTP_TRANSFER_CLOSE_SOCKET:
            // Response will finish when socket is closed
        default:
            // Transfer mode is unknown
            break;
        }
    }

    return nData;
}


//
// receiveUntilSocketClose
// =======
//
// Receive a http response until connection is closed
//
int CHTTPconn::receiveUntilSocketClose(int size)
{

    int nData = 0;

    nData = receiveData(size);

    return nData;
}


//
// receiveUntilSpecifiedLength
// =======
//
// Receive a http response until a known length
//
int CHTTPconn::receiveUntilSpecifiedLength(int size)
{

    int nData = 0;

    nData = receiveData(size);

    if ( Document.getNbBytesLoaded() >= Document.getTotalLength() )
    {
        responseLoaded();
    }

    return nData;
}



//
// receiveUntilAllChunkLoaded
// =======
//
// Receive a http response until all chunk are loaded
//
int CHTTPconn::receiveUntilAllChunkLoaded(int size)
{
    int nData = 0;

    nData = receiveData(size);

    int chunkSize = 0;
    // Size field of the chunk is here
    while ( (chunkSize = Document.getChunkSize()) > 0 )
    {
        // Do not include chunk header field in result document
        if (Document.removeChunkInformation() == 0)
            Document.incTotalLength(chunkSize + 2);
    }

    // If last chunk is here
    if ( (chunkSize == 0) && (!Document.isLastChunk()) )
    {
        if (Document.removeChunkInformation() == 0)
            Document.setLastChunk(true);
    }

    // If we have receive the last chunk and we can find an empty line, this is the end
    if ( (Document.isLastChunk())
            && ( (Document.find("\r\n\r\n", Document.getTotalLength()) != string::npos) // trailers followed by an empty line
                 || (Document.find("\r\n", Document.getTotalLength()) == Document.getTotalLength()) ) // no trailers, just an empty line
       )
    {
        responseLoaded();
    }

    return nData;
}



//
// receive
// =======
//
// Call the good receive method regard to the transfer method
//
int CHTTPconn::receive(int size)
{
    // Reset any error text we may of set
    setError("");

    // If receiving a new document, destroy the last
    // (isResponseLoaded() is true if all document was loaded,
    // CF_RECV is set if we just send a request)
    if ( (Document.isResponseLoaded()) && (nConnectionFlags & CF_RECV) )
    {
        Document.setResponseLoaded(false);
//        Document.cleanup();
    }

    CTTime t;
    t.now();
    Document.setLastReceiveTime(t);

    if ( isConnected() )
    {
        switch(Document.getTransferType())
        {
        case HTTP_TRANSFER_NONE:            // Header not already loaded
            return receiveHeader(size);
            break;
        case HTTP_TRANSFER_CLOSE_SOCKET:    // Response will finish when socket is closed
            return receiveUntilSocketClose(size);
            break;
        case HTTP_TRANSFER_LENGTH_SPECIFIED:// Response size is known
            return receiveUntilSpecifiedLength(size);
            break;
        case HTTP_TRANSFER_CHUNK:           // Response is chunked
            return receiveUntilAllChunkLoaded(size);
            break;
        default:
            return 0;
            break;
        }
    }
    else
    {
        setError("Not connected to host.");
        return -1;
    }
}


//
// receiveData
// ==========
//
// Receive size octets on this connection
//
int CHTTPconn::receiveData(int size)
{
    setError("");

    int nData = 0;
    char *buffer = new char[size + 1];

    try
    {
        if (secureType == HTTP_NOT_SECURE)
            nData = connection->recvAvailableData(buffer, size);
        else
        {
            // There is something to read
            nData = SSL_read (SSLStruct->ssl, buffer, size);

            // SSL_pending return the number of remaining bytes
            int ret = SSL_pending(SSLStruct->ssl);

            // Loop and receive until end of datas
            int loop = 0;
            if ((ret%size) != 0)
                loop = 1;

            for (int j=0; (ret != 0) && (j < (ret/size) + loop); j++)
            {
                if (nData > 0)
                    Document.append(buffer, nData);
                nData = SSL_read (SSLStruct->ssl, buffer, size);
            }
        }


        // we received something, copy it in document
        if (nData > 0)
            Document.append(buffer, nData);
        // If connection was closed by server
        // (select tells us there is something to read on
        // this connection but there is nothing)
        else
        {
            responseLoaded();
            nConnectionFlags &= ~CF_CONNECTED;
        }

        if (buffer)
        {
            delete[] buffer;
            buffer = NULL;
        }

    }
    catch(CTSocketException &msg)
    {
        if (Document.getTransferType() != HTTP_TRANSFER_CLOSE_SOCKET)
            setError("Connection ended while receiving header: " + string(msg.what()));
        if (buffer)
        {
            delete[] buffer;
            buffer = NULL;
        }

        Document.setResponseLoaded(true);
        nConnectionFlags &= ~CF_RECV;
        nConnectionFlags &= ~CF_CONNECTED;

        updateStatistics();
    }

    return nData;
}




//
// getHttpResponse
// ==========
//
// Get the document
//
string CHTTPconn::getHttpResponse(int nType, unsigned int size) const
{
    if (nType == HEADER_VAR)
        return Document.getHeaderValue(size);
    else // nType == BODY_VAR
        return Document.getBodyValue(size);
}


//
// getHttpResponse
// ==========
//
// Get a value from a variable
//
string CHTTPconn::getHeaderResponse(string szIdent) const
{
    string returnValue;

    // If szIdent begins with "Set-Cookie", cookie's name follows.
    if ( begins_with(szIdent, "SET-COOKIE"))
    {
        int beginName;
        int endName;
        beginName = szIdent.find_first_not_of(", \r\n", strlen("Set-Cookie"));
        if (beginName == string::npos)
            return "";
        endName = szIdent.find_first_of(" \r\n", beginName);
        if (endName == string::npos)
            endName = szIdent.size();
        string szCookieName = szIdent.substr(beginName, endName - beginName);

        returnValue = Document.getHeaderValue("Set-Cookie", szCookieName);

        // If this cookie-name is not available, show the others Set-Cookie fields
        // (only for unique replay)
        if (returnValue == "")
        {
            int i = 1;
            logr->AddTrace(USERTRACE, ERRORLEVEL, (string("Cannot find cookie " + szCookieName
                                                   + ", must be one of the following: ")).c_str());
            while ( (returnValue = Document.getHeaderValue("Set-Cookie", i++)) != "")
            {
                logr->AddTrace(USERTRACE, ERRORLEVEL, returnValue.c_str());
            }
        }

        return returnValue;
    }
    else
    {
        returnValue = Document.getHeaderValue(szIdent);

        if (returnValue == "")
            logr->AddTrace(USERTRACE, ERRORLEVEL, ("Cannot find header field  " + szIdent).c_str());

        return returnValue;
    }
}


bool CHTTPconn::getBodyResponse(AddressQueryList & aql, Parser::AddressValueList & avl)
{
	return Document.getBodyValue(aql, avl) ;
}

//
// disconnect
// ==========
//
// Disconnect from the host
//
int CHTTPconn::disconnect()
{
    // Deallocate SSLStruct
    if (SSLStruct)
    {
        delete SSLStruct;
        SSLStruct = NULL;
    }

    if (connection)
    {
        try
        {
            connection->close();
        }
        catch(CTSocketException&)
        {}


        delete connection;
        connection = NULL;
    }

    // And take off all flags
    nConnectionFlags &= CF_DEFAULT;

    return 0;
}



//
// getHTTPReplyMsg
// ===============
//
// Translate the HTTP error codes
//
string CHTTPconn::getHTTPReplyText(int nCode) const
{
    string szReplyText;

    switch (nCode)
    {
    case 100 :
        szReplyText = "100: Continue";
        break;
    case 101 :
        szReplyText = "101: Switching Protocols";
        break;
    case 200 :
        szReplyText = "200: OK";
        break;
    case 201 :
        szReplyText = "201: Created";
        break;
    case 202 :
        szReplyText = "202: Accepted";
        break;
    case 203 :
        szReplyText = "203: Non-Authoritative Information";
        break;
    case 204 :
        szReplyText = "204: No Content";
        break;
    case 205 :
        szReplyText = "205: Reset Content";
        break;
    case 206 :
        szReplyText = "206: Partial Content";
        break;
    case 300 :
        szReplyText = "300: Multiple Choices";
        break;
    case 301 :
        szReplyText = "301: Moved Permanently";
        break;
    case 302 :
        szReplyText = "302: Found";
        break;
    case 303 :
        szReplyText = "303: See Other";
        break;
    case 304 :
        szReplyText = "304: Not Modified";
        break;
    case 305 :
        szReplyText = "305: Use Proxy";
        break;
    case 307 :
        szReplyText = "307: Temporary Redirect";
        break;
    case 400 :
        szReplyText = "400: Bad Request";
        break;
    case 401 :
        szReplyText = "401: Unauthorized";
        break;
    case 402 :
        szReplyText = "402: Payment Required";
        break;
    case 403 :
        szReplyText = "403: Forbidden";
        break;
    case 404 :
        szReplyText = "404: Not Found";
        break;
    case 405 :
        szReplyText = "405: Method Not Allowed";
        break;
    case 406 :
        szReplyText = "406: Not Acceptable";
        break;
    case 407 :
        szReplyText = "407: Proxy Authentication Required";
        break;
    case 408 :
        szReplyText = "408: Request Time-out";
        break;
    case 409 :
        szReplyText = "409: Conflict";
        break;
    case 410 :
        szReplyText = "410: Gone";
        break;
    case 411 :
        szReplyText = "411: Length Required";
        break;
    case 412 :
        szReplyText = "412: Precondition Failed";
        break;
    case 413 :
        szReplyText = "413: Request Entity Too Large";
        break;
    case 414 :
        szReplyText = "414: Request-URI Too Large";
        break;
    case 415 :
        szReplyText = "415: Unsupported Media Type";
        break;
    case 416 :
        szReplyText = "416: Requested range not satisfiable";
        break;
    case 417 :
        szReplyText = "417: Expectation Failed";
        break;
    case 500 :
        szReplyText = "500: Internal Server Error";
        break;
    case 501 :
        szReplyText = "501: Not Implemented";
        break;
    case 502 :
        szReplyText = "502: Bad Gateway";
        break;
    case 503 :
        szReplyText = "503: Service Unavailable";
        break;
    case 504 :
        szReplyText = "504: Gateway Time-out";
        break;
    case 505 :
        szReplyText = "505: HTTP Version not supported";
        break;
    default :
        szReplyText = "XXX: Unexpected Error occurred";
        break;
    }

    return szReplyText;
}


//
// receiveOnAllConnections
// ===============
//
// Receive datas from each connection
//
CHTTPconn::enumReceiveCode CHTTPconn::receiveOnAllConnections(TContext &ctxt,
        unsigned long timeout /*= 0*/,
        CHTTPconn* pConnection /*= NULL*/,
        int location /*= 1*/,
        bool dontSleep /* = false */)
{
    try
    {
        //--- Get connections map
        cons_map_t &conMap = ctxt.get_connections();

        CTSocketSelectorList waitOn;
        CTSocketSelectorList ready;

        CTTime  finish;
        CTTime  now;
        double  selectTimeOut;
        bool    receiveEnd = false;
        bool    timeoutEnd = false;
        CTTime  globalTimeoutValue;
        CTTime  cttimeout(0, timeout);

        //--- Get value for REQUEST_TIMOUT
        try
        {
            globalTimeoutValue = atoi((ctxt.get_variable(REQUEST_TIMEOUT)).c_str());
        }
        catch(mdbParseException &)
        {
            // REQUEST_TIMOUT variable does not exists, default is 0 (ie no timeout)
            globalTimeoutValue = 0;
        }

        //--- If we want to receive all data or only the header on a connection
        if (pConnection)
        {
            if (location == 0)
                receiveEnd = pConnection->isHeaderLoaded();
            else
                receiveEnd = pConnection->isReceiveFinished();
        }

        //--- If we want to receive data until a specified delay
        if (timeout != 0)
        {
            selectTimeOut = timeout/1000.;
            now.now();
            finish = cttimeout + now;
            if (finish <= now) // seems to be impossible !
                timeoutEnd = true;
        }
        else
            selectTimeOut = TIMEOUT;

        timeval timevalStruct;
        timevalStruct.tv_sec = 0;
        timevalStruct.tv_usec = (timeout * CLOCKS_PER_SEC) / 1000.;


        //--- While the response is not completely loaded or given time is not elapsed,
        // wait for something to read on all connected sockets
        while ( ! receiveEnd && ! timeoutEnd)
        {
            waitOn.reset () ;

            // ------------------------------
            // Put all other connections with valid sockets and something to receive
            for(cons_map_t::iterator j = conMap.begin(); j != conMap.end(); j++)
            {
                if ( !(static_cast<CHTTPconn*>((*j).second))->isConnected()
                        || (static_cast<CHTTPconn*>((*j).second))->isReceiveFinished() )
                    continue;

                // If a timeout has occured on connection, exit and goto end of script
                if ( (static_cast<CHTTPconn*>((*j).second))->hasTimedOut(globalTimeoutValue) == true )
                {
                    ctxt.set_variable(ON_TIMOUT_MESSAGE, string("Receive has timed out on conid ") + int_to_string((static_cast<CHTTPconn*>((*j).second))->getID())
                                      + string(": ") + (static_cast<CHTTPconn*>((*j).second))->getInfo());
                    return HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER;
                }

                // Add sockets to the selector.
                waitOn.add (*((static_cast<CHTTPconn*>((*j).second))->getSocket()), _S_RECEIVE_ ) ;
                // Set the polling waiting mode.
                waitOn.setTimeOut (&timevalStruct) ;
            }
            // ------------------------------

            // ------------------------------
            // Select connections
            int iTypeOfEvent = 0;
            if (!waitOn.isEmpty())
                waitOn(ready, iTypeOfEvent);
            else
                // possible if pConnection is NULL
                // and there is nothing more to receive on all others connections
                // (typically on a TRcrdWait or on a TRcrdSynchronize with bigs delays)
                break;
            // ------------------------------

			if (ready.isEmpty())
				CyrSleep(3);
			else
			{
				// ------------------------------
				// Retreive the connection associates with each socket
				// and call receive(DEFAULT_RECEIVE_SIZE)
				CHTTPconn *pCon;
				for(cons_map_t::iterator k = conMap.begin(); k != conMap.end(); k++)
				{
					iTypeOfEvent = 0;
					pCon = (static_cast<CHTTPconn*>((*k).second));
					if ( (ready.contains(*(pCon->getSocket()), iTypeOfEvent)) && (iTypeOfEvent & _S_RECEIVE_) )
					{
						pCon->receive(DEFAULT_RECEIVE_SIZE);

						// We have received something, update statistics
						pCon->updateStatistics();

						// If retcod or retsts variables were specified when request was done
						if ( (pCon->getTimerName() != "") && (pCon->isReceiveFinished()) )
							ctxt.StopTimer(pCon->getTimerName());
						if ( ((pCon->getCodeName() != "")||(pCon->getStatusName() != "")) && (pCon->isStatusCodeLoaded()) )
						{
							ctxt.set_variable(pCon->getCodeName(), int_to_string(pCon->getReplyCode()));
							ctxt.set_variable(pCon->getStatusName(), int_to_string(pCon->getStatus()));
						}
					}
				}
				// ------------------------------
			}

            // ------------------------------
            // Clear vectors
            waitOn.reset();
            ready.reset();
            // ------------------------------

            // ------------------------------
            // Check if we have to receive again
            if (pConnection)
            {
                // If we search for something in the http header
                if (location == 0)
                    receiveEnd = pConnection->isHeaderLoaded();
                // If the information is in the body
                else
                    receiveEnd = pConnection->isReceiveFinished();
            }

            // If the timeout is reached
            now.now();
            if ( (timeout != 0) && (finish <= now) )
            {
                timeoutEnd = true;
                // if dontSleep is true and timeout is reached
                // (ex: a TRcrdSynchronize object with a little timeout)
                if (dontSleep)
                    return HTTP_RECEIVE_EXIT_ON_LOCAL_TIMER;
            }
            // ------------------------------
        }

        // If timeout is not elapsed, sleep

        now.now();
        if ( (timeout != 0) && (finish > now) && (dontSleep == false) )
            CyrSleep((unsigned long)(finish - now));
    }
    catch(CTException& e)
    {
        ctxt.log( string("Exception while receiving data: ") + e.what(), ERRMSG);
        return HTTP_RECEIVE_ERROR;
    }

    return HTTP_RECEIVE_END;
}


//
// responseLoaded
// =======
//
// Set statistic values for this response
//
void CHTTPconn::responseLoaded(void)
{
    Document.setResponseLoaded(true);
    nConnectionFlags &= ~CF_RECV;

    updateStatistics();

    timeval t;
    t.tv_sec = 0;
    t.tv_usec = CLOCKS_PER_SEC * 10;
    if ( (connection->select(t, _S_RECEIVE_)) == CTSocket::RECEIVE )
    {
        // something happened !!
        char buf[5];
        int nData = 0;
        try
        {
            nData = connection->recvAvailableData(buf, 5);
        }
        catch (const CTSocketException& )
        {}

        if (nData == 0)
            nConnectionFlags &= ~CF_CONNECTED;
    }
}


//
// updateStatistics
// ===========
//
// actualize statistics
//
void CHTTPconn::updateStatistics(void)
{
#ifndef __DLLSA__
    if ( ! bCollectStat )
        return;

    // Update last statistic object
    // (only one request at a time for each connection)
    if (!statisticList.empty())
    {
        CStatistics &stats = statisticList.back();
        CTTime t;
        t.now();
        stats.setResponseTime(t);
        stats.setReplyCode(Document.getReplyCode());
        stats.setReplySize(Document.getNbBytesLoaded());
    }
#endif
}


//
// hasTimedOut
// ===========
//
// Cancel the request if time since last receive is too long
//
bool CHTTPconn::hasTimedOut(const CTTime &timeoutVal)
{
    if (timeoutVal.isNull())
        return false;
    CTTime t;
    t.now();
    if ( (t - Document.getLastReceiveTime()) > timeoutVal )
        return true;

    return false;
}




//
// cleanup
// =======
//
// Cleanup HTML document
//
inline void CHTMLDocument::cleanup(void)
{
    szData.erase(0, szData.size());
    // do not modify lastRequestType
    transferType = HTTP_TRANSFER_NONE;
    nbBytesLoaded = 0;
    lTotalLength = -1;
    bHeaderLoaded = false;
    bResponseLoaded = false;
    bLastChunk = false;
    sReplyCode = DEFAULT_REPLY_CODE;
    bSaveDocument = true; 
    lastReceiveTime.now();
}


//
// remove1XXResponse
// =================
//
// Erase a 1xx response from HTML document
//
void CHTMLDocument::remove1XXResponse(void)
{
    int endHeaderSize = strlen("\r\n\r\n");

    int endHeader = szData.find("\r\n\r\n");
    if (endHeader == string::npos)
    {
        endHeader = szData.find("\n\n");
        endHeaderSize = strlen("\n\n");
    }

    if (endHeader == string::npos)
    {
        // Trying to remove an incomplete header !
        // Remove all
        szData.erase();
        nbBytesLoaded = 0;
    }
    else
    {
        szData.erase(0, endHeader + endHeaderSize);
        nbBytesLoaded -= (endHeader + endHeaderSize);
    }

    // do not modify lastRequestType
    transferType = HTTP_TRANSFER_NONE;
    lTotalLength = -1;
    bHeaderLoaded = false;
    bResponseLoaded = false;
    bLastChunk = false;
    sReplyCode = 0;
//    bSaveDocument = true;
}


//
// append
// ======
//
// Append new datas to HTML document
//
void CHTMLDocument::append(const char* cs, int nData)
{
    // Keep document in memory only if it is HTML
    // (or if we cannot identify its type)
    if ( (bSaveDocument) || (!isHeaderLoaded()) )
        szData.append(cs, nData);
    else
        // Problem with chunks: we don't want to keep the document
        // but we need each chunk size, datas are appended here and
        // will be removed later in removeChunkInformation()
        // (only one chunk is saved)
        if (getTransferType() == HTTP_TRANSFER_CHUNK)
            szData.append(cs, nData);

    nbBytesLoaded += nData;
}


//
// setTransferType
// =======
//
// Extract information from header to set the transfer type
//
void CHTMLDocument::setTransferType(void)
{
    int CLPos = 0; // position of Content-Length field in header
    int TCPos = 0; // position of Transfer-Coding field in header

    // Set defaukt transfer type
    transferType = HTTP_TRANSFER_CLOSE_SOCKET;

    // If a Content-Length field is specified
    if ( (CLPos = find_in_string(szData, "Content-Length", false, 0)) != string::npos )
    {
        transferType = HTTP_TRANSFER_LENGTH_SPECIFIED;
        // determine the total length of the response (Content-Length + header length)
        int headerSize;
        if ( (headerSize = szData.find("\r\n\r\n")) != string::npos )
            headerSize += 4;
        else if ( (headerSize = szData.find("\n\n")) != string::npos )
            headerSize += 2;

        lTotalLength = headerSize + atoi(getHeaderValue("Content-Length").c_str());
    }

    // If a Transfer-Coding field is specified
    if ( ((TCPos = find_in_string(szData, "Transfer-Encoding", false, 0)) != string::npos)
            && ( find_in_string(getHeaderValue("Transfer-Encoding"), "chunked", false, 0) != string::npos)
            && ((CLPos == string::npos) || (CLPos > TCPos))
       )
    {
        // if there is no Content-Length field or if it appears
        // after the Transfer-Coding one
        transferType = HTTP_TRANSFER_CHUNK;

        int headerSize;
        if ( (headerSize = szData.find("\r\n\r\n")) != string::npos )
            headerSize += 4;
        else if ( (headerSize = szData.find("\n\n")) != string::npos )
            headerSize += 2;

        // Set lTotalLength to header size, it will be incremented
        // when we receive each chunk
        lTotalLength = headerSize;
    }
}


//
// setHeaderLoaded
// =======
//
// Header is comletely loaded, extract reply code from first line
// and look at the content-type header field to check if we have to
// save the document or not.
// Set transferType to the correct value
//
void CHTMLDocument::setHeaderLoaded(bool b)
{
    if ( (bHeaderLoaded = b) == false )
        return;

    // Header is loaded, extract the reply code
    string value;
    long beginReply;
    long beginCode;
    long endCode;

    if (((beginReply = find_in_string(szData, "HTTP", false, 0)) == string::npos)
            || ((beginCode = szData.find(" ", beginReply) + 1) == string::npos)
            || ((endCode = szData.find(" ", beginCode)) == string::npos)
       )
        return;

    value = szData.substr(beginCode, endCode - beginCode);
    sReplyCode = atoi(value.c_str());


/*    // Header is loaded, look at the document type (we have to keep it if it is HTML)
    long CTPos = find_in_string(szData, "Content-Type", false, 0);

    // if there is a Content-Type field and "html" word is not specified
    // we don't have to save the document
    if ( (CTPos != string::npos)
            &&(find_in_string(getHeaderValue("Content-Type"), "html", false, 0) == string::npos))
        bSaveDocument = false;
    else
        bSaveDocument = true;
*/

    setTransferType();
}


//
// getHeaderValue
// =======
//
// Extract a field value from header
//
string CHTMLDocument::getHeaderValue(const string& field, unsigned int index /* = 1*/) const
{
    string value;
    unsigned int i = 0;

    long endHeader = szData.find("\r\n\r\n");
    if (endHeader == string::npos)
        endHeader = szData.find("\n\n");
    long beginField = 0;

    string szHead = szData.substr(0, endHeader + 4);

    // Loop through header
    while (i < index)
    {
        beginField = find_in_string(szHead, field, false, beginField + 1);
        i++;
    }

    if ( (endHeader == string::npos) || (beginField == string::npos) || (beginField > endHeader) )
        return "";

    long beginValue = szHead.find(":", beginField);
    if (beginValue == string::npos)
        return "";
    beginValue = szHead.find_first_not_of(" ", beginValue + 1);
    long endValue = szHead.find_first_of("\r\n", beginValue);
    if (endValue == string::npos)
        return "";

    value = szData.substr(beginValue, endValue - beginValue);

    return value;
}

string CHTMLDocument::getHeaderValue(unsigned int size) const
{
    int endHeader = szData.find("\r\n\r\n");
    if (endHeader == string::npos)
        endHeader = szData.find("\n\n");

    if ( (endHeader != -1) && (endHeader < size) )
        return szData.substr(0, endHeader);
    else
        return szData.substr(0, size);
}

//
// Find a cookie value.
// This function expects there is only one cookie per "Set-Cookie" line in the header
// (ie: GET ...
//      Set-Cookie: name1 = value1
//      Set-Cookie: name2 = value2
// )
//
string CHTMLDocument::getHeaderValue(const string& field, const string& subfield) const
{
    string szSubfield = subfield;
    szSubfield.insert(0, "^");
    int i = 0;
    while ( (i = szSubfield.find_first_of(".", i)) != string::npos )
    {
        szSubfield.insert(i, "\\");
        i += 2;
    }
    i = 0;
    while ( (i = szSubfield.find_first_of("*", i)) != string::npos )
    {
        szSubfield.insert(i, ".");
        i += 2;
    }
    i = 0;
    while ( (i = szSubfield.find_first_of("?", i)) != string::npos )
    {
        szSubfield.replace(i, 1, ".");
        i += 2;
    }

    // Add '=' at end of cookie's name to differentiate cookies like "foo"
    // from those like "foofoo"
    szSubfield.append("=");


    CRegExp regExp;
    regExp.RegComp( szSubfield.c_str() );

    string value;
    string subValue;
    int beginSubField = string::npos;
    int endSubField;
    int len = 0;
    i = 1;

    do
    {
        value = getHeaderValue(field, i++).c_str();
    }
    while ( (value != "") && ((beginSubField = find_in_string(value, regExp, true, 0)) == -1) );

    // If subfield was not found in the header, show what can be found instead
    if (value == "")
    {
        return "";
    }

    endSubField = value.find_first_of(";\r\n", beginSubField);
    if (endSubField == string::npos)
        endSubField = value.size();

    subValue = value.substr(beginSubField, endSubField - beginSubField);

    return subValue;
}


//
// getBodyValue
// ============
//
// Get the body of document
//
string CHTMLDocument::getBodyValue(unsigned int size) const
{
    int endHeaderSize = strlen("\r\n\r\n");
    int endHeader = szData.find("\r\n\r\n");
    if (endHeader == string::npos)
    {
        endHeader = szData.find("\n\n");
        endHeaderSize = strlen("\n\n");
    }

    if (endHeader != -1)
        return szData.substr(endHeader + endHeaderSize, size);
    else
        return "";
}



//
// getBodyValue
// ============
//
// Get an element value in a HTML page
//
bool CHTMLDocument::getBodyValue(AddressQueryList & aql, Parser::AddressValueList & avl)
{
    int endHeaderSize = strlen("\r\n\r\n");
    int beginBody = szData.find("\r\n\r\n");
    if (beginBody == string::npos)
    {
        beginBody = szData.find("\n\n");
        endHeaderSize = strlen("\n\n");
    }
    beginBody += endHeaderSize;
	unsigned char * pBody = (unsigned char *)szData.c_str() + beginBody;

	PhLexer<unsigned char *>     lexer    (pBody);
	PhParser<unsigned char *>    parser   (lexer, TRUE);

	return parser.Parse(aql, avl) ;
}




//
// getChunkSize
// ============
//
// Extract the size of the new chunk
//
int CHTMLDocument::getChunkSize(void) const
{
    int chunkSize = -1;
    // lTotalLength point to the beginning of the new chunk
    long endSize = szData.find_first_of(" \r\n", lTotalLength);

    // Convert chunk size from hexadecimal to decimal
    if (endSize != string::npos)
        sscanf(szData.substr(lTotalLength, endSize).c_str(), "%x", &chunkSize);

    return chunkSize;
}


//
// removeChunkInformation
// ======================
//
// Remove the first line of each chunk since it is only length information
// (return -1 if this first line is not already loaded)
// Remove also the last chunk body if it is not an HTML document
// (each chunk is replaced by the new one)
//
int CHTMLDocument::removeChunkInformation(void)
{
    int endChunkInfo = szData.find("\r\n", lTotalLength);

    if (endChunkInfo != string::npos)
    {
        // Erase this chunk informations
        szData.erase(lTotalLength, endChunkInfo - lTotalLength + 2);
        // Erase last chunk if it is not HTML and decrease lTotalLength
        if (bSaveDocument == false)
        {
            int endHeaderSize = strlen("\r\n\r\n");
            int headerSize = szData.find("\r\n\r\n");
            if (headerSize == string::npos)
            {
                headerSize = szData.find("\n\n");
                endHeaderSize = strlen("\n\n");
            }
            headerSize += endHeaderSize;
            szData.erase(headerSize, lTotalLength - headerSize);
            lTotalLength = headerSize;
        }
        return 0;
    }
    else
        return -1;
}


//
// Ascii2Binary
// ============
//
// Convert an ascii body to its binary value or generate a body with given size
//
int Ascii2Binary(const string &inBuf, string &outBuf)
{
    // if inBuf begin with "binary("
    if (begins_with(inBuf, "BINARY("))
    {
        // extract the body length
        int beginLen = strlen("binary(");
        int endLen = inBuf.find_last_not_of(")", beginLen);
        string szLen = inBuf.substr(beginLen, endLen - beginLen);
        int length = atoi(szLen.c_str());
        outBuf.assign(0x41, length);
        return 0;
    }

    // if inBuf does not begin with "0x", its not a binary body
    if (!begins_with(inBuf, "0X"))
    {
        outBuf = inBuf;
        return 0;
    }

    int bufLen = inBuf.size();

    char tmp[2];
    int hexcode;
    for (int i = 2; i < bufLen; i += 2)
    {
        tmp[0] = inBuf[i];
        tmp[1] = inBuf[i+1];
        sscanf(tmp, "%x", &hexcode);
        outBuf.append(1, hexcode);
    }

    return 0;
}



unsigned short swaps(unsigned short s)
{
    return (unsigned short) ( ( (s >> 8) & 0x00FF) | ( (s << 8) & 0xFF00) ) ;
}

unsigned long swapl(unsigned long l)
{
    union
    {
        long    l ;
        short   s[2] ;
    } unLongSrc, unLongDst ;
    unLongSrc.l = l ;
    unLongDst.s[1] = swaps (unLongSrc.s[0]) ;
    unLongDst.s[0] = swaps (unLongSrc.s[1]) ;

    return unLongDst.l ;
}

unsigned int swapi(unsigned int i)
{
    union
    {
        int i ;
        short   s[2] ;
    } unIntSrc, unIntDst ;
    unIntSrc.i = i ;
    unIntDst.s[1] = swaps (unIntSrc.s[0]) ;
    unIntDst.s[0] = swaps (unIntSrc.s[1]) ;

    return unIntDst.i ;
}


unsigned short littles(unsigned short s)
{
#if MCH_ENDIAN == LITTLE_ENDIAN
    return s ;
#endif  // LITTLE_ENDIAN
#if MCH_ENDIAN == BIG_ENDIAN
    return swaps (s) ;
#endif  // BIG_ENDIAN
}


unsigned long littlel(unsigned long l)
{
#if MCH_ENDIAN == LITTLE_ENDIAN
    return l ;
#endif  // LITTLE_ENDIAN
#if MCH_ENDIAN == BIG_ENDIAN
    return swapl (l) ;
#endif  // BIG_ENDIAN
}


unsigned int littlei(unsigned int i)
{
#if MCH_ENDIAN == LITTLE_ENDIAN
    return i ;
#endif  // LITTLE_ENDIAN
#if MCH_ENDIAN == BIG_ENDIAN
    return swapi (i) ;
#endif  // BIG_ENDIAN
}


