/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.cpp
// -----------------
//=========================================================================

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif


#include  <strstream>
#include "TEHttpRecords.h"
#include "TEHttpRecords.inl"



static const char * HTTP_NONE       = "None";
static const char * HTTP_CONNECT    = "Connect";
static const char * HTTP_REQUEST    = "Request";
static const char * HTTP_DISCONNECT = "Disconnect";
static const char * HTTP_RESPONSE   = "Response";
static const char * HTTP_SYNCHRONIZE= "Synchronize";
static const char * HTTP_WAIT       = "Wait";

//
// TRcrdConnect::prepare
// =====================
//
// perform any preparation before conecting the threads
int TRcrdConnect::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prm = param("server");
    if(prm == NULL)
        return -1;
    string l_data = prm->strVal();
    _exprP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_exprP->Parse(l_data, 0)))
    {
        MSGERR("Connect : " + l_data +
               " Failed during 'server' PREPARE with: " + _exprP->Reason());
        return -1;
    }
    return 0;
}

//
// TRcrdConnect::execute
// =====================
//
// execute the threads in turn
int TRcrdConnect::execute(TContext& ctxt)
{
    ctxt.set_phase_text("Starting Connect execute");

    // Get the variables from the tof
    int         nConid      =   intPrmVal("conid", ctxt);
    string      szServer    =   "";
    {
        CTLockGuard<CTMutex> guard(m_mutex);
        szServer=_exprP->Evaluate(ctxt);
    }

    ctxt.log("CONNECT(" + int_to_string(nConid) + "): Server: " + szServer, TRACEMSG);


    // ------------------------------
    // Get connection object, create new if necessary.
    // If it already exists, receive all remaining datas, disconnect and reconnect
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if(pConnection == NULL)
    {
        string statLevel;
        try
        {
            statLevel = ctxt.get_variable(STAT_LEVEL);
        }
        catch(mdbParseException &)
        {
            statLevel = "L";
        }

        pConnection = new CHTTPconn(nConid, ctxt.GetThreadID(), statLevel);
        if(pConnection == NULL)
        {
            // print statistics before exiting
            CHTTPconn::printAllStatistics(ctxt);

            ctxt.log("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object", ERRMSG);
            throw mdbParseException("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object");
        }
        // Add new connection to list
        ctxt.addCon(nConid, pConnection);
    }
    else
    {
        // Get all remaining data on this connection
        if ( !(pConnection->isReceiveFinished()) )
        {
            if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
            {
                return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
            }
        }

        // Disconnect this connection object
        if( pConnection->disconnect() != 0 )
        {
            // If error
            ctxt.log("Disconnect failed:" + pConnection->getLastError(), ERRMSG);
        }
    }
    // ------------------------------


    // ------------------------------
    // Perform connection
    string szHost;
    int nPort = 80;
    int endServer = szServer.find(":");
    if (endServer != string::npos)
    {
        szHost = szServer.substr(0, endServer).c_str();
        nPort = atoi(szServer.substr(endServer + 1, szServer.size() - endServer - 1).c_str());
    }

    if((pConnection->connect(szHost, nPort)) != 0)
    {
        string lastError = pConnection->getLastError();
        // Log client error
        ctxt.log("Connection Failed:" + lastError, ERRMSG);
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);
        // Remove connection again
        ctxt.remCon(nConid);

        throw mdbParseException("Connection Failed:" + lastError);
    }
    // ------------------------------

    // --- A Connect object implies that connection was made through a proxy
    pConnection->setProxy(true);

    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}

//
// TRcrdRequest::prepare
// =====================
//
// perform any preparation before sending the request
int TRcrdRequest::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prm = param("uri");
    if(prm == NULL)
        return -1;
    string l_uri = prm->strVal();
    _uriP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_uriP->Parse(l_uri, 0)))
    {
        MSGERR("Request : " + l_uri +
               " Failed during 'uri' PREPARE with: " + _uriP->Reason());
        return -1;
    }


    prm = param("header");
    if(prm == NULL)
        return -1;
    string l_header = prm->strVal();
    _headerP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_headerP->Parse(l_header, 0)))
    {
        MSGERR("Request : " + l_header +
               " Failed during 'header' PREPARE with: " + _headerP->Reason());
        return -1;
    }


    prm = param("with");
    if(prm == NULL)
        return -1;
    string l_with = prm->strVal();
    _withP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_withP->Parse(l_with, 0)))
    {
        MSGERR("Request : " + l_with +
               " Failed during 'with' PREPARE with: " + _withP->Reason());
        return -1;
    }


    prm = param("wout");
    if(prm == NULL)
        return -1;
    string l_wout = prm->strVal();
    _woutP = new TtValList(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_woutP->Parse(l_wout, 0)))
    {
        MSGERR("Request : " + l_wout +
               " Failed during 'wout' PREPARE with: " + _woutP->Reason());
        return -1;
    }


    prm = param("body");
    if(prm == NULL)
        return -1;
    string l_body = prm->strVal();
    _bodyP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_bodyP->Parse(l_body, 0)))
    {
        MSGERR("Request : " + l_body +
               " Failed during 'body' PREPARE with: " + _bodyP->Reason());
        return -1;
    }


	prm = param("conid");
    string l_conid = prm->strVal();
	_iConid = atoi(l_conid.c_str());


    return 0;
}

//
// TRcrdRequest::getRequestType
// ============================
//
// translate the type into the actual Command string
string TRcrdRequest::getRequestString(int eType)
{
    string szCommand;

    switch (eType)
    {
    case HTTP_REQUEST_GET:
        {
            szCommand = "GET";
            break;
        }
    case HTTP_REQUEST_HEAD:
        {
            szCommand = "HEAD";
            break;
        }
    case HTTP_REQUEST_POST:
        {
            szCommand = "POST";
            break;
        }
    default:
        {
            szCommand = "GET";
        }
    }

    return szCommand;
}

//
// TRcrdRequest::execute
// =====================
//
// execute the request
int TRcrdRequest::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_REQUEST);

    // --- Get the variables from the tof
    int     nType       =   intPrmVal("type", ctxt);
    string  szUri       =   "";
    int     nConid      =   intPrmVal("conid", ctxt);
    string  szBody      =   "";
    string  szTimer     =   strPrmVal("timer", ctxt);
    string  szStatus    =   strPrmVal("retsts", ctxt);
    string  szCode      =   strPrmVal("retcod", ctxt);

    list<string> listHeader;
    list<string> listWithHeader;
    list<string> listWoutHeader;
    {
        CTLockGuard<CTMutex> guard(evaluateMtx);
        szUri       =   _uriP->Evaluate(ctxt);
        szBody      =   _bodyP->Evaluate(ctxt);
        _headerP->EvaluateValues(ctxt, listHeader);
        _withP->EvaluateValues(ctxt, listWithHeader);
        _woutP->EvaluateValues(ctxt, listWoutHeader);
    }

    ctxt.log("REQUEST(" + int_to_string(nConid) + "): "
             + getRequestString(nType) + " " + szUri, TRACEMSG);

    // --- Is it a http or a https request ?
    bool secureRequest = (begins_with(szUri, "HTTPS://"))
                         ? true
                         : false;

    // ------------------------------
    // Get connection object, create new if necessary (if there is no proxy)
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));

    if(pConnection == NULL)
    {
        string statLevel;
        try
        {
            statLevel = ctxt.get_variable(STAT_LEVEL);
        }
        catch(mdbParseException &)
        {
            statLevel = "L";
        }

        pConnection = new CHTTPconn(nConid, ctxt.GetThreadID(), statLevel);
        if(pConnection == NULL)
        {
            // print statistics before exiting
            CHTTPconn::printAllStatistics(ctxt);

            ctxt.log("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object", ERRMSG);
            throw mdbParseException("Connection alloc failed(" + int_to_string(nConid) + ") for this Connect object");
        }
        // --- Add new connection to list
        ctxt.addCon(nConid, pConnection);


        // --- Is it a http or a https request ?
        int nPort;
        string RequestProtocol;
        if (secureRequest)
        {
            RequestProtocol = "https://";
            nPort = 443; // Default port for SSL
        }

        else
        {
            RequestProtocol = "http://";
            nPort = 80;
        }

        // --- Extract server name from request line
        int server_begin    = find_in_string(szUri, RequestProtocol, false, 0) + RequestProtocol.size();
        int server_end      = szUri.find_first_of(":/ ", server_begin);
        string server_name  = szUri.substr(server_begin, server_end - server_begin);

        // --- If port is specified
        if (szUri[server_end] == ':')
        {
            int port_begin  = server_end + 1;
            int port_end    = szUri.find_first_of("/ ", port_begin);
            string szPort   = szUri.substr(port_begin, port_end - port_begin);
            nPort           = atoi(szPort.c_str());
        }

        // --- Perform connection
        if((pConnection->connect(server_name, nPort)) != 0)
        {
            string lastError = pConnection->getLastError();
            // Log client error
            ctxt.log("Connection Failed:" + lastError, ERRMSG);
            // print statistics before exiting
            CHTTPconn::printAllStatistics(ctxt);
            // Remove connection again
            ctxt.remCon(nConid);

            throw mdbParseException("Connection Failed:" + lastError);
        }
    }
    // ------------------------------

    // ------------------------------
    // Get all remaining data on this connection
    if ( !(pConnection->isReceiveFinished()) )
    {
        if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }
    }
    // ------------------------------

    // If some variables names are here
    pConnection->setTimerName(szTimer);
    pConnection->setStatusName(szStatus);
    pConnection->setCodeName(szCode);

    // ------------------------------
    // Reconnect if necessary
    if (!pConnection->isConnected())
    {
        ctxt.log("Have to reconnect", TRACEMSG);
        if (pConnection->connect(pConnection->getHost(),pConnection->getPort()) != 0)
        {
            string lastError = pConnection->getLastError();
            // print statistics before exiting
            CHTTPconn::printAllStatistics(ctxt);

            ctxt.log("Connection Failed:" + lastError, ERRMSG);
            throw mdbParseException("Connection Failed:" + lastError);
        }
    }
    // ------------------------------


    if (secureRequest)
        pConnection->setProtocol(HTTP_SECURE);


    // ------------------------------
    // Convert body to binary if necessary
    string newBody;
    if (szBody.length())
    {
        Ascii2Binary(szBody, newBody);
    }
    // ------------------------------


    // ------------------------------
    // Format headers
    string szHeader;
    string szWithHeader;
    string szWoutHeader;
    szHeader.reserve(DEFAULT_HEADER_SIZE);

    list<string>::iterator j;
    for(j = listHeader.begin(); j != listHeader.end(); j++)
    {
        szHeader.append(*j);
        remove_trailing(szHeader, "\n");
        szHeader.append("\r\n");
    }

    for(j = listWithHeader.begin(); j != listWithHeader.end(); j++)
    {
        szWithHeader.append(*j);
        remove_trailing(szWithHeader, "\n");
        szWithHeader.append("\r\n");
    }

    for(j = listWoutHeader.begin(); j != listWoutHeader.end(); j++)
    {
        szWoutHeader.append(*j);
        remove_trailing(szWoutHeader, "\n");
        szWoutHeader.append("\r\n");
    }

    string_lf2crlf(szHeader);
    string_cr2crlf(szHeader);
    string_lf2crlf(szWithHeader);
    string_cr2crlf(szWithHeader);
    string_lf2crlf(szWoutHeader);
    string_cr2crlf(szWoutHeader);

    // Remove the "with" header fields
    int offset = 0;
    int endLine = 0;
    int beginField = 0;
    int endField = 0;
    while ( (offset != string::npos) && ((endField = szWithHeader.find_first_of(" :", offset)) != string::npos) )
    {
        beginField = find_in_string(szHeader, szWithHeader.substr(offset, endField - offset), false, 0);
        endLine = szHeader.find("\r\n", beginField);
        if ( (beginField != string::npos) && (endLine != string::npos) )
            szHeader.erase(beginField, endLine - beginField + 2);
        offset = szWithHeader.find("\r\n", offset);
        offset += 2;
    }

    // Add the "with" header fields
    szHeader.append(szWithHeader);


    // If it is a POST request, calculate the Content-Length
    if (nType == 2)
    {
        beginField = find_in_string(szHeader, "Content-Length", false, 0);
        endLine = szHeader.find("\r\n", beginField);
        if ( (beginField != string::npos) && (endLine != string::npos) )
            szHeader.erase(beginField, endLine - beginField + 2);
        // Add the new Content-Length
        char buf[30];
        sprintf(buf, "Content-Length: %d\r\n", newBody.size());
        szHeader.append(buf);
    }


    // And remove the "wout" ones
    offset = 0;
    endLine = 0;
    beginField = 0;
    endField = 0;
    while ( (offset != string::npos) && ((endField = szWoutHeader.find_first_of(" \r\n", offset)) != string::npos) )
    {
        beginField = find_in_string(szHeader, szWoutHeader.substr(offset, endField - offset), false, 0);
        endLine = szHeader.find("\r\n", beginField);
        if ( (beginField != string::npos) && (endLine != string::npos) )
            szHeader.erase(beginField, endLine - beginField + 2);
        offset = szWoutHeader.find("\r\n", offset);
        offset += 2;
    }

    // Add an empty line
    szHeader.append("\r\n");
    // ------------------------------


    // ------------------------------
    // Remove server name if there is no proxy
    if ( !pConnection->isProxy() )
    {
        int endServerName = szUri.find_first_of("/", strlen("http://") + 2) ;
        if (endServerName != string::npos)
            szUri.erase(0, endServerName);
        else
            szUri = "/";
    }
    // ------------------------------


    // ------------------------------
    // Do the request
	ctxt.StartTimer(pConnection->getTimerName());

    if ((pConnection->request(nType, getRequestString(nType) + " " + szUri + "\r\n" + szHeader, newBody, _bSaveDocument) == -1))
    {
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);

        // Log client error
        ctxt.log("Request Failed:" + pConnection->getLastError(), ERRMSG);
        throw mdbParseException("Request Failed:" + pConnection->getLastError());
        // return -1;
    }
    // ------------------------------


    // ------------------------------

    if ( (pConnection->getCodeName() != "") || (pConnection->getStatusName() != "") )
    {
        if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, 0) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }
        // CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, 0);
    }
    // ------------------------------



    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}



//
// TRcrdDisconnect::prepare
// ========================
//
// perform any preparation before disconnecting
int TRcrdDisconnect::prepare(TScript* script, int objIndex, bool verbose)
{
    return 0;
}

//
// TRcrdDisconnect::execute
// ========================
//
// execute the disconnect
int TRcrdDisconnect::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_DISCONNECT);

    // Get the variables from the tof
    unsigned nConid = intPrmVal("conid", ctxt);
    unsigned nCancel = intPrmVal("options", ctxt);

    ctxt.log("DISCONNECT(" + int_to_string(nConid) + ")", TRACEMSG);


    // ------------------------------
    // If there is no connection Id, disconnect all connections
    if (nConid == 0)
    {
        if (nCancel == 0)
        {
            if (CHTTPconn::receiveOnAllConnections(ctxt, 0, NULL) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
            {
                return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
            }
        }

        // Get connections map
        cons_map_t &conMap = ctxt.get_connections();
        cons_map_t::iterator j;

        // Disconnect all connections
        for(j = conMap.begin(); j != conMap.end(); j++)
        {
            (static_cast<CHTTPconn*>((*j).second))->disconnect();
        }

        // Clear the map (destroy each element)
        while (!conMap.empty())
        {
            j = conMap.begin();
            ctxt.remCon((*j).first);
        }
        conMap.clear();

        ctxt.set_phase_text(HTTP_NONE);
        return 0;
    }
    // ------------------------------


    // ------------------------------
    // Get connection object
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if( pConnection == NULL )
    {
        ctxt.log("Invalid connection id(" + int_to_string(nConid) + ") for disconnect", ERRMSG);
        ctxt.set_phase_text(HTTP_NONE);
        return 0; // Safe to carry on
    }
    // ------------------------------


    // ------------------------------
    // Get all remaining data on this connection

    if ( !(pConnection->isReceiveFinished()) && (nCancel == 0) )
    {
        if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }
        // CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection);
    }
    // ------------------------------


    // ------------------------------
    // Disconnect and remove this connection object from map

    if( pConnection->disconnect() != 0 )
    {
        ctxt.log("Disconnect failed:" + pConnection->getLastError(), ERRMSG);
    }

    ctxt.remCon(nConid);
    // ------------------------------

    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}


//
// TRcrdHttpResponse::prepare
// ========================
//
// perform any preparation before reading the response
int TRcrdHttpResponse::prepare(TScript* script, int objIndex, bool verbose)
{
	const TParam* prmP = param("data");
	_responseType = prmP->intVal();

	prmP = param("conid");
    string l_conid = prmP->strVal();

	_iConid = atoi(l_conid.c_str());
	for (int i=0; i<l_conid.size(); i++)
	{
		if (!isdigit(l_conid[i]))
		{
			_iConid = -1 ;
			break ;
		}
	}

    // Get variable string and expression string from 'name' parameters.
    prmP = param("into");
    string l_name = prmP->strVal();

    // Create variable reference object, for variable to be set.
    _varP = new TtVarRef(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_varP->Parse(l_name, 0)))
    {
        MSGERR("SET VARIABLE: " + l_name + " Failed during 'into' PREPARE with: " + _varP->Reason());
        return -1;
    }

    // Get pointer to TRecord for variable - as we need information about the variable
    // to process the expression string.
    TRecord* trecP = (*script).GetVardef(_varP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
        MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable '" + _varP->Varnam() + "' not defined");
        return -1;
    }

    // Get variable options mask so that we can check the variable type and so
    // determine what type of expression we need (integer or character).
    prmP = trecP->param(TParam::stOptions);
    if (prmP == NULL) // If no 'options'  parameter (should always be present).
    {
        MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Variable options missing");
        return -1;
    }

    int options = prmP->intVal();   // Variable options mask

    // Create an expression object (of the appropriate type)
    if (_exprP) delete _exprP;  // Delete any existing object first


    if (options & VTCHAR)
    {
        _exprP = new TtChrExpr(script);
    }
    else if (options & VTINT)
    {
        _exprP = new TtIntExpr(script);
    }
    else // Inrecognised variable type.
    {
        MSGERR("SET VARIABLE: " + l_name + " Failed during PREPARE with: Unrecognised variable type");
        return -1;
    }

    // Keep the variable size
    prmP = trecP->param(TParam::stSize);
    if (prmP)
        _varSize = prmP->intVal();

    prmP = param("identifier");
    if(prmP == NULL)
        return -1;
    string l_ident = prmP->strVal();
    _identP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_identP->Parse(l_ident, 0)))
    {
        MSGERR("Connect : " + l_ident +
               " Failed during 'identifier' PREPARE with: " + _identP->Reason());
        return -1;
    }

	if (_responseType == HEADER_VAR)
		return 0 ;

	// if the conid is -1, build a new map.
	if (_iConid == -1)
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}
	else
	{
		int newIndex = objIndex;
		while (newIndex > 0)
		{
			newIndex--;
			TRecord * prevRecord = script->rec_at_index(newIndex) ;
			if (prevRecord && prevRecord->isType(TRcrdRequest::KEY))
			{
				TRcrdRequest *prevRequestRecord = (TRcrdRequest *)prevRecord;
				if ( (prevRequestRecord->_iConid == _iConid) || (prevRequestRecord->_iConid == 0) )
				{
					prevRequestRecord->_bSaveDocument = true;
					break;
				}
			}
		}

		

		// Try to find a previous Response Record with the same conid 
		// and the same responseType
		while (objIndex > 0)
		{
			objIndex -- ;
			TRecord * prevRecord = script->rec_at_index(objIndex) ;
			// check that the previous record is also a TRcrdHttpResponse Record
			// with type==BODY
			if (prevRecord->isType(KEY))
			{
				TRcrdHttpResponse * prevResponseRecord = (TRcrdHttpResponse *)prevRecord ; 
				if (prevResponseRecord->_iConid == _iConid && 
					prevResponseRecord->_responseType == _responseType)
				{
					_addressQueryList = prevResponseRecord->_addressQueryList ;
					_varAddressMap = prevResponseRecord->_varAddressMap ;
					prevResponseRecord->_addressQueryList.clear() ;
					prevResponseRecord->_varAddressMap.clear() ;
					break ;
				}
			}
			else 
				break ;
		}
	}

	// If the address begins by " and finishes by " and 
	// there are no other " in the address, remove the "
	// else has to evaluate the address in execute later
	string address = l_ident ;
	if (address[0] == '"' && address.find("\"", 1) == address.size() - 1)
		address = address.substr(1, address.size() - 2);
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
		return 0 ;
	}

	if (!address.empty())
	{
		AddressQuery * aq = new AddressQuery(l_name, address.c_str());
		// Normally, it should be a valid address, if not, will be evaluated in
		// the next
		if (aq->IsValidAddress())
		{
			_addressQueryList.push_back(aq) ;
			delete _identP ;
			_identP = NULL ;
		}
		else
		{
			delete aq ;
			_varAddressMap[l_name] = _identP ;
			_identP = NULL ;
			return 0 ;
		}
	}
	else
	{
		_varAddressMap[l_name] = _identP ;
		_identP = NULL ;
	}
    return 0;
}


//
// TRcrdHttpResponse::execute
// ========================
//
// read the response and set variables values
int TRcrdHttpResponse::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_RESPONSE);
	int ret ;

	// if both map are empty, return

    int nConid      = intPrmVal("conid", ctxt);
    int nType       = intPrmVal("data", ctxt);
    string szName ;
	string szIdent ;

	if (nType == HEADER_VAR)
	{
		szName   = strPrmVal("into", ctxt);
		szIdent  = "";
		{
			CTLockGuard<CTMutex> guard(m_mutex);
			szIdent =_identP->Evaluate(ctxt);
		}
		ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: "
			 + szName + " Identity: " + szIdent, TRACEMSG);
	}
	else if (_addressQueryList.empty() && _varAddressMap.empty())
		return 0;


    // ------------------------------
    // Get connection object
    CHTTPconn* pConnection = static_cast<CHTTPconn*>(ctxt.connection(nConid));
    if(pConnection == NULL)
    {
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);

        ctxt.log("Invalid connection id(" + int_to_string(nConid) + ") for HttpResponse", ERRMSG);
        throw mdbParseException("Invalid connection id(" + int_to_string(nConid) + ") for HttpResponse");
        //return -1;
    }
    // ------------------------------

    // ------------------------------
    // Get header or all remaining data on this connection

    if ( !(pConnection->isReceiveFinished()) )
    {
        if (CHTTPconn::receiveOnAllConnections(ctxt, 0, pConnection, nType) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }
    }

	// Header variable
	if (nType == HEADER_VAR)
	{
		int ret = 0;
		// If szIdent is empty, put the document in variable regards to its size
		if (szIdent == "")
			ret = ctxt.set_variable(szName, pConnection->getHttpResponse(HEADER_VAR, _varSize));
		//
		else
			ret = ctxt.set_variable(szName, pConnection->getHeaderResponse(szIdent));
	}
	// Body variables
	else
	{
		// ------------------------------
		//
		Parser::AddressValueList avl;
		TtChrExpr  * pIdent ;

		AddressQueryList tmpAddressQueryList = _addressQueryList;
		AddressQueryList newAddressQueryList;

		// Evaluate the variable addresses and add fill an AddressQueryList
		for (variableAddressMap::iterator itr = _varAddressMap.begin(); itr != _varAddressMap.end(); itr ++)
		{
			pIdent = itr->second ;
			if (pIdent)
			{
				string address ;
				{
					CTLockGuard<CTMutex> guard(m_mutex);
					address =pIdent->Evaluate(ctxt);
				}
				// The address is empty, get all the document
				if (address.empty())
				{
					ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: "
							 + itr->first + " Identity: " + address, TRACEMSG);
					ret = ctxt.set_variable(itr->first, pConnection->getHttpResponse(BODY_VAR, _varSize));
					if(ret == -1)
					{
						// print statistics before exiting
						CHTTPconn::printAllStatistics(ctxt);

						ctxt.log("Set variable failed(" + itr->first + ")", ERRMSG);
						throw mdbParseException("Set variable failed(" + itr->first + ")");
					}
					continue ;
				}
				// put the addressQuery in the addressQueryList.
				AddressQuery *newAddressQuery  = new AddressQuery(itr->first, address.c_str());
				newAddressQueryList.push_back(newAddressQuery) ;
				tmpAddressQueryList.push_back(newAddressQuery) ;
			}
		}


		// Use the HTML parser to  get the values associated with addresses.
		if (pConnection->getBodyResponse(tmpAddressQueryList, avl))
		{
			// assciate each addressValue to each variable.
			for (Parser::AddressValueList::iterator avlItr = avl.begin(); avlItr != avl.end(); avlItr ++)
			{
				// get the variable name associated with the AddressQuery
				const AddressQuery * aq = avlItr->pAq_ ;
				string vName = aq->strVariable_ ;	
				ctxt.log("HTTPRESPONSE(" + int_to_string(nConid) + "): Name: "
							 + vName + " Identity: " + (aq->address_).ToString(), TRACEMSG);
				ret = ctxt.set_variable(vName, string((const char *)(avlItr->pcBegin_), avlItr->ulSize_));
				if(ret == -1)
				{
					// print statistics before exiting
					CHTTPconn::printAllStatistics(ctxt);

					ctxt.log("Set variable failed(" + vName + ")", ERRMSG);
					throw mdbParseException("Set variable failed(" + vName + ")");
				}
			}
		}

		for (AddressQueryList::iterator newitr = newAddressQueryList.begin(); newitr != newAddressQueryList.end(); newitr ++)
		{
			delete *newitr;
		}
	}


    if(ret == -1)
    {
        // print statistics before exiting
        CHTTPconn::printAllStatistics(ctxt);

        ctxt.log("Set variable failed(" + szName + ")", ERRMSG);
        throw mdbParseException("Set variable failed(" + szName + ")");
    }
    // ------------------------------


    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}


//
// TRcrdSynchronize::prepare
// ========================
//
// prepare
int TRcrdSynchronize::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param("timeout");
    if (prmP == NULL)
        return 0;
    int tgtvar = prmP->intVal();

    if (tgtvar == 0)
        return 0;
    _labelIndex = PrepareGoto(this, TParam::stTmolab, script, true);
    return 0;
}


//
// TRcrdSynchronize::execute
// ========================
//
// read the response on all connections
int TRcrdSynchronize::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_SYNCHRONIZE);

    int nTimeout = intPrmVal("timeout", ctxt);


    ctxt.log("SYNCHRONIZE", TRACEMSG);


    CHTTPconn::enumReceiveCode retCode;

    if (nTimeout == 0)
        retCode = CHTTPconn::receiveOnAllConnections(ctxt, 0, NULL);
    else
        retCode = CHTTPconn::receiveOnAllConnections(ctxt, nTimeout, NULL, 0, true);

    if (retCode == CHTTPconn::HTTP_RECEIVE_EXIT_ON_LOCAL_TIMER)
    {
        ctxt.set_phase_text(HTTP_NONE);
        return ExecuteGoto(this, ctxt, TParam::stTmolab, &_labelIndex, &m_mutex);
    }
    else
        if (retCode == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
        {
            return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
        }

    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}


//
// TRcrdWait::execute
// ========================
//
// receive some data from each connection during an amount of time
int TRcrdWait::execute(TContext& ctxt)
{
    ctxt.set_phase_text(HTTP_WAIT);

    int nDelay = intPrmVal("tdelay", ctxt);

    ctxt.log("WAIT : Delay: " + int_to_string(nDelay), TRACEMSG);

    unsigned percent = ctxt.get_delay_percent();
    unsigned long realDelay = (nDelay*percent)/100 + (nDelay*percent)%100;

    if (CHTTPconn::receiveOnAllConnections(ctxt, realDelay, NULL) == CHTTPconn::HTTP_RECEIVE_EXIT_ON_GLOBAL_TIMER)
    {
        return ExecuteGotoLabel(this, ctxt, "ERR_LABEL");
    }

    ctxt.set_phase_text(HTTP_NONE);

    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// HELPER FUNCTIONS
// ================
///////////////////////////////////////////////////////////////////////////////
int server_specific_init(void)
{
    char* maxthr = getenv("TE_MAX_THREADS");
    int imaxthr;
    if(maxthr != NULL)
    {
        imaxthr = atoi(maxthr);
    }
    else
    {
        imaxthr = DEFAULT_THREAD_MAX;
    }

    CHTTPconn::statmtx = new CTMutex[(imaxthr/10) + 1];

	CHTTPconn::initHTMLParser();

    return 0;
}

// Perform any server specific one time cleanup
int server_specific_cleanup(void)
{
    if (CHTTPconn::statmtx)
        delete[] CHTTPconn::statmtx;

    return 0;
}
