/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     NT_Performance
// @module  NT Performance Class | This module contains the implementation of
//          the NT Performance class.  This class allows the retrieval of NT
//          Performance Statistics. Obviously this class will only work on a
//          Windows platform.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "OtaDefs.h"
#include "OtaNTPerf.h"
#include <CORBA/CORBAServer.h>
#include "TPLog.h"
#include <pdhmsg.h>
#include <winbase.h>

///////////////////////////////////////////////////////////////////////////////
#include <CECSimpleParse.h>
///////////////////////////////////////////////////////////////////////////////
// Constants
//const char* OtaNTPerform::CLASS_NAME = "OtaNTPerform";

// Globals
static HQUERY hQuery = NULL;      // Static query handle (only one required)

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor initialises the member variables and then opens a
//          performance query.  This query will be used throughout the lifetime
//          of the object.  The constructor is also responsible for registering
//          the object with the NameService.
// @parm    Pointer to CORBA helper functionality.
// @parm    Name of the object.
OtaNTPerform::OtaNTPerform()
{
    // Error information
    const char* where = "OtaNTPerform::OtaNTPerform";

    // Open query if we don't already have one
    if (hQuery == NULL)
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Opening NT Performance query");

        // Open performance query - assume success
        PDH_STATUS pdhStatus = ERROR_SUCCESS;

        try
        {
            // Make the query
            pdhStatus = PdhOpenQuery( NULL, 0, &hQuery );
            if ( pdhStatus != ERROR_SUCCESS )
                throw CyrQueryException( "PdhOpenQuery failed", where, pdhStatus );
        }
        catch (...)
        {
            // And throw an exception if not possible - as we may be finishing
            throw CyrQueryException( "PdhOpenQuery failed", where, pdhStatus );
        }
    }

    m_initialized = false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor removes the counters from the query
OtaNTPerform::~OtaNTPerform()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method executes the passed string as a Windows NT Performance
//          Query.  This 'counter' is added to the open query, executed and
//          then removed.
// @rdesc   true if query successful.
// @parm    String representing the requested Performance Query.
// @parm    The result of the query.

bool OtaNTPerform::RunSimpleQuery( const string query, long *result )
{
    PDH_STATUS pdhStatus;
    HCOUNTER hCounter;
    if (!m_initialized)
    {
        // Add counter to query
        try
        {
            pdhStatus = PdhAddCounter( hQuery, query.c_str(), NULL, &hCounter );
            if ( pdhStatus != ERROR_SUCCESS )
            {
                LogPDHError("PdhAddCounter", query, pdhStatus);
                return false;
            }
        }
        catch (...)
        {
            OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhAddCounter");
            return false;
        }
        m_hCounters.push_back(hCounter);
        m_initialized = true;
    }

    // Collect data
    try
    {
        // add a sleep to allow the CPU performance statistics
        // to settle down
        Sleep(1);   // milliseconds

        pdhStatus = PdhCollectQueryData( hQuery );
        if ( pdhStatus != ERROR_SUCCESS )
        {
            LogPDHError("PdhCollectQueryData", query, pdhStatus);
            return false;
        }
    }
    catch (...)
    {
        OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhCollectQueryData");
        return false;
    }

    // Get formatted data
    PDH_FMT_COUNTERVALUE value;
    try
    {
        pdhStatus = PdhGetFormattedCounterValue( m_hCounters[0],
                    PDH_FMT_LONG, NULL, &value );
        if ( pdhStatus != ERROR_SUCCESS )
        {
            LogPDHError("PdhGetFormattedCounterValue", query, pdhStatus);
            return false;
        }
    }
    catch (...)
    {
        OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhGetFormattedCounterValue");
        return false;
    }

    // Return result
    *result = value.longValue;
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Like the RunSimpleQuery method, this method executes a Windows NT
//          performance query.  The return string however is formatted for
//          output to a detailed statistics file.
//          The query string is split into individual counters and added to the
//          query.
// @rdesc   true if query successful.
// @parm    The detailed query string.
// @parm    A formatted results string (including newline characters).
bool OtaNTPerform::RunDetailedQuery( const string query, string *result )
{
    PDH_STATUS pdhStatus;
    HCOUNTER hCounter;
    if (!m_initialized)
    {
        // Parse query string into individual counters
        CECSimpleParse token( query);
        string entry = token.getNext( '\n' );
        while( !entry.empty() )
        {
            // Remove any leading/trailing whitespace
            while(entry[0]==' '||entry[0]=='\r')entry=entry.erase(0,1);
            while(entry[entry.size()]==' '||entry[entry.size()]=='\r')entry=entry.erase(entry.size(),1);


            // If we still have an entry
            if ( entry.length() > 0 )
            {
                // Add counter to query
                bool added = true;
                try
                {
                    pdhStatus = PdhAddCounter( hQuery, entry.data(), NULL, &hCounter );
                    if ( pdhStatus != ERROR_SUCCESS )
                    {
                        LogPDHError("PdhAddCounter", query, pdhStatus);
                        added = false;
                    }
                }
                catch (...)
                {
                    OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhAddCounter");
                    added = false;
                }
                if (added)
                {
                    m_hCounters.push_back(hCounter);
                }
            }

            // Get next entry
            entry = token.getNext( '\n' );
        }


        m_initialized = true;
    }

    // Get current time and format it
    CTTime now;
    now.now();
    string res = now.Format( "2\t%Y%m%d %H:%M:%S\r\n3" );

    // Collect data
    try
    {
        pdhStatus = PdhCollectQueryData( hQuery );
        if ( pdhStatus != ERROR_SUCCESS )
        {
            LogPDHError("PdhCollectQueryData", query, pdhStatus);
            return false;
        }
    }
    catch (...)
    {
        OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhCollectQueryData");
        return false;
    }

    // Get formatted data
    PDH_FMT_COUNTERVALUE value;
    for ( int i = 0; i < m_hCounters.size(); i++ )
    {
        try
        {
            pdhStatus = PdhGetFormattedCounterValue( m_hCounters[i], PDH_FMT_LONG, NULL, &value );
            if ( pdhStatus != ERROR_SUCCESS )
            {
                LogPDHError("PdhGetFormattedCounterValue", query, pdhStatus);
                return false;
            }
        }
        catch (...)
        {
            OtaAddTrace(CYR_OTA,ERRORLEVEL,"Exception in PdhGetFormattedCounterValue");
            return false;
        }

        char buf[30];
        sprintf( buf, "\t%d", value.longValue);
        res += buf;
    }
    res += "\r\n";

    // Return result
    *result = res;
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a tab separated string containing a list of
//          column headers for the detailed query.
// @rdesc   The header string.
// @parm    The query that will produce the headers.
string OtaNTPerform::GetDetailedHeaders( const string query )
{
    string headers( "" );

    for ( int i = 0; i < m_hCounters.size(); i++ )
    {
        bool fRes = true;
        DWORD dwBufferSize;
        PPDH_COUNTER_INFO pInfo;
        string name;

        dwBufferSize = 2048;
        pInfo = (PPDH_COUNTER_INFO) LocalAlloc(LPTR, dwBufferSize);
        try
        {
            if ( ERROR_SUCCESS != PdhGetCounterInfo(m_hCounters[i],
                                                    FALSE, &dwBufferSize, pInfo))
            {
                fRes = false;
            }
        }
        catch (...)
        {
            fRes = false;
        }

        if (fRes)
        {
            name = pInfo->szCounterName;
        }
        else
        {
            name = "Unknown";
        }
        if (pInfo != NULL) LocalFree(pInfo);
        headers += name + "\t";
    }

    // Return the generated string
    return headers;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method logs a PDH error
// @parm    The PDH function name
// @parm    The query
// @parm    The error status
void OtaNTPerform::LogPDHError (const string func, const string query, PDH_STATUS error)
{
    string errtxt = "Error from " + func + " for " + query + ": ";
    char    pdherr[256];
    if (error == PDH_CSTATUS_BAD_COUNTERNAME)
    {
        strcpy (pdherr, "The counter name path string could not be parsed or interpreted.");
    }
    else if (error == PDH_CSTATUS_NO_COUNTER)
    {
        strcpy (pdherr, "The specified counter was not found.");
    }
    else if (error == PDH_CSTATUS_NO_COUNTERNAME)
    {
        strcpy (pdherr, "An empty counter name path string was passed in.");
    }
    else if (error == PDH_CSTATUS_NO_MACHINE)
    {
        strcpy (pdherr, "A machine entry could not be created.");
    }
    else if (error == PDH_CSTATUS_NO_OBJECT)
    {
        strcpy (pdherr, "The specified object could not be found.");
    }
    else if (error == PDH_FUNCTION_NOT_FOUND)
    {
        strcpy (pdherr, "The calculation function for this counter could not be determined.");
    }
    else if (error == PDH_INVALID_ARGUMENT)
    {
        strcpy (pdherr, "One or more arguments are invalid.");
    }
    else if (error == PDH_INVALID_HANDLE)
    {
        strcpy (pdherr, "The query handle is not valid.");
    }
    else if (error == PDH_INVALID_DATA)
    {
        strcpy (pdherr, "The specified counter does not contain valid data or a successful status code.");
    }
    else if (error == PDH_MEMORY_ALLOCATION_FAILURE)
    {
        strcpy (pdherr, "A memory buffer could not be allocated.");
    }
    else if (error == PDH_NO_DATA)
    {
        strcpy (pdherr, "The query does not currently have any counters.");
    }
    else
    {
        sprintf (pdherr, "Unknown error %d", error);
    }
    errtxt += pdherr;
    OtaAddTrace(CYR_OTA,ERRORLEVEL,errtxt);
}
