/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc
// @module  OtaQuery | This module contains the implementation of the
//          OtaQuery class.  This class is used to store details about an
//          OTA query.  these details include executor, the query string and
//          any results generated.
//
///////////////////////////////////////////////////////////////////////////////

#include "OtaQuery.h"
#include "OtaDefs.h"
#include <CyrHelper.inl>
#include "TPLog.h"

const char * OtaQuery::PARAM_DEBUG  = "TraceLevel";
const char * OtaQuery::PARAM_DETAIL  = "Details";
const char * OUTPUT_TYPE  = "OtaStat.txt";
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor simply assigns member variables and allocates
//          memory for a simple or detailed query object.
// @parm    The string to be used in the query.
// @parm    The description for the query.
// @parm    The interval between queries.
// @parm    True if query is a difference query.
// @parm    True if a detailed query.
// @parm    Query type.
// @parm    Reference to a query target.
// @parm    Reference to a query handler.
OtaQuery::OtaQuery( CECOrbTestRun* pTestRun, string query, string nodename, string desc,
                    int interval, bool delta, bool detail, int querytype,
                    OtaTarget* pTarget, OtaQueryHandler* pHandler )
{
    static int index = 0;

    // Assign member variables
    m_node = nodename;
    m_query = query;
    m_desc = desc;
    m_interval = interval;
    m_delta = delta;
    m_detail = detail;
    m_type = querytype;
    m_pTarget = pTarget;
    m_pHandler = pHandler;
    m_time.now();
    m_active = false;
    m_disabled = false;
    m_savedLastResult = false;

    m_dQuery = NULL;
    m_sQuery = NULL;
    m_sHeader = NULL;

    // Create either a simple or detailed query object
    if ( detail == false )
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Creating new OTA simple query");
        m_dQuery = NULL;

        // Add a header to the test run results
        m_index = index;
        AddResultHeader( pTestRun, nodename, desc );

        // Increment the index ready for the next header
        index += 1;
    }
    else
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Creating new OTA detail query");
        m_sQuery = NULL;
        m_dQuery = new OtaDetailQuery( query, desc );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor simply tidies up the memory allocated in the
//          constructor and has no other tasks.
OtaQuery::~OtaQuery()
{
    // Tidy memory
    if ( m_dQuery != NULL )
    {
        delete m_dQuery;
    }
    if ( m_sQuery != NULL )
    {
        delete m_sQuery;
    }
    if ( m_pHandler != NULL )
    {
        delete m_pHandler;
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the actual query string for this object.
// @rdesc   The query string.
string OtaQuery::GetQuery()
{
    return m_query;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description string for this object.
// @rdesc   The description string.
string OtaQuery::GetDescription()
{
    return m_desc;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the handler for this object.
OtaQueryHandler* OtaQuery::GetHandler()
{
    return m_pHandler;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the target for this object.
OtaTarget* OtaQuery::GetTarget()
{
    return m_pTarget;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the requested interval between queries.
// @rdesc   The interval value in seconds.
long OtaQuery::GetInterval()
{
    return m_interval;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the last time the query was executed.
// @rdesc   The last time the query was executed.
CTTime OtaQuery::GetLastTimestamp()
{
    return m_time;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns true of the query is a difference query.
//          Difference queries store the delta between results instead of the
//          actual result.
// @rdesc   True if the query is a difference query, otherwise false.
bool OtaQuery::IsDifferenceQuery()
{
    return m_delta;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns true if the query is a detailed query.  If so
//          the results will be strings and not integers.
// @rdesc   True if the query is a detailed one, otherwise false.
bool OtaQuery::IsDetailQuery()
{
    return (m_dQuery != NULL);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a detiled result to the array of results.  It also
//          updates the time when the query was last executed.
// @parm    The start time of the query in seconds.
// @parm    The stop time of the query in seconds.
// @parm    The result of the query.
void OtaQuery::AddResult( long start, long stop, const char* result )
{
    // Set time of last check to be the time when the query stopped
    m_time = CTTime( stop );

    // Store result
    m_dQuery->AddResult( result );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a simple result header to the test output.
void OtaQuery::AddResultHeader( CECOrbTestRun* pTestRun, string nodename, string desc )
{
    m_sHeader = new CECOtaHeader( nodename.c_str(), m_index, desc.c_str() );
    StoreResults(pTestRun);
    delete m_sHeader;
    m_sHeader = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a simple result to the array of results.  It
//          determines if the query is a diff query and calculates the correct
//          value automatically.  The method also updates the time when the
//          query was last executed.
// @parm    The start time of the query in seconds.
// @parm    The stop time of the query in seconds.
// @parm    The result of the query.
void OtaQuery::AddResult( CECOrbTestRun* pTestRun, long start, long stop, long result )
{
    // Set time of last check to be the time when the query stopped
    m_time = CTTime( stop );

    // If a difference query then calculate result
    int actual = result;
    if ( IsDifferenceQuery() )
    {
        // If first time, just store value for subsequent delta
        if (!m_savedLastResult)
        {
            m_lastResult = result;
            m_savedLastResult = true;
            return;
        }
        else
        {
            actual -= m_lastResult;
            m_lastResult = result;
        }
    }

    // Store result in test output object
    m_sQuery = new CECOtaSimpleQuery( m_node.c_str(), m_index, start, stop, actual );
    StoreResults(pTestRun);
    delete m_sQuery;
    m_sQuery = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Whatever the type of query, this method will store any generated
//          results into the current test run object.
// @parm    The test run that will receive the results.
void OtaQuery::StoreResults( CECOrbTestRun* pTestRun )
{
    // If the query is a simple query then simply add the CECTestOutput data
    if ( m_sQuery != NULL )
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Storing simple test results");
        pTestRun->AddTestOutput( m_sQuery );
    }
    // If it is a header then simply add the CECTestOutput data
    else if ( m_sHeader != NULL )
    {
        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Storing OTA Header");
        pTestRun->AddTestOutput( m_sHeader );
    }
    else
    {
        // Get headers (note that we do this here rather than when creating query,
        // as individual elements of an NT query may fail to be added, so the
        // headers will correspond only to those elements which are valid)
        string headers = m_pHandler->GetDetailedHeaders(m_query);
        m_dQuery->SetHeaders(headers);

        OtaAddTrace(DBGSTD,DEBUGLEVEL,"Storing detail test results");
        // Create a temporary file and write data to file
        string filename = CyrGetTempFileName();
        m_dQuery->WriteToFile( filename.c_str() );

        // Calculate name of file within OpenSTARep .  This will be the standard
        // filename, plus a unique numeric value with an extension of .std
        long count, index = 0;
        char buf[30];
        string filelist = pTestRun->GetFiles( count );

        do
        {
            // Calculate name of file inOpenSTARep
            sprintf( buf, "OTADetail%d.std", index );
            index++;
        }
        while( filelist.find( buf )!=string::npos);

        // Create a new OpenSTARep entry with the contents of the file
        CECOrbFile* pFile;
        try
        {
            pFile = new CECOrbFile( pTestRun, buf );
        }
        catch(...)//assume this is beacuse it is not there and try and create a new one
        {
            pFile = new CECOrbFile( pTestRun, buf, "OTA Detailed Statistics" );
        }
        pFile->Lock();
        pFile->CheckIn( filename.c_str(), true );
        delete pFile;

        // Delete temporary file
        CyrDeleteFile( filename.c_str() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns true if the query object is currently active.
// @rdesc   True if the query is active else false.
bool OtaQuery::IsActive()
{
    return m_active;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows the query to be set active or not.
// @parm    True to set the active flag or false to de-activate.
void OtaQuery::SetActive( bool active )
{
    m_active = active;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows the query to be set disabled.
void OtaQuery::Disable()
{
    m_disabled = true;
}

// @mfunc   This method returns true if the query object is disabled.
// @rdesc   True if the query is disabled.
bool OtaQuery::IsDisabled()
{
    return m_disabled;
}

/////////////////////////////////////////////////////////////////////////////
// Log
///////////////////////////////////////////////////////////////////////////////
// @mfunc Log the query.
// @param
// @param
void    OtaQuery::Log ()
{
    string msg;
    static const char *types[] =
        {"SQL", "Exec", "NT Performance", "SNMP"
        };
    string type = types[m_type];
    char interval[32];
    sprintf(interval, "%d", m_interval);
    if (m_detail)
    {
        if (m_pTarget == NULL)
        {
            msg = "Type = Detail " + type +
                  ", Interval = " + interval +
                  ", Heading = " + m_desc +
                  ", Text = " + m_query;
        }
        else
        {
            msg = "Type = Detail " + type +
                  ", Target = " + m_pTarget->m_name +
                  ", Interval = " + interval +
                  ", Heading = " + m_desc +
                  ", Text = " + m_query;
        }
    }
    else
    {
        char delta[4];
        if (m_delta)
        {
            strcpy(delta, "Yes");
        }
        else
        {
            strcpy(delta, "No");
        }
        if (m_pTarget == NULL)
        {
            msg = "Type = Simple " + type +
                  ", Interval = " + interval +
                  ", Delta = " + delta +
                  ", Heading = " + m_desc +
                  ", Text = " + m_query;
        }
        else
        {
            msg = "Type = Simple " + type +
                  ", Target = " + m_pTarget->m_name +
                  ", Interval = " + interval +
                  ", Delta = " + delta +
                  ", Heading = " + m_desc +
                  ", Text = " + m_query;
        }
    }
    OtaAddTrace(CYR_OTA,TRACE_CONFIG,msg);
}
